(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure CeXL1Abs =
  struct
    (* This entire data structure represents the Naked CeXL syntax *)
    type label = string
    type tyvarname = string
    type tycon = string
    type longtycon = string
    type longvid = string
    type vid = string
    type strid = string
    type longstrid = string

    (* Restrictions for type variables *)
    datatype 'a restrtypat =
      RstrTyPat of {args : {name : tyvarname, pat : 'a restrtypat, ann : 'a} list,
                    name : longtycon,
                    ann : 'a}
    
    datatype 'a restr =
      RstrExclude of {labs : label list, ann : 'a}
    | RstrTyPats of {pats : 'a restrtypat list, ann : 'a}
    | RstrLongVid of {longVid : longvid, ann : 'a}

    type 'a crestr = 'a restr list

    (* Types *)
    datatype fieldstatus =
      Absent
    | Present
    | Optional of tyvarname

    datatype 'a ty =
      TyVar of {name : tyvarname, ann : 'a} 
    | TyFun of {from : 'a ty, to : 'a ty, ann : 'a}
    | TyCon of {args : {name : tyvarname, ty : 'a ty, ann : 'a} list,
                longTyCon : longtycon, ann : 'a}
    | TyRecord of {fields : {status : fieldstatus,
                             lab : label,
                             ty : 'a ty,
                             ann : 'a} list,
                   ext : tyvarname option,
                   ann : 'a}
    
    (* Type parameters with restrictions on type variables *)
    type 'a typarams =
              {tyVars : tyvarname list,
               crestr : 'a crestr,
               ann : 'a} list                     

    (* Patterns *)
    datatype 'a pat =
      PatWild of {ann : 'a}
    | PatConst of {const : KsiAbs.scon, ann : 'a}
    | PatLongVid of {longVid : longvid, ann : 'a}
    | PatRecord of {fields : {isOpt : bool,
                              lab : label,
                              pat : 'a pat,
                              ann : 'a} list,
                    isExt : bool,
                    ext : 'a pat option,
                    ann : 'a}
    | PatCon of {longVid : longvid, pat : 'a pat, ann : 'a}
    | PatTyped of {pat : 'a pat, ty : 'a ty, ann : 'a}
    | PatLayered of {vid : vid, pat : 'a pat, ty : 'a ty option, ann : 'a}

    (* Expressions *)
    datatype 'a exp =
      ExpTyped of {exp : 'a exp, ty : 'a ty, ann : 'a}
    | ExpApp of {funExp : 'a exp, argExp : 'a exp, ann : 'a}
    | ExpFun of {matches : {pat : 'a pat,
                            exp : 'a exp,
                            ann : 'a} list,
                 ann : 'a}
    | ExpCase of {exp : 'a exp,
                  matches : {pat : 'a pat,
                             exp : 'a exp,
                             ann : 'a} list,
                  ann : 'a}
    | ExpFieldCase of {exp : 'a exp,
                       tyVar : tyvarname,
                       absExp : 'a exp,
                       prePat : 'a pat,
                       preExp : 'a exp,
                       tyOut : 'a ty,
                       ann : 'a}
    | ExpHandle of {exp : 'a exp,
                    matches : {pat : 'a pat,
                               exp : 'a exp,
                               ann : 'a} list,
                    ann : 'a}
    | ExpRaise of {exp : 'a exp, ann : 'a}
    | ExpConst of {const : KsiAbs.scon, ann : 'a}
    | ExpLongVid of {longVid : longvid, ann : 'a}
    | ExpRecord of {fields : {isOpt : bool,
                              lab : label,
                              exp : 'a exp,
                              ann : 'a} list,
                    ext : 'a exp option,
                    ann : 'a}
    | ExpLet of {decs : 'a dec list, exp : 'a exp, ann : 'a}

    (* Declarations *)
    and 'a dec =
      DecVal of {tyParams : 'a typarams,
                 isRec : bool,
                 binds : {pat : 'a pat,
                          exp : 'a exp,
                          ann : 'a} list,
                 ann : 'a}
    | DecRes of {vid : vid, crestr : 'a crestr, ann : 'a}
    | DecTy of     {binds : {tyParams : 'a typarams,
                             tyCon : tycon,
                             ty : 'a ty,
                             ann : 'a} list,
                    ann : 'a}
    | DecDataty of {binds : {tyParams : 'a typarams,
                             tyCon : tycon,
                             conbinds : {vid : vid,
                                         ty : 'a ty option,
                                         ann : 'a} list,
                             ann : 'a} list,
                    ann : 'a}
    | DecDatatyRepl of {tyCon : tycon, longTyCon : longtycon, ann : 'a}
    | DecExc of {binds : {vid : vid,
                          ty : 'a ty option,
                          longVid : longvid option,
                          ann : 'a} list,
                 ann : 'a}
                
    (* Simple nested structures *)
    datatype 'a strdec =
      SDecDec of 'a dec
    | SDecStruct of {strId : strid, strDecs : 'a strdec list, ann : 'a}
    | SDecStructRepl of {strId : strid, longStrId : longstrid, ann : 'a}

    (* Necessary for the reduction from Naked CeXL to xi-Calculus and
       for declaring the built-in environments. *)
    datatype idstatus = SV | SC | SE


    (* These last functions directly reduce
       Full CeXL syntax to Naked CeXL syntax.
       This is only the part of this reduction, which is actually needed
       by the parser (i.e. which the parser doesn't reduce itself). *)
    
    exception InternalError

    (* Reduces the syntax: () *)
    fun makeExpUnitValue ann =
          ExpRecord {fields = nil, ext = NONE, ann = ann}

    (* Reduces the syntax: #lab  (vid must be a fresh variable) *)
    fun makeSelectLab lab vid ann =
          ExpFun
            {matches =
               [{pat = PatRecord {fields = [{isOpt = false,
                                             lab = lab,
                                             pat = PatLongVid {longVid = vid,
                                                               ann = ann},
                                             ann = ann}],
                                  isExt = true,
                                  ext = NONE,
                                  ann = ann},
                 exp = ExpLongVid {longVid = vid, ann = ann},
                 ann = ann}],
             ann = ann}

    (* Reduces the syntax: (exp_1, ... , exp_n) *)
    fun makeExpTuple [exp] ann =
          exp
      | makeExpTuple exps ann =
          let
            (* Lexicographical ordering is not the same as the numerical
               string ordering in tuples. *)
            val (_, fields) =  foldl (fn (exp, (n, flds)) =>
                                        (n + 1,
                                         {isOpt = false,
                                          lab = Int.toString n,
                                          exp = exp,
                                          ann = ann} :: flds))
                                     (1, nil) exps

            fun ltField ({lab, exp, isOpt, ann},
                         {lab = lab', exp = _, isOpt = _, ann = _}) =
                  String.< (lab, lab')                                 

            val sortedFields = Util.quickSort ltField fields
          in
            ExpRecord
              {fields = sortedFields,
               ext = NONE,
               ann = ann}
          end

    (* Reduces the syntax: (exp_1; ... ; exp_n; exp) *)
    fun makeExpSeq [exp] ann =
          exp
      | makeExpSeq (exp::exps) ann =
          ExpCase {exp = exp,
                   matches = [{pat = PatWild {ann = ann},
                               exp = makeExpSeq exps ann,
                               ann = ann}],
                   ann = ann}
      | makeExpSeq nil ann =
          (* This function must only be called if there
             is at least one expression in the list! *)
          raise InternalError

    (* Reduces the syntax: exp1 infvid exp2 *)
    fun makeExpInfix exp1 infvid exp2 ann =
          ExpApp {funExp = ExpLongVid {longVid = infvid, ann = ann},
                  argExp = makeExpTuple [exp1, exp2] ann,
                  ann = ann}

    (* Reduces the syntax: [exp_1, ... , exp_n] *)
    fun makeExpList exps ann =
          foldr (fn (exp, acc) => makeExpInfix exp "::" acc ann)
                (ExpLongVid {longVid = "nil", ann = ann}) exps

    (* Reduces the syntax: if exp then exp1 else exp2 *)
    fun makeExpIf exp exp1 exp2 ann =
          ExpCase
            {exp = exp,
             matches = [{pat = PatLongVid {longVid = "true", ann = ann},
                         exp = exp1,
                         ann = ann},
                        {pat = PatLongVid {longVid = "false", ann = ann},
                         exp = exp2,
                         ann = ann}],
            ann = ann}

    (* Reduces the syntax: lExp orelse rExp *)
    fun makeExpOrelse lExp rExp ann =
          makeExpIf lExp (ExpLongVid {longVid = "true", ann = ann}) rExp ann

    (* Reduces the syntax: lExp andalso rExp *)
    fun makeExpAndalso lExp rExp ann =
          makeExpIf lExp rExp (ExpLongVid {longVid = "false", ann = ann}) ann

    (* Reduces the syntax: let decs in exp_1 ; ... ; exp_n end *)
    fun makeExpLetWithExps decs exps ann =
          ExpLet {decs = decs,
                  exp = makeExpSeq exps ann,
                  ann = ann}

    (* Reduces the syntax: while exp do decs in iterBinds end
       (vid must be a fresh variable) *)
    fun makeExpFWhile exp decs iterBinds vid ann =
          let
            fun ltIterBind ({vid, exp, ann}, {vid = vid', exp = _, ann = _}) =
                  String.< (vid, vid')                                 

            val iterBinds = Util.quickSort ltIterBind iterBinds
            
            fun expVid vid = ExpLongVid {longVid = vid, ann = ann}
            fun patVid vid = PatLongVid {longVid = vid, ann = ann}

            val iterFields = map (fn {vid, exp, ann} =>
                                    {isOpt = false,
                                     lab = vid,
                                     exp = expVid vid,
                                     ann = ann})
                                 iterBinds
            
            val expFields = map (fn {vid, exp, ann} =>
                                   {isOpt = false,
                                    lab = vid,
                                    exp = exp,
                                    ann = ann})
                                iterBinds

            val patFields = map (fn {vid, exp, ann} =>
                                   {isOpt = false,
                                    lab = vid,
                                    pat = patVid vid,
                                    ann = ann})
                                iterBinds

            val iterRecord = ExpRecord
                               {fields = iterFields,
                                ext = NONE,
                                ann = ann}

            val expRecord = ExpRecord
                              {fields = expFields,
                               ext = NONE,
                               ann = ann}
            
            fun recursiveCall record =
                  ExpApp {funExp = expVid vid,
                          argExp = record,
                          ann = ann}
          in
            ExpLet
              {decs =
                 [DecVal
                    {tyParams = nil,
                     isRec = true,
                     binds =
                       [{pat = patVid vid,
                         exp = ExpFun
                                 {matches =
                                    [{pat = PatRecord
                                              {fields = patFields,
                                               isExt = false,
                                               ext = NONE,
                                               ann = ann},
                                      exp = makeExpIf exp
                                              (ExpLet
                                                 {decs = decs,
                                                  exp = recursiveCall
                                                          expRecord,
                                                  ann = ann})
                                              iterRecord ann,
                                      ann = ann}],
                                  ann = ann},
                         ann = ann}],
                     ann = ann}],
               exp = recursiveCall iterRecord,
               ann = ann}
          end

    (* Reduces the syntax: () *)
    fun makePatUnitValue ann =
          PatRecord {fields = nil, isExt = false, ext = NONE, ann = ann}

    (* Reduces the syntax: (pat_1, ... , pat_n) *)
    fun makePatTuple [pat] ann =
          pat
      | makePatTuple pats ann =
          let
            (* Lexicographical ordering is not the same as the numerical
               string ordering in tuples. *)
            val (_, fields) =  foldl (fn (pat, (n, flds)) =>
                                        (n + 1,
                                         {isOpt = false,
                                          lab = Int.toString n,
                                          pat = pat,
                                          ann = ann} :: flds))
                                     (1, nil) pats

            fun ltField ({lab, pat, isOpt, ann},
                         {lab = lab', pat = _, isOpt = _, ann = _}) =
                  String.< (lab, lab')                                 

            val sortedFields = Util.quickSort ltField fields
          in
            PatRecord
              {fields = sortedFields,
               isExt = false,
               ext = NONE,
               ann = ann}
          end

    (* Reduces the syntax: pat1 infvid pat2 *)
    fun makePatInfix pat1 infvid pat2 ann =
          PatCon {longVid = infvid,
                  pat = makePatTuple [pat1, pat2] ann,
                  ann = ann}

    (* Reduces the syntax: [pat_1, ... , pat_n] *)
    fun makePatList exps ann =
          foldr (fn (pat, acc) => makePatInfix pat "::" acc ann)
                (PatLongVid {longVid = "nil", ann = ann}) exps

    (* There is no reduction for this pattern row short-hand:
       vid < : ty > < as pat > < , patrow >
       because it's possibly simpler to extend this directly in the parser. *)

    (* Reduces the syntax: (tyvar_1, ... , tyvar_n) *)
    fun makeTyPmsTyVars tyVars ann =
          [{tyVars = tyVars, crestr = nil, ann = ann}]

    (* Reduces the syntax: vid atpat_11 ... atpat_1n < : ty > = exp_1  | ... *)
    fun makeFValBind (fValBind as {vid, pats, ty, exp, ann = _}::_) makeVid ann =
          let
            val vids = map (fn _ => makeVid ()) pats
            val expVids = map (fn vid => ExpLongVid {longVid = vid, ann = ann})
                              vids
            val patVids = map (fn vid => PatLongVid {longVid = vid, ann = ann})
                              vids
            val matches = map (fn {vid, pats, ty = tyOpt, exp, ann} =>
                                 {pat = makePatTuple pats ann,
                                  exp = case tyOpt of
                                          NONE =>
                                            exp
                                        | SOME ty =>
                                            ExpTyped
                                              {exp = exp, ty = ty, ann = ann},
                                  ann = ann})
                              fValBind
            val caseExp = ExpCase {exp = makeExpTuple expVids ann,
                                   matches = matches,
                                   ann = ann}
          in
            {pat = PatLongVid {longVid = vid, ann = ann},
             exp = foldr (fn (patVid, accExp) =>
                            ExpFun {matches =
                                      [{pat = patVid,
                                        exp = accExp,
                                        ann = ann}],
                                    ann = ann})
                         caseExp patVids,
             ann = ann}
          end
      | makeFValBind nil _ _ =
          (* This function may only be called if there
             is at least one function clause in the list! *)
          raise InternalError

    (* Reduces the syntax: < and fvalbind > (i.e. a sequence of fvalbinds)  *)
    fun makeFValBinds fValBinds makeVid ann =
          map (fn fValBind => makeFValBind fValBind makeVid ann) fValBinds

    (* Reduces the syntax: fun tyPms fvalbind *)
    fun makeDecFunTyParams tyPms fValBind makeVid ann =
          DecVal
            {tyParams = tyPms,
             isRec = true,
             binds = makeFValBinds fValBind makeVid ann,
             ann = ann}
    
    (* Reduces the syntax: ty_1 * ... * ty_n *)
    fun makeTyTuple tys ann =
          let
            (* Lexicographical ordering is not the same as the numerical
               string ordering in tuples. *)
            val (_, fields) =  foldl
                                 (fn (ty, (cnt, acc)) =>
                                    (cnt + 1,
                                     {status = Present,
                                      lab = Int.toString cnt,
                                      ty = ty,
                                      ann = ann} :: acc))
                                 (1, nil) tys

            fun ltField ({lab, ty, status, ann},
                         {lab = lab', ty = _, status = _, ann = _}) =
                  String.< (lab, lab')                                 

            val sortedFields = Util.quickSort ltField fields
          in
            TyRecord
              {fields = sortedFields,                                    
               ext = NONE,
               ann = ann}
          end

    (* Reduces the syntax: (ty_1, ... ty_n)
       as part of: (ty_1, ... ty_n) longtycon *)
    fun makeTyUnnamedTyConArgs tys ann =
          let
            val prod = KsiTypeInfer.lexicogTyvarProducer ()
          in
            map (fn ty => {name = prod (), ty = ty, ann = ann}) tys
          end

    (* Reduces the syntax: (typat_1, ... typat_n)
       as part of: (typat_1, ... typat_n) longtycon *)
    fun makeTyPatUnnamedTyConArgs typats ann =
          let
            val prod = KsiTypeInfer.lexicogTyvarProducer ()
          in
            map (fn typat => {name = prod (), pat = typat, ann = ann})
                typats
          end
  end
