(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure NakedToKsi =
  struct
    (* exception InternalError *)

    structure ST = SymbolTable

    type ann = {pos : {fileName : string, fileLoc : {line : int, pos : int}}}

    type veelem = {tyscheme : unit KsiAbs.tyscheme,
                   status : CeXL1Abs.idstatus,
                   vid : KsiAbs.vid}

    type tyfun = {params : {name : KsiAbs.tyvarname,
                            restr : KsiAbs.restr} list,
                  ty : unit KsiAbs.ty}

    (* Generation of type names and constructor names *)
    (* FIXME: Add predeclared names *)
    val generatedTypeNames = ref (ST.empty ()) : unit ST.t ref
    val generatedConNames = ref (ST.empty ()) : unit ST.t ref

    fun generateName refVar original =
          let
            fun iter cnt =
                  let
                    val suggestion = if cnt = 1 then
                                       original
                                     else
                                       original ^ "#" ^ (Int.toString cnt)
                  in
                    case ST.lookup (!refVar) (ST.symbol suggestion) of
                      NONE =>
                        (refVar := ST.insert (ST.symbol suggestion) () (!refVar);
                         suggestion)
                    | SOME _ =>
                        iter (cnt + 1)
                  end
          in
            iter 1
          end

    fun generateTypeName original =
          generateName generatedTypeNames original

    fun generateConName original =
          generateName generatedConNames original

    (* Functions for reporting errors *)
    (* val errorReportedRef = ref 0 *)

    fun reset () =
          ( (* errorReportedRef := 0; *)
           (* FIXME: Add predeclared names *)
           generatedTypeNames := (ST.empty ());
           generatedConNames := (ST.empty ()))

    val subSystemCode = (0wx11 : Word32.word)
    
    val errorReported =
          KsiReportTypeErrors.errorReported

    val reportError =
          KsiReportTypeErrors.reportError subSystemCode
    val reportErrorStrs =
          KsiReportTypeErrors.reportErrorStrs subSystemCode

    val reportErrorFurther =
          KsiReportTypeErrors.reportErrorFurther
    val reportErrorFurtherStrs =
          KsiReportTypeErrors.reportErrorFurtherStrs

    val reportErrorAtPos =
          KsiReportTypeErrors.reportErrorAtPos subSystemCode
    val reportErrorAtPosStrs =
          KsiReportTypeErrors.reportErrorAtPosStrs subSystemCode

    fun reportInternalError errCode errMsg =
          KsiReportTypeErrors.reportInternalError
            subSystemCode errCode errMsg
    fun reportInternalErrorStrs errCode strs =
          KsiReportTypeErrors.reportInternalErrorStrs
            subSystemCode errCode strs

    fun reportInternalErrorAtPos pos errCode errMsg =
          KsiReportTypeErrors.reportInternalErrorAtPos
            subSystemCode pos errCode errMsg
    fun reportInternalErrorAtPosStrs pos errCode strs =
          KsiReportTypeErrors.reportInternalErrorAtPosStrs
            subSystemCode pos errCode strs
    

    (*
    fun errorReported () =
          (!errorReportedRef) > 0

    fun reportError str =
          (errorReportedRef := (!errorReportedRef) + 1;
           print ("#" ^
                  (Int.toString (!errorReportedRef)) ^ ": " ^
                  str ^ "\n"))

    val reportErrorStrs = reportError o String.concat

    fun reportErrorAtPos {fileName, fileLoc = {line, pos}} str =
          (errorReportedRef := (!errorReportedRef) + 1;
           print ("#" ^
                  (Int.toString (!errorReportedRef)) ^
                  " " ^ fileName ^
                  " line " ^ (Int.toString line) ^
                  " pos " ^ (Int.toString pos) ^
                  ": " ^ str ^ "\n"))

    fun reportErrorAtPosStrs pos strs =
          reportErrorAtPos pos (String.concat strs)

    fun reportErrorFurther str =
          (if (!errorReportedRef) <= 0 then
             errorReportedRef := 1
           else
             ();
           print (str ^ "\n"))

    val reportErrorFurtherStrs = reportErrorFurther o String.concat

    fun reportInternalError str =
          (errorReportedRef := (!errorReportedRef) + 1;
           (* We do not print error number on internal errors,
              because internal errors are not supposed to be "normal" -
              so hopefully it signifies to the user that something is off. *)
           print ("Internal error in ksi-translation: " ^ str ^ "\n");
           raise InternalError)

    fun reportInternalErrorStrs strs =
          reportInternalError (String.concat strs)

    fun reportInternalErrorAtPos {fileName, fileLoc = {line, pos}} str =
          (errorReportedRef := (!errorReportedRef) + 1;
           (* We do not print error number on internal errors,
              because internal errors are not supposed to be "normal" -
              so hopefully it signifies to the user that something is off. *)
           print ("Internal error in ksi-translation in file " ^ fileName ^
                  " line " ^ (Int.toString line) ^
                  " pos " ^ (Int.toString pos) ^
                  ": " ^ str ^ "\n");
           raise InternalError)

    fun reportInternalErrorAtPosStrs pos strs =
          reportInternalErrorAtPos pos (String.concat strs)
    *)
    
    (* Make abstract syntax utilities *)
    fun makePatPresentCon ann =
          KsiAbs.PatCon
            {tag = "present",
             tyScheme = KsiBuiltinStatic.presentTyScheme,
             ann = ann}

    fun makeExpPresentCon ann =
          KsiAbs.ExpCon
            {tag = "present",
             tyScheme = KsiBuiltinStatic.presentTyScheme,
             ann = ann}

    val dummyTyScheme = KsiBuiltinStatic.unitTyScheme

    (* Utilities *)
    fun isLongVid vid =
          Substring.foldl (fn (c, acc) => acc orelse c = #".")
                          false (Substring.full vid)
    
    (* Remove first strId from a structure prefix variable *)
    fun inn str =
          case String.tokens (fn x => x = #".") str of
            nil =>
              (reportInternalError 0wx0101
                 "Empty string given for function inn"; str)
          | [x] =>
              ""
          | x::xs =>
              foldl (fn (x, acc) => acc ^ x ^ ".") "" xs

    (* Environment copying with replacing of prefix *)
    (* The simple brute-force implementation *)
    fun copyAnyEnvReplacePrefix fromPrefix fromEnv toPrefix toEnv =
          let
            val fromPrefixLen = String.size fromPrefix

            fun copyF (ident, accEnv) =
                  if String.isPrefix fromPrefix (ST.name ident) then
                    let
                      val newIdent = toPrefix ^ (String.extract (ST.name ident, fromPrefixLen, NONE))
                    in
                      case ST.lookup fromEnv ident of
                        NONE =>
                          (reportInternalErrorStrs 0wx0102
                             ["Identifier ", ST.name ident,
                              " not found in environment as expected"];
                           accEnv)
                      | SOME value =>
                          ST.insert (ST.symbol newIdent) value accEnv
                    end
                  else
                    accEnv
          in
            foldl copyF toEnv (ST.dimensions fromEnv)
          end

    (* Supposed-to-be-more efficient implementation, taking advantage of
       the fact that we are using a ternary search tree, where getting all
       keys with a certain prefix is quite efficient. But this turned out
       to be less efficient than the brute-force approach! *)
    (*
    fun copyAnyEnvReplacePrefix fromPrefix fromEnv toPrefix toEnv =
          let
            val fromPrefixLen = String.size fromPrefix

            fun copyF ((key, value), accEnv) =
                  let
                    val newIdent = toPrefix ^ (String.extract (ST.name key, fromPrefixLen, NONE))
                  in
                    ST.insert (ST.symbol newIdent) value accEnv
                  end
          in
            foldl copyF toEnv
                  (ST.lookupKeyedPrefix fromEnv fromPrefix)
          end
    *)

    fun copyEnvReplacePrefix fromPrefix (fromVE, fromRE, fromTE, fromDelta)
                             toPrefix   (toVE, toRE, toTE, toDelta) =
          (copyAnyEnvReplacePrefix fromPrefix fromVE toPrefix toVE,
           copyAnyEnvReplacePrefix fromPrefix fromRE toPrefix toRE,
           copyAnyEnvReplacePrefix fromPrefix fromTE toPrefix toTE,
           toDelta)

    (* Combine restrictions *)
    fun combineRestr _ (KsiAbs.RstrNone, r) =
          r
      | combineRestr _ (r, KsiAbs.RstrNone) =
          r
      | combineRestr _ (KsiAbs.RstrExclude {labs}, KsiAbs.RstrExclude {labs = l}) =
          KsiAbs.RstrExclude
            {labs = KsiTypeInfer.makeUnionSortedSets op< op= labs l}
      | combineRestr pos (r1 as (KsiAbs.RstrTyPats {pats}),
                          r2 as (KsiAbs.RstrTyPats {pats = p})) =
          (case KsiTypeInfer.makeIntersectSet KsiTypeInfer.eqTyPat pats p of
             nil =>
               (reportErrorAtPosStrs pos 0wx0103
                      ["Combination of restriction: ",
                       KsiPrettyPrint.restrToStr r1,
                       " and restriction: ",
                       KsiPrettyPrint.restrToStr r2,
                       " is illegal, ",
                       "since the intersection of restrictions is empty"];
                KsiAbs.RstrTyPats {pats = pats})
           | combinedPats =>
               KsiAbs.RstrTyPats {pats = combinedPats}
          )
      | combineRestr pos (r1, r2) =
          (reportErrorAtPosStrs pos 0wx0104
                 ["Incompatiple combination of restriction: ",
                  KsiPrettyPrint.restrToStr r1,
                  " with restriction: ",
                  KsiPrettyPrint.restrToStr r2];
           KsiAbs.RstrNone)

    (* Check that a type does not have any free types (i.e. meta variables) *)
    fun noFreeTypes ty =
          let
            val metas = KsiTypeInfer.metavarsTy ty
          in
            case ST.dimensions metas of
              nil =>
                true
            | _ =>
                false
          end

    fun noTyVars ty =
          let
            val tyVars = KsiTypeInfer.tyvarsTy ty
          in
            case ST.dimensions tyVars of
              nil =>
                true
            | _ =>
                false
          end

    (* The closure operation on a datatype constructor *)
    fun closureConsT errPos tyvProd (_, {tyscheme = {params, ty, ann}, status, vid}) =
          let
            (* Tidy up type for closure *)
            val ty' = KsiTypeInfer.closureTidy (tyvProd, ST.empty ()) ty
          in
            (* We keep the parameter list *)
            {tyscheme = {params = params,
                         ty = ty',
                         ann = ann},
             status = status,
             vid = vid}
          end

    (* The closure operation for a constructor value environment *)
    fun closureConsVE errPos vE =
          let
            val tyvProd = KsiTypeInfer.makeTyvarProducer (ST.empty ()) nil
            val vE' = ST.map (fn {tyscheme = {params, ty, ann}, status, vid} =>
                                {tyscheme = {params = params,
                                             ty = KsiTypeInfer.closureTidyNoInst ty,
                                             ann = ann},
                                 status = status,
                                 vid = vid}) vE
          in
            ST.mapi (closureConsT errPos tyvProd) vE'
          end

    (* Utility *)
    fun ltTyRecordField ({lab, ty, status, ann},
                           {lab = lab2, ty = _, status = _, ann = _}) =
            String.< (lab, lab2)

    fun eqTyRecordField ({lab : string, ty, status, ann},
                           {lab = lab2, ty = _, status = _, ann = _}) =
            lab = lab2

    fun ltPatRecordField ({lab, pat, isOpt, ann},
                           {lab = lab2, pat = _, isOpt = _, ann = _}) =
            String.< (lab, lab2)

    fun eqPatRecordField ({lab : string, pat, isOpt, ann},
                           {lab = lab2, pat = _, isOpt = _, ann = _}) =
            lab = lab2

    fun ltExpRecordField ({lab, exp, isOpt, ann},
                           {lab = lab2, exp = _, isOpt = _, ann = _}) =
            String.< (lab, lab2)

    fun eqExpRecordField ({lab : string, exp, isOpt, ann},
                           {lab = lab2, exp = _, isOpt = _, ann = _}) =
            lab = lab2

    fun checkSortedDistinct ltF eqF getLab elems msg {pos} =
            if KsiTypeInfer.isSortedDistinct ltF elems then
              true
            else
              (reportErrorAtPosStrs pos 0wx0105
                 [msg, KsiPrettyPrint.labsToStr
                         (map getLab (Util.getSortedDuplicates eqF elems))];
               false)

    (* Translation *)
    fun transTyVar (VE, RE, TE, Delta) name =
          (case ST.lookup Delta (ST.symbol name) of
             NONE =>
               KsiAbs.TyVar {name = name, restr = KsiAbs.RstrNone, ann = ()}
           | SOME restr =>
               KsiAbs.TyVar {name = name, restr = restr, ann = ()}
          )

    fun transTyFieldStatus E (CeXL1Abs.Absent) =
          KsiBuiltinStatic.absentTy
      | transTyFieldStatus E (CeXL1Abs.Present) =
          KsiBuiltinStatic.presentTy
      | transTyFieldStatus E (CeXL1Abs.Optional tyvar) =
          transTyVar E tyvar
    
    fun transTy E (CeXL1Abs.TyVar {name, ann}) =
          transTyVar E name
      | transTy E (CeXL1Abs.TyRecord {fields, ext, ann = ann as {pos}}) =
          (case fields of
             nil =>
               (case ext of
                  NONE =>
                    KsiAbs.TyEmptyRecord {ann = ()}
                | (SOME _) =>
                    (reportErrorAtPosStrs pos 0wx0106
                       ["An empty record type may not be extensible"];
                     KsiAbs.TyEmptyRecord {ann = ()})
               )
           | _::_ =>
               (checkSortedDistinct ltTyRecordField eqTyRecordField #lab
                  fields "Duplicate labels in record type: " ann;
                KsiAbs.TyExtRecord
                  {fields = map (transTyField E) fields,
                   ext = case ext of
                           NONE =>
                             KsiAbs.TyEmptyRecord {ann = ()}
                         | SOME tyvar =>
                             transTyVar E tyvar,
                   ann = ()
                  }
               )
          )
      | transTy E (CeXL1Abs.TyFun {from, to, ann}) =
          KsiAbs.TyFun
            {from = transTy E from,
             to = transTy E to,
             ann = ()}
      | transTy E (CeXL1Abs.TyCon {args, longTyCon = "?", ann = ann as {pos}}) =
          (* The field constructor is handled specially *)
          (case args of
             nil =>
               KsiAbs.TyField
                 {present = KsiTypeInfer.newMetaTy (),
                  ty = KsiTypeInfer.newMetaTy (),
                  ann = ()}
           | [{name = "'a", ty, ann}] =>
               KsiAbs.TyField
                 {present = transTy E ty,
                  ty = KsiTypeInfer.newMetaTy (),
                  ann = ()} 
           | [{name = "'b", ty, ann}] =>
               KsiAbs.TyField
                 {present = KsiTypeInfer.newMetaTy (),
                  ty = transTy E ty,
                  ann = ()} 
           | [{name = "'a", ty = ty1, ann = _}, {name = "'b", ty = ty2, ann}] =>
               KsiAbs.TyField
                 {present = transTy E ty1,
                  ty = transTy E ty2,
                  ann = ()} 
           | [{name = "'b", ty = ty2, ann}, {name = "'a", ty = ty1, ann = _}] =>
               KsiAbs.TyField
                 {present = transTy E ty1,
                  ty = transTy E ty2,
                  ann = ()} 
           | _ =>
               (reportErrorAtPosStrs pos 0wx0107
                  ["Illegal arguments for the infix ",
                   "type constructor ? for optional field types"];
                KsiAbs.TyField
                  {present = KsiTypeInfer.newMetaTy (),
                   ty = KsiTypeInfer.newMetaTy (),
                   ann = ()})
          )
      | transTy (E as (VE, RE, TE, Delta)) (CeXL1Abs.TyCon {args, longTyCon, ann = {pos}}) =
          (* General type function application where longTyCon <> "?" *)
          let
            val ksiArgs = map (fn {name, ty, ann} =>
                                 {name = name, ty = transTy E ty, ann = ann})
                              args
          in
            case ST.lookup TE (ST.symbol longTyCon) of
              NONE =>
                (reportErrorAtPosStrs pos 0wx0108
                   ["Unbound type constructor: ", longTyCon];
                 KsiTypeInfer.newMetaTy ())
            | (SOME ({params, ty} : tyfun)) => 
                let
                  (* First create fresh types with appropriate restrictions *)
                  val envParams = KsiTypeInfer.paramsToFreshMetaVars
                                    (ST.empty ()) params
                  
                  (* We only require that Dom args \subseteq Dom params,
                     so traversing all args and finding each param is enough. *)
                  val _ = List.app (fn {name = n, ty = t, ann = _} =>
                                      (case ST.lookup envParams (ST.symbol n) of
                                         NONE =>
                                           reportErrorAtPosStrs pos 0wx0109
                                                 ["Cannot instantiate type variable ",
                                                  n, " since it it does not exist in ",
                                                  "the parameter list for the type constructor ",
                                                  longTyCon]
                                       | SOME ty =>
                                           (KsiTypeInfer.unifyReport pos (ty, t); ())
                                      )
                                   )
                                   ksiArgs
                in
                  (* Instantiate type variables in ty with all the fresh
                     meta variables in envParams. *)
                  KsiTypeInfer.instantiateTyVars envParams ty
                end
          end

    and transTyField E {status, lab, ty, ann} =
          {lab = lab,
           ty = KsiAbs.TyField
                  {present = transTyFieldStatus E status,
                   ty = transTy E ty,
                   ann = ()},
           ann = ()}

    fun transTyPat (E as (_, _, TE, _)) (CeXL1Abs.RstrTyPat {args, name, ann = {pos}}) =
          (case ST.lookup TE (ST.symbol name) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx010a
                  ["Undeclared type: ", name];
                KsiAbs.RstrTyPat
                  {args = nil,
                   name = name,
                   tags = nil})
           | SOME ({params, ty} : tyfun) =>
               (case ty of
                  KsiAbs.TyCon {args = tyConArgs, name = tyName, tags, ann = _} =>
                    if tyName = "exn" then
                      (reportErrorAtPosStrs pos 0wx010b
                         ["Restriction type exn may not be used ",
                          "in a sequence of disjoint restrictions"];
                       KsiAbs.RstrTyPat
                         {args = nil,
                          name = tyName,
                          tags = tags})
                    else
                      let
                        val xiArgs = transTyPatArgs E args

                        fun eqXiArgTyConArg ({name, pat}, {name = tyName, ty}) =
                              name = tyName

                        fun checkTyConArgXiArg ({name = tyName, ty}, {name = argName, pat}) =
                              (case ty of
                                 KsiAbs.TyVar {name = _, restr, ann} =>
                                   if KsiTypeInfer.tyPatRespectsRestr (pat, restr) then
                                     tyName = argName
                                   else
                                     (reportErrorAtPosStrs pos 0wx010c
                                        ["Type pattern does not respect restriction ",
                                         "of datatype parameter in constructor ",
                                         tyName];
                                      tyName = argName)
                               | _ =>
                                   (* This should never happen *)
                                   (reportInternalErrorStrs 0wx010d
                                      ["datatype ", tyName,
                                       " is not parametrized with type ",
                                       "variables as expected"];
                                    tyName = argName)
                              )
                      in
                        (* Check restrictions of tyConArgs against xiArgs
                           and check that the number of parameters of the datatype
                           and the number of arguments in the type pattern are
                           the same. *)
                        if KsiTypeInfer.isSubset eqXiArgTyConArg xiArgs tyConArgs then
                          if KsiTypeInfer.isSubset checkTyConArgXiArg tyConArgs xiArgs then
                            KsiAbs.RstrTyPat
                              {args = xiArgs,
                               name = tyName,
                               tags = tags}
                          else
                            (reportErrorAtPosStrs pos 0wx010e
                               ["datatype ", tyName,
                                " has parameters which are not supplied as arguments ",
                                "to type pattern"];
                                KsiAbs.RstrTyPat
                                  {args = xiArgs,
                                   name = tyName,
                                   tags = tags})
                        else
                          (reportErrorAtPosStrs pos 0wx010f
                             ["Type pattern has more parameters supplied than ",
                              "there are parameters for the datatype ", tyName];
                           KsiAbs.RstrTyPat
                              {args = xiArgs,
                               name = tyName,
                               tags = tags})
                      end
                | _ =>
                    (reportErrorAtPosStrs pos 0wx0110
                       ["The type ", name, " is not a datatype as expected"];
                     KsiAbs.RstrTyPat
                       {args = nil,
                        name = name,
                        tags = nil})
               )
          )

    and transTyPatArgs E args =
          map (fn {name, pat, ann} =>
                 {name = name,
                  pat = transTyPat E pat})
              args

    fun transRestr E (CeXL1Abs.RstrExclude {labs, ann}) =
          (* We assume (but verfy to be sure) that the parser has sorted labs. *)
          (checkSortedDistinct op< op= (fn x => x)
             labs "Duplicate labels in restriction: " ann;
           KsiAbs.RstrExclude {labs = labs}
          )
      | transRestr E (CeXL1Abs.RstrTyPats {pats, ann = {pos}}) =
          (case pats of
             nil =>
               (reportErrorAtPosStrs pos 0wx0111
                  ["Sequence of disjoint restriction types may not be empty"];
                KsiAbs.RstrNone)
           | _ =>
               KsiAbs.RstrTyPats {pats = map (transTyPat E) pats}
          )
      | transRestr (VE, RE, TE, Delta) (CeXL1Abs.RstrLongVid {longVid, ann = {pos}}) =
          (case ST.lookup RE (ST.symbol longVid) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx0112
                  ["Undeclared restriction: ", longVid];
                KsiAbs.RstrNone)
           | SOME restr =>
               restr
          )

    fun transCRestr pos E (restr::crestr) =
          let
            fun foldF (re, combRe) =
                  let
                    val newRe = transRestr E re
                    val _ = KsiTypeInfer.checkRestr newRe
                  in
                    combineRestr pos (newRe, combRe)
                  end

            val initRe = transRestr E restr
            val _ = KsiTypeInfer.checkRestr initRe
          in
            foldl foldF initRe crestr
          end
      | transCRestr pos _ nil =
          (reportErrorAtPosStrs pos 0wx0113
             ["Sequence of combined restrictions may not be empty"];
           KsiAbs.RstrNone)

    fun epsTyParams (E as (_, _, _, Delta)) tyParams =
          foldl (epsTyParamsF E) Delta tyParams

    and epsTyParamsF E ({tyVars, crestr, ann = ann as {pos}}, accDelta) =
          (case tyVars of
             _::_ =>
               let
                 val restr = case crestr of
                               nil =>
                                 KsiAbs.RstrNone
                             | _ =>
                                 transCRestr pos E crestr

                 fun insertTyVar (tyVar, acc) =
                       ST.insert (ST.symbol tyVar) restr acc
               in
                 foldl insertTyVar accDelta tyVars
               end
           | nil =>
               (reportErrorAtPosStrs pos 0wx0114
                    ["Encountered empty list of type variables in ",
                     "type parameter list"];
                accDelta)
          )

    fun epsPat (E, isRec, _) (CeXL1Abs.PatWild {ann}) =
          E
      | epsPat (E, isRec, _) (CeXL1Abs.PatConst {const, ann}) =
          E
      | epsPat ((VE, RE, TE, Delta), isRec, prefix)
                 (CeXL1Abs.PatLongVid {longVid, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol longVid) of
             NONE =>
               if isLongVid longVid then
                 (reportErrorAtPosStrs pos 0wx0115
                        ["The identifier ", longVid,
                         " is neither a constructor ",
                         "nor an exception and may therefore ",
                         "not be a long identifier"];
                  (VE, RE, TE, Delta))
               else
                 (ST.insert (ST.symbol longVid)
                            {tyscheme = dummyTyScheme,
                             status = CeXL1Abs.SV,
                             vid = prefix ^ longVid}
                            VE,
                  RE, TE, Delta)
           | SOME ({tyscheme, status, vid} : veelem) =>
               (case (status, isRec) of
                  (CeXL1Abs.SC, false) =>
                    (VE, RE, TE, Delta)
                | (CeXL1Abs.SE, false) =>
                    (* FIXME: We just assume that the type scheme for
                       exceptions is exn as expected *)
                    (VE, RE, TE, Delta)
                | _ =>
                    (* This is what we do if either status=SV or isRec=true *)
                    if isLongVid longVid then
                      (reportErrorAtPosStrs pos 0wx0116
                             ["The identifier ", longVid,
                              " is neither a constructor ",
                              "nor an exception and may therefore ",
                              "not be a long identifier"];
                       (VE, RE, TE, Delta))
                    else
                      (ST.insert (ST.symbol longVid)
                                 {tyscheme = dummyTyScheme,
                                  status = CeXL1Abs.SV,
                                  vid = prefix ^ longVid}
                                 VE,
                       RE, TE, Delta)
                
               )
          )
      | epsPat (E, isRec, prefix) (CeXL1Abs.PatRecord {fields, isExt, ext, ann}) =
          let
            val E' = foldl (fn ({lab, isOpt, pat, ann}, acc) =>
                              epsPat (acc, isRec, prefix) pat)
                           E fields
          in
            case ext of
              NONE =>
                E'
            | SOME patExt =>
                epsPat (E', isRec, prefix) patExt
          end
      | epsPat ((E as (VE, RE, TE, Delta)), isRec, prefix)
                 (CeXL1Abs.PatCon {longVid, pat, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol longVid) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx0117
                  ["Unbound pattern constructor variable ", longVid];
                epsPat (E, isRec, prefix) pat)
           | SOME ({tyscheme = {params, ty, ann = _}, status, vid}) =>
               (case status of
                  CeXL1Abs.SV =>
                    (reportErrorAtPosStrs pos 0wx0118
                           ["The identifier ", longVid,
                            " occurs as a constructor pattern but ",
                            "is neither a constructor ",
                            "nor an exception"];
                     epsPat (E, isRec, prefix) pat)
                | CeXL1Abs.SC =>
                    epsPat (E, isRec, prefix) pat
                | CeXL1Abs.SE =>
                    (* FIXME: We just assume that the type scheme is
                       ty -> exn as expected *)
                    epsPat (E, isRec, prefix) pat
               )
          )
      | epsPat E (CeXL1Abs.PatTyped {pat, ty, ann}) =
          epsPat E pat
      | epsPat (E, isRec, prefix) (CeXL1Abs.PatLayered {vid, pat, ty = tyOpt, ann = ann as {pos}}) =
          let
            val (VE, RE, TE, Delta) = epsPat (E, isRec, prefix) pat
            val E' = (ST.insert (ST.symbol vid)
                                {tyscheme = dummyTyScheme,
                                 vid = prefix ^ vid,
                                 status = CeXL1Abs.SV}
                                VE,
                      RE, TE, Delta)
          in
            case ST.lookup VE (ST.symbol vid) of
              NONE =>
                E'
            | SOME ({tyscheme = _, status, vid = vid'}) =>
                (case (status, isRec) of
                   (CeXL1Abs.SC, false) =>
                     (reportErrorAtPosStrs pos 0wx0119
                              ["The identifier ", vid,
                               " occurs in a layered pattern and must ",
                               "therefore be a variable, but ",
                               "it is declared as a constructor"];
                      E')
                 | (CeXL1Abs.SE, false) =>
                     (reportErrorAtPosStrs pos 0wx011a
                              ["The identifier ", vid,
                               " occurs in a layered pattern and must ",
                               "therefore be a variable, but ",
                               "it is declared as an exception"];
                      E')
                 | _ =>
                     (* This is what we do if either status=SV or isRec=true *)
                     E'
                 
                )
          end

    fun transPat E (CeXL1Abs.PatWild {ann}) =
          KsiAbs.PatWild {ann = ann}
      | transPat E (CeXL1Abs.PatConst {const, ann}) =
          KsiAbs.PatSCon {scon = const, ann = ann}
      | transPat ((VE, RE, TE, Delta), prefix)
                 (CeXL1Abs.PatLongVid {longVid, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol longVid) of
             NONE =>
               (* epsPat is always called on a pattern before transPat,
                  so unbound variables _should_ already be reported.
                  So we only report something if nothing at all has
                  been reported - and then it's an internal error! *)
               if errorReported () then
                 KsiAbs.PatVar {name = longVid, ann = ann}
               else
                 (reportInternalErrorAtPosStrs pos 0wx011b
                    ["Unbound pattern variable ", longVid];
                  KsiAbs.PatVar {name = longVid, ann = ann})
           | SOME ({tyscheme = tyscheme as {params, ty, ann = _}, status, vid} : veelem) =>
               (case status of
                  CeXL1Abs.SV =>
                    if isLongVid longVid then
                      (reportErrorAtPosStrs pos 0wx011c
                             ["The identifier ", longVid,
                              " is neither a constructor ",
                              "nor an exception and may therefore ",
                              "not be a long identifier"];
                       KsiAbs.PatVar {name = longVid, ann = ann})
                    else
                      KsiAbs.PatVar {name = vid, ann = ann}
                | CeXL1Abs.SC =>
                    KsiAbs.PatCon {tag = vid, tyScheme = tyscheme, ann  = ann}
                | CeXL1Abs.SE =>
                    (* FIXME: We just assume that the type scheme is exn as expected *)
                    KsiAbs.PatEx {tag = vid, ann = ann}
               )
          )
      | transPat E (CeXL1Abs.PatRecord {fields = nil, isExt, ext, ann}) =
          (* Empty record *)
          (if isExt then
             reportInternalError 0wx011d
               "An empty record may not be extensible"
           else
             (case ext of
                NONE =>
                  ()
              | SOME _ =>
                  reportInternalError 0wx011e
                    "An empty record may not have an extension"
             );
           KsiAbs.PatEmptyRecord {ann = ann})
      | transPat E (CeXL1Abs.PatRecord {fields, isExt, ext, ann}) =
          (* Non-empty record *)
          (checkSortedDistinct ltPatRecordField eqPatRecordField #lab
             fields "Duplicate labels in record pattern: " ann;
           KsiAbs.PatExtRecord {fields = map (transPatField E) fields,
                                ext = if isExt then
                                        case ext of
                                          NONE =>
                                            KsiAbs.PatWild {ann = ann}
                                        | SOME patExt =>
                                            transPat E patExt
                                      else
                                        KsiAbs.PatEmptyRecord {ann = ann},
                                ann = ann}
          )
      | transPat (E as ((VE, RE, TE, Delta), prefix))
                 (CeXL1Abs.PatCon {longVid, pat, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol longVid) of
             NONE =>
               (* epsPat is always called on a pattern before transPat,
                  so unbound variables _should_ already be reported.
                  So we only report something if nothing at all has
                  been reported - and then it's an internal error! *)
               if errorReported () then
                 transPat E pat
               else
                 (reportInternalErrorAtPosStrs pos 0wx011f
                    ["Unbound pattern constructor variable ", longVid];
                  transPat E pat)
           | SOME ({tyscheme = tyscheme as {params, ty, ann = _}, status, vid}) =>
               (case status of
                  CeXL1Abs.SV =>
                    (reportErrorAtPosStrs pos 0wx0120
                           ["The identifier ", longVid,
                            " occurs as a constructor pattern but ",
                            "is neither a constructor ",
                            "nor an exception"];
                     transPat E pat)
                | CeXL1Abs.SC =>
                    KsiAbs.PatParamCon
                      {tag = vid,
                       tyScheme = tyscheme,
                       pat = transPat E pat,
                       ann = ann}
                | CeXL1Abs.SE =>
                    (* FIXME: We just assume that the type scheme is
                       ty -> exn as expected *)
                    KsiAbs.PatParamEx
                      {tag = vid,
                       ty = case ty of
                              KsiAbs.TyFun {from, to, ann = _} =>
                                from
                            | _ =>
                                (reportErrorAtPosStrs pos 0wx0121
                                   ["Type scheme for exception constructor ",
                                    "is not of the expected form"];
                                 ty),
                       pat = transPat E pat,
                       ann = ann}
               )
          )
      | transPat (E, prefix) (CeXL1Abs.PatTyped {pat, ty, ann}) =
          KsiAbs.PatTyped
            {pat = transPat (E, prefix) pat,
             ty = transTy E ty,
             ann = ann}
      | transPat (E as (VE, RE, TE, Delta), prefix) (CeXL1Abs.PatLayered {vid, pat, ty = tyOpt, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol vid) of
             NONE =>
               (* epsPat is always called on a pattern before transPat,
                  so unbound variables _should_ already be reported.
                  So we only report something if nothing at all has
                  been reported - and then it's an internal error! *)
               if errorReported () then
                 transPat (E, prefix) pat
               else
                 (reportInternalErrorAtPosStrs pos 0wx0122
                    ["Unbound layered pattern variable ", vid];
                  transPat (E, prefix) pat)
           | SOME ({tyscheme = {params, ty = _, ann = _}, status, vid = vid'}) =>
               (case status of
                  CeXL1Abs.SV =>
                    (* The layered variable is not a constructor or exception *)
                    (case tyOpt of
                       NONE =>
                         KsiAbs.PatLayered
                           {vid = vid',
                            pat = transPat (E, prefix) pat,
                            ann = ann}
                     | SOME ty =>
                         KsiAbs.PatLayered
                           {vid = vid',
                            pat = KsiAbs.PatTyped
                                    {pat = transPat (E, prefix) pat,
                                     ty = transTy E ty,
                                     ann = ann},
                            ann = ann}
                    )
                | _ =>
                    (reportErrorAtPosStrs pos 0wx0123
                             ["The identifier ", vid,
                              " occurs in a layered pattern and must ",
                              "therefore be a variable, but ",
                              "it is declared as a constructor ",
                              "or an exception"];
                     transPat (E, prefix) pat)
               )
          )

    and transPatField E {isOpt, lab, pat, ann} =
          {lab = lab,
           pat = if isOpt then
                   transPat E pat
                 else
                   KsiAbs.PatOptField
                     {present = makePatPresentCon ann,
                      pat = transPat E pat,
                      ann = ann},
           ann = ann}

    fun transExp E (CeXL1Abs.ExpConst {const, ann}) =
          KsiAbs.ExpSCon {scon = const, ann = ann}
      | transExp ((VE, RE, TE, Delta), prefix)
                 (CeXL1Abs.ExpLongVid {longVid, ann = ann as {pos}}) =
          (case ST.lookup VE (ST.symbol longVid) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx0124
                  ["Unbound variable ", longVid];
                KsiAbs.ExpVar {name = longVid, ann = ann})
           | SOME ({tyscheme = tyscheme as {params, ty, ann = _}, status, vid} : veelem) =>
               (case status of
                  CeXL1Abs.SV =>
                    KsiAbs.ExpVar {name = vid, ann = ann}
                | CeXL1Abs.SC =>
                    KsiAbs.ExpCon {tag = vid, tyScheme = tyscheme, ann = ann}
                | CeXL1Abs.SE =>
                    (case params of
                       nil =>
                         KsiAbs.ExpEx {tag = vid, ty = ty, ann = ann}
                     | _ =>
                         (reportInternalErrorStrs 0wx0125
                            ["Exception constructor ", longVid,
                             " does not have a closed type scheme"];
                          KsiAbs.ExpEx {tag = vid, ty = ty, ann = ann})
                    )
               )
          )
      | transExp E (CeXL1Abs.ExpRecord {fields = nil, ext, ann}) =
          (* Empty record *)
          (case ext of
             SOME _ =>
               reportInternalError 0wx0126
                 "An empty record may not be extensible"
           | NONE =>
               ();
           KsiAbs.ExpEmptyRecord {ann = ann})
      | transExp E (CeXL1Abs.ExpRecord {fields, ext, ann}) =
          (* Non-empty record *)
          (checkSortedDistinct ltExpRecordField eqExpRecordField #lab
             fields "Duplicate labels in record expression: " ann;
           KsiAbs.ExpExtRecord {fields = map (transExpField E) fields,
                                ext = case ext of
                                        NONE =>
                                          KsiAbs.ExpEmptyRecord {ann = ann}
                                      | SOME expExt =>
                                          transExp E expExt,
                                ann = ann}
          )
      | transExp E (CeXL1Abs.ExpLet {decs, exp, ann}) =
          transDecs E decs exp
      | transExp (E as ((VE, RE, TE, Delta), prefix))
                 (CeXL1Abs.ExpFieldCase
                    {exp, tyVar,
                     absExp, prePat, preExp,
                     tyOut, ann = ann as {pos}}) =
          (case ST.lookup Delta (ST.symbol tyVar) of
             NONE =>
               ()
           | SOME restr =>
               if KsiTypeInfer.eqRestrict
                    (restr, KsiTypeInfer.makeAbsentPresentRestr ()) then
                 ()
               else
                 reportErrorAtPosStrs pos 0wx0126
                   ["The type variable ", tyVar, " in fieldcase expression ",
                    "has a restriction which is not absent | present as required"];
           let
             val E' = epsPat ((VE, RE, TE, Delta), false, prefix) prePat
           in
             KsiAbs.ExpFieldCase
               {exp = transExp E exp,
                tyvar = tyVar,
                absExp = transExp E absExp,
                prePat = transPat (E', prefix) prePat,
                preExp = transExp (E', prefix) preExp,
                tyOut = transTy (VE, RE, TE, Delta) tyOut,
                ann = ann}
           end
          )
      | transExp E (CeXL1Abs.ExpApp {funExp, argExp, ann = ann as {pos}}) =
          KsiAbs.ExpApp
            {funExp = transExp E funExp,
             argExp = transExp E argExp,
             ann = ann}
      | transExp (E, prefix) (CeXL1Abs.ExpTyped {exp, ty, ann}) =
          KsiAbs.ExpTyped
            {exp = transExp (E, prefix) exp,
             ty = transTy E ty,
             ann = ann}
      | transExp E (CeXL1Abs.ExpRaise {exp, ann}) =
          KsiAbs.ExpRaise
            {exp = transExp E exp,
             ann = ann}
      | transExp E (CeXL1Abs.ExpHandle {exp, matches, ann}) =
          KsiAbs.ExpHandle
            {exp = transExp E exp,
             matches = transMatches E matches,
             ann = ann}
      | transExp E (CeXL1Abs.ExpCase {exp, matches, ann}) =
          KsiAbs.ExpCase
            {exp = transExp E exp,
             matches = transMatches E matches,
             ann = ann}
      | transExp E (CeXL1Abs.ExpFun {matches, ann}) =
          let
            val vid = CeXL1FrontEndCommon.makeFreshVid ()
          in
            KsiAbs.ExpFun
              {param = vid,
               exp = KsiAbs.ExpCase
                       {exp = KsiAbs.ExpVar {name = vid, ann = ann},
                        matches = transMatches E matches,
                        ann = ann},
               ann = ann}
          end
      
    and transExpField E {isOpt, lab, exp, ann} =
          {lab = lab,
           exp = if isOpt then
                   transExp E exp
                 else
                   KsiAbs.ExpOptField
                     {present = makeExpPresentCon ann,
                      exp = transExp E exp,
                      ann = ann},
           ann = ann}

    and transMatches E matches =
          map (transMRule E) matches

    and transMRule (E, prefix) {pat, exp, ann} =
          let
            val E' = epsPat (E, false, prefix) pat
          in
            {pat = transPat (E', prefix) pat,
             exp = transExp (E', prefix) exp,
             ann = ann}
          end

    (* The recursion with exp, fnXiExp and xiExp' is quite tricky here :-) *)
    (* transDecs accumulates a Naked CeXL expression in its parameter exp.
       transDec generates a function (fnXiExp), which when applied
       to a xi-Calculus expression (xiExp') yields a new xi-Calculus
       expression, consisting of the code in fnXiExp with xiExp' at the end. *)
    and transDecs (E, prefix) (dec::decs) exp =
          let
            val E' = epsDec (E, prefix, false) dec
            val fnXiExp = transDec (E, E', prefix) dec
            val xiExp' = transDecs (E', prefix) decs exp
          in
            fnXiExp xiExp'
          end
      | transDecs E nil exp =
          transExp E exp

    and transDec (E as (VE, RE, TE, _), E' as (VE', RE', TE', _), prefix)
                 (CeXL1Abs.DecVal {tyParams, isRec, binds, ann}) =
          if isRec then
            let
              val Delta2' = epsTyParams E' tyParams
              val binds' = transValBinds ((VE', RE', TE', Delta2'),
                                        (VE', RE', TE', Delta2'), prefix)
                                        binds
                                       
            in
              fn xiExp =>
              KsiAbs.ExpLetRec
                {binds = binds',
                 exp = xiExp,
                 ann = ann}
            end
          else
            let
              val Delta2 = epsTyParams E tyParams
              val Delta2' = epsTyParams E' tyParams
              val binds' = transValBinds ((VE', RE', TE', Delta2'),
                                        (VE, RE, TE, Delta2), prefix)
                                       binds
            in
              fn xiExp =>
              KsiAbs.ExpLet
                {binds = binds',
                 exp = xiExp,
                 ann = ann}
            end
      | transDec (E, E', prefix)
                 (CeXL1Abs.DecExc {binds, ann}) =
          let
            val binds' = map (transExBind E') binds
          in
          (fn xiExp =>
             KsiAbs.ExpLetEx
               {binds = binds',
                exp = xiExp,
                ann = ann})
          end
      | transDec _ (CeXL1Abs.DecRes {vid, crestr, ann}) =
          (fn xiExp => xiExp)
      | transDec _ (CeXL1Abs.DecTy {binds, ann}) =
          (fn xiExp => xiExp)
      | transDec _ (CeXL1Abs.DecDataty {binds, ann}) =
          (fn xiExp => xiExp)
      | transDec _ (CeXL1Abs.DecDatatyRepl {tyCon, longTyCon, ann}) =
          (fn xiExp => xiExp)
    
    and transValBinds E binds =
          map (transValBind E) binds

    and transValBind (E1, E2, prefix) {pat, exp, ann} =
          {pat = transPat (E1, prefix) pat,
           exp = transExp (E2, prefix) exp,
           ann = ann}
          
    and transExBind (VE, RE, TE, Delta) {vid, ty, longVid, ann = ann as {pos}} =
          (case ST.lookup VE (ST.symbol vid) of
             NONE =>
               (reportInternalErrorAtPosStrs pos 0wx0127
                  ["Unbound declared exception ", vid];
                {tag = vid, ty = KsiBuiltinStatic.unitTy, ann = ann})
           | SOME ({tyscheme = tyscheme as {params, ty, ann = _},
                    status, vid = vid'} : veelem) =>
               (case status of
                  CeXL1Abs.SE =>
                    (case params of
                       nil =>
                         {tag = vid', ty = ty, ann = ann}
                     | _ =>
                         (reportInternalErrorStrs 0wx0128
                            ["Exception constructor ", vid,
                             " does not have a closed type scheme"];
                          {tag = vid', ty = ty, ann = ann})
                    )
                | _ =>
                    (reportInternalErrorStrs 0wx0129
                       ["Variable ", vid,
                        " is not bound to an exception as expected"];
                     {tag = vid', ty = ty, ann = ann})
               )
          )

    and epsDec (E as (VE, RE, TE, Delta), prefix, isTopLevel)
               (CeXL1Abs.DecVal {tyParams, isRec, binds, ann}) =
          let
            val Delta' = epsTyParams E tyParams
            val (VE', RE', TE', Delta'') =
                   epsValBind ((VE, RE, TE, Delta'), isRec, prefix) binds
          in
            (VE', RE', TE', Delta)
          end
      | epsDec (E as (VE, RE, TE, Delta), prefix, isTopLevel)
               (CeXL1Abs.DecRes {vid, crestr, ann = {pos}}) =
          (case crestr of
             nil =>
               (VE, ST.insert (ST.symbol vid) KsiAbs.RstrNone RE,
                TE, Delta)
           | _ =>
               (VE, ST.insert (ST.symbol vid) (transCRestr pos E crestr) RE,
                TE, Delta)
          )
      | epsDec (E as (VE, RE, TE, Delta), prefix, isTopLevel)
               (CeXL1Abs.DecTy {binds, ann}) =
          epsTypBind (E, prefix) binds
      | epsDec (E as (VE, RE, TE, Delta), prefix, isTopLevel)
               (CeXL1Abs.DecDataty {binds, ann = {pos}}) =
          let
            val (TE', binds') = epsDatBindPreRecurse
                                  ((ST.empty (), RE, TE, Delta), prefix) binds

            val (VE', RE', _, Delta') =
                   epsDatBind ((ST.empty (), RE, TE', Delta), prefix) binds'
          in
            (ST.append (VE, closureConsVE pos VE'),
             RE', TE', Delta')
          end
      | epsDec (E as (VE, RE, TE, Delta), prefix, isTopLevel)
               (CeXL1Abs.DecDatatyRepl {tyCon, longTyCon, ann = {pos}}) =
          (case ST.lookup TE (ST.symbol longTyCon) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx012a
                  ["Unbound datatype ", longTyCon];
                E)
           | SOME (tyFun as {params, ty} : tyfun) =>
               (case ty of
                  KsiAbs.TyCon _ =>
                    (VE, RE, ST.insert (ST.symbol tyCon) tyFun TE, Delta)
                | _ =>
                    (reportErrorAtPosStrs pos 0wx012b
                       ["Type constructor ", longTyCon,
                        " which is to be replicated is not a datatype"];
                     E)
               )
          )
      | epsDec (E, prefix, isTopLevel)
               (CeXL1Abs.DecExc {binds, ann}) =
          epsExBind (E, prefix, isTopLevel) binds

    and epsValBind (E, isRec, prefix) ({pat, exp, ann}::binds) =
          epsValBind (epsPat (E, isRec, prefix) pat, isRec, prefix) binds
      | epsValBind (E, _, _) nil =
          E

    and epsTypBind (E, prefix) binds =
          foldl (epsTypBindF prefix) E binds

    and epsTypBindF prefix ({tyParams, tyCon, ty, ann = {pos}}, (VE, RE, TE, Delta)) =
          let
            val Delta' = epsTyParams (VE, RE, TE, ST.empty ()) tyParams
            val ty' = transTy (VE, RE, TE, Delta') ty

            val _ = case KsiTypeInfer.checkTySpec pos Delta' ty' of
                      KsiTypeInfer.NoErr =>
                        if noFreeTypes ty' then
                          ()
                        else
                          reportErrorAtPosStrs pos 0wx012c
                            ["Declaration of the type ", tyCon,
                             " may not contain free types (i.e. meta variables):\n  ",
                             KsiPrettyPrint.tyToStr ty']
                    | KsiTypeInfer.NewErr =>
                        reportErrorAtPosStrs pos 0wx012d
                          ["Declaration of the type ", tyCon,
                           " does not form a valid type:\n  ",
                           KsiPrettyPrint.tyToStr ty']
                    | KsiTypeInfer.ReportedErr =>
                        reportErrorFurtherStrs
                          ["Declaration of the type ", tyCon,
                           " does not form a valid type:\n  ",
                           KsiPrettyPrint.tyToStr ty']

            val ksiTyParams = KsiTypeInfer.tyvarsToParams Delta'
          in
            (VE, RE,
             ST.insert (ST.symbol tyCon) {params = ksiTyParams, ty = ty'} TE,
             Delta)
          end

    and epsDatBindPreRecurse (E as (VE, RE, TE, Delta), prefix) binds =
          (* Notice that the resulting binds will be reversed here,
             but it shouldn't matter *)
          foldl (epsDatBindPreRecurseF (E, prefix)) (TE, nil) binds

    and epsDatBindPreRecurseF ((VE, RE, _, Delta), prefix)
                              ({tyParams, tyCon, conbinds, ann = ann as {pos}}, (TE, accBinds)) =
          let
            val tyName = generateTypeName (prefix ^ tyCon)
            val NN = epsConNames prefix conbinds
            val Delta' = epsTyParams (VE, RE, TE, ST.empty ()) tyParams
            val ksiTyParams = KsiTypeInfer.tyvarsToParams Delta'
            val ksiConArgs = KsiTypeInfer.tyvarsToConArgs Delta'
            val tags = map ST.name (ST.dimensions NN)
            val conTy = KsiAbs.TyCon
                          {args = ksiConArgs,
                           name = tyName,
                           tags = tags,
                           ann = ()}
          in
            (ST.insert (ST.symbol tyCon) {params = ksiTyParams, ty = conTy} TE,
             {ksiTyParams = ksiTyParams,
              newDelta = Delta',
              conTy = conTy,
              tyCon = tyCon,
              tyName = tyName,
              conbinds = conbinds,
              NN = NN,
              ann = ann}
                ::accBinds)
          end

    and epsDatBind (E, prefix) binds =
          foldl (epsDatBindF prefix) E binds

    and epsDatBindF prefix ({ksiTyParams, newDelta, conTy, tyCon, tyName,
                             conbinds, NN, ann = {pos}},
                            (VE, RE, TE, Delta)) =
          (epsConBind ((VE, RE, TE, newDelta), (prefix, NN, ksiTyParams, conTy))
                      conbinds,
           RE, TE, Delta)

    and epsConBind (E as (VE, RE, TE, Delta), params) binds =
          foldl (epsConBindF (E, params)) VE binds

    and epsConBindF (E as (VE, RE, TE, Delta), (prefix, NN, ksiTyParams, conTy))
                    ({vid, ty, ann = {pos}}, accVE) =
          let
            val ty'' = case ty of
                         NONE =>
                           conTy
                       | SOME ty =>
                           let
                             val ty' = transTy (VE, RE, TE, Delta) ty

                             val _ =  case KsiTypeInfer.checkTySpec pos Delta ty' of
                                        KsiTypeInfer.NoErr =>
                                          if noFreeTypes ty' then
                                            ()
                                          else
                                            reportErrorAtPosStrs pos 0wx012e
                                              ["Declaration of the datatype constructor ", vid,
                                               " may not contain free types (i.e. meta variables):\n  ",
                                               KsiPrettyPrint.tyToStr ty']
                                      | KsiTypeInfer.NewErr =>
                                          reportErrorAtPosStrs pos 0wx012f
                                            ["Declaration of the datatype constructor ", vid,
                                             " does not form a valid type:\n  ",
                                             KsiPrettyPrint.tyToStr ty']
                                      | KsiTypeInfer.ReportedErr =>
                                          reportErrorFurtherStrs
                                            ["Declaration of the datatype constructor ", vid,
                                             " does not form a valid type:\n  ",
                                             KsiPrettyPrint.tyToStr ty']
                           in
                             KsiAbs.TyFun
                               {from = ty', to = conTy, ann = ()}
                           end
          in
            case ST.lookup NN (ST.symbol vid) of
              NONE =>
                (reportInternalErrorAtPosStrs pos 0wx0130
                   ["Constructor ", vid,
                    " not found where it is supposed to be!"];
                 accVE)
            | SOME vid' =>
                ST.insert (ST.symbol vid)
                          {tyscheme =
                             {params = ksiTyParams, ty = ty'', ann = ()},
                           status = CeXL1Abs.SC,
                           vid = vid'}
                          accVE
          end

    and epsConNames prefix conbinds =
          foldl (epsConNamesF prefix) (ST.empty ()) conbinds

    and epsConNamesF prefix ({vid, ty, ann}, NN) =
          ST.insert (ST.symbol vid) (generateConName (prefix ^ vid)) NN

    and epsExBind (E, prefix, isTopLevel) binds =
          foldl (epsExBindF (prefix, isTopLevel)) E binds

    and epsExBindF (prefix, isTopLevel)
                   ({vid, ty, longVid, ann = {pos}}, (VE, RE, TE, Delta)) =
          let
            val ty'' = case ty of
                         NONE =>
                           KsiBuiltinStatic.exnTy
                       | SOME ty =>
                           let
                             val ty' = transTy (VE, RE, TE, Delta) ty

                             val _ =  if noFreeTypes ty' then
                                        if (not isTopLevel) orelse noTyVars ty' then
                                          ()
                                        else
                                          reportErrorAtPosStrs pos 0wx0131
                                            ["Declaration of the exception ", vid,
                                             " may not contain type variables:\n  ",
                                             KsiPrettyPrint.tyToStr ty']
                                      else
                                        reportErrorAtPosStrs pos 0wx0132
                                          ["Declaration of the exception ", vid,
                                           " may not contain free types (i.e. meta variables):\n  ",
                                           KsiPrettyPrint.tyToStr ty']
                           in
                             KsiAbs.TyFun
                               {from = ty',
                                to = KsiBuiltinStatic.exnTy,
                                ann = ()}
                           end
          in
            case longVid of
              NONE =>
                (ST.insert (ST.symbol vid)
                   {tyscheme = {params = nil, ty = ty'', ann = ()},
                    status = CeXL1Abs.SE,
                    vid = prefix ^ vid}
                   VE,
                 RE, TE, Delta)
            | SOME longVid =>
                (case ST.lookup VE (ST.symbol longVid) of
                   NONE =>
                     (reportErrorAtPosStrs pos 0wx0133
                        ["The exception ", longVid, " is not declared!"];
                      (VE, RE, TE, Delta))
                 | SOME {tyscheme, status, vid = vid'} =>
                     (case status of
                        CeXL1Abs.SE =>
                          (ST.insert (ST.symbol vid)
                             {tyscheme = tyscheme, status = status, vid = vid'}
                             VE,
                           RE, TE, Delta)
                      | _ =>
                          (reportErrorAtPosStrs pos 0wx0134
                             ["The variable ", longVid, " is not declared as an exception!"];
                           (VE, RE, TE, Delta))
                     )
                )
          end

    fun epsPackPat (VE, E', longVid) (CeXL1Abs.PatWild {ann}) =
          VE
      | epsPackPat (VE, E', longVid) (CeXL1Abs.PatConst {const, ann}) =
          VE
      | epsPackPat (VE, (VE', _, _, _), longVid')
                   (CeXL1Abs.PatLongVid {longVid, ann = ann as {pos}}) =
          (case ST.lookup VE' (ST.symbol (inn(longVid') ^ longVid)) of
             NONE =>
               if isLongVid longVid then
                 (reportErrorAtPosStrs pos 0wx0135
                        ["The identifier ", longVid,
                         " is neither a constructor ",
                         "nor an exception and may therefore ",
                         "not be a long identifier"];
                  VE)
               else
                 (reportErrorAtPosStrs pos 0wx0136
                        ["Error while packing the variable ", longVid,
                         " into the structure ", longVid'];
                  VE)
           | SOME (v as {tyscheme, status, vid} : veelem) =>
               (case status of
                  CeXL1Abs.SV =>
                    if isLongVid longVid then
                      (reportErrorAtPosStrs pos 0wx0137
                             ["The identifier ", longVid,
                              " is neither a constructor ",
                              "nor an exception and may therefore ",
                              "not be a long identifier"];
                       VE)
                    else
                      ST.insert (ST.symbol (longVid' ^ longVid)) v VE
                | CeXL1Abs.SC =>
                    VE
                | CeXL1Abs.SE =>
                    VE
               )
          )
      | epsPackPat (VE, E' as (VE', _, _, _), longVid)
                   (CeXL1Abs.PatRecord {fields, isExt, ext, ann}) =
          let
            val VE'' = foldl (fn ({lab, isOpt, pat, ann}, acc) =>
                              epsPackPat (acc, E', longVid) pat)
                             VE fields
          in
            case ext of
              NONE =>
                VE''
            | SOME patExt =>
                epsPackPat (VE'', E', longVid) patExt
          end
      | epsPackPat (VE, E', longVid')
                   (CeXL1Abs.PatCon {longVid, pat, ann = ann as {pos}}) =
          epsPackPat (VE, E', longVid') pat
      | epsPackPat E (CeXL1Abs.PatTyped {pat, ty, ann}) =
          epsPackPat E pat
      | epsPackPat (VE, E' as (VE', _, _, _), longVid)
                   (CeXL1Abs.PatLayered {vid, pat, ty, ann = ann as {pos}}) =
          let
            val VE'' = case ST.lookup VE (ST.symbol (inn(longVid) ^ vid)) of
                         NONE =>
                           (reportErrorAtPosStrs pos 0wx0138
                              ["Error while packing the layered variable ", vid,
                               " into the structure ", longVid];
                            VE)
                       | SOME (v as {tyscheme, status, vid = vid'}) =>
                           (case status of
                              CeXL1Abs.SV =>
                                ST.insert (ST.symbol (longVid ^ vid)) v VE
                            | CeXL1Abs.SC =>
                                VE
                            | CeXL1Abs.SE =>
                                VE
                           )
          in
            epsPackPat (VE'', E', longVid) pat
          end

    fun epsPackValBindF (E', longVid) ({pat, exp, ann}, (VE, RE, TE, Delta)) =
          let
            val VE' = epsPackPat (VE, E', longVid) pat
          in
            (VE', RE, TE, Delta)
          end

    fun epsPackValBind (E, E', longVid) binds =
          foldl (epsPackValBindF (E', longVid)) E binds

    fun epsPackTypBindF ((_, _, TE', _), longVid)
                        ({tyParams, tyCon, ty, ann = {pos}}, (VE, RE, TE, Delta)) =
          (case ST.lookup TE' (ST.symbol (inn(longVid) ^ tyCon)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx0139
                  ["Error while packing the type ",
                   tyCon, " into the structure ", longVid];
                (VE, RE, TE, Delta))
           | SOME tyFun =>
               (VE, RE,
                ST.insert (ST.symbol (longVid ^ tyCon)) tyFun TE,
                Delta)
          )

    fun epsPackTypBind (E, E', longVid) binds =
          foldl (epsPackTypBindF (E', longVid)) E binds

    fun epsPackExBindF ((VE', _, _, _), longVid)
                       ({vid, ty, longVid = _, ann = {pos}}, (VE, RE, TE, Delta)) =
          (case ST.lookup VE' (ST.symbol (inn(longVid) ^ vid)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx013a
                  ["Error while packing the exception ",
                   vid, " into the structure ", longVid];
                (VE, RE, TE, Delta))
           | SOME (v as {tyscheme, status, vid = vid'}) =>
               (case status of
                  CeXL1Abs.SE =>
                    (ST.insert (ST.symbol (longVid ^ vid)) v VE,
                     RE, TE, Delta)
                | _ =>
                    (* The identifier is not bound to an exception,
                       so don't store anything. *)
                    (VE, RE, TE, Delta)
               )
          )

    fun epsPackExBind (E, E', longVid) binds =
          foldl (epsPackExBindF (E', longVid)) E binds

    fun epsPackConBindF (VE', longVid)
                       ({vid, ty, ann = {pos}}, VE) =
          (case ST.lookup VE' (ST.symbol (inn(longVid) ^ vid)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx013b
                  ["Error while packing the datatype constructor ",
                   vid, " into the structure ", longVid];
                VE)
           | SOME (v as {tyscheme, status, vid = vid'}) =>
               (case status of
                  CeXL1Abs.SC =>
                    ST.insert (ST.symbol (longVid ^ vid)) v VE
                | _ =>
                    (* The identifier is not bound to a constructor,
                       so don't store anything. *)
                    VE
               )
          )

    fun epsPackConBind (VE', longVid) (conbinds, VE) =
          foldl (epsPackConBindF (VE', longVid)) VE conbinds

    fun epsPackDatBindF ((VE', _, TE', _), longVid)
                        ({tyParams, tyCon, conbinds, ann = {pos}}, (VE, RE, TE, Delta)) =
          (case ST.lookup TE' (ST.symbol (inn(longVid) ^ tyCon)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx013c
                  ["Error while packing the datatype ",
                   tyCon, " into the structure ", longVid];
                (VE, RE, TE, Delta))
           | SOME tyFun =>
               (epsPackConBind (VE', longVid) (conbinds, VE),
                RE,
                ST.insert (ST.symbol (longVid ^ tyCon)) tyFun TE,
                Delta)
          )

    fun epsPackDatBind (E, E', longVid) binds =
          foldl (epsPackDatBindF (E', longVid)) E binds

    fun epsPackDec (E, E', longVid) (CeXL1Abs.DecVal {tyParams, isRec, binds, ann}) =
          epsPackValBind (E, E', longVid) binds
      | epsPackDec ((VE, RE, TE, Delta), (_, RE', _, _), longVid)
                   (CeXL1Abs.DecRes {vid, crestr, ann = {pos}}) =
          (case ST.lookup RE' (ST.symbol (inn(longVid) ^ vid)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx013d
                  ["Error while packing the restriction ",
                   vid, " into the structure ", longVid];
                (VE, RE, TE, Delta))
           | SOME restr =>
               (VE,
                ST.insert (ST.symbol (longVid ^ vid)) restr RE,
                TE, Delta)
          )
      | epsPackDec (E, E', longVid) (CeXL1Abs.DecTy {binds, ann}) =
          epsPackTypBind (E, E', longVid) binds
      | epsPackDec (E, E', longVid) (CeXL1Abs.DecDataty {binds, ann}) =
          epsPackDatBind (E, E', longVid) binds
      | epsPackDec ((VE, RE, TE, Delta), (_, _, TE', _), longVid)
                   (CeXL1Abs.DecDatatyRepl {tyCon, longTyCon, ann = {pos}}) =
          (case ST.lookup TE' (ST.symbol (inn(longVid) ^ tyCon)) of
             NONE =>
               (reportErrorAtPosStrs pos 0wx013e
                  ["Error while packing the replicated datatype ",
                   tyCon, " into the structure ", longVid];
                (VE, RE, TE, Delta))
           | SOME tyFun =>
               (VE, RE,
                ST.insert (ST.symbol (longVid ^ tyCon)) tyFun TE,
                Delta)
          )   
      | epsPackDec (E, E', longVid) (CeXL1Abs.DecExc {binds, ann}) =
          epsPackExBind (E, E', longVid) binds

    fun epsPackStruct (E, E', longVid) (strdec::strdecs) =
          let
            val E'' = epsPackSDec (E, E', longVid) strdec
          in
            epsPackStruct (E'', E', longVid) strdecs
          end
      | epsPackStruct (E, E', longVid) nil =
          E

    and epsPackSDec (E, E', longVid) (CeXL1Abs.SDecDec dec) =
          epsPackDec (E, E', longVid) dec
      | epsPackSDec (E, E', longVid) (CeXL1Abs.SDecStruct {strId, strDecs, ann}) =
          epsPackStruct (E, E', longVid ^ strId ^ ".") strDecs
      | epsPackSDec (E, E', longVid) (CeXL1Abs.SDecStructRepl {strId, longStrId, ann}) =
          copyEnvReplacePrefix (inn(longVid) ^ strId ^ ".") E' (longVid ^ strId ^ ".") E

    fun transEpsSDec (E, prefix) (CeXL1Abs.SDecDec dec) =
          let
            val E' = epsDec (E, prefix, true) dec
            val fnXiCode = transDec (E, E', prefix) dec
          in
            (fnXiCode, E')
          end
      | transEpsSDec (E, prefix) (CeXL1Abs.SDecStruct {strId, strDecs, ann}) =
          let
            (* val _ = print ("declare structure " ^ (strId) ^ "\n") *)
            val (fnXiCode, E') = transEpsSDecs (E, prefix ^ strId ^ ".") strDecs
            val E'' = epsPackStruct (E, E', strId ^ ".") strDecs
            (* val _ = print ("- declared structure " ^ (strId) ^ "\n") *)
          in
            (fnXiCode, E'')
          end
      | transEpsSDec (E, _) (CeXL1Abs.SDecStructRepl {strId, longStrId, ann}) =
          (fn xiExp => xiExp,
           copyEnvReplacePrefix (longStrId ^ ".") E (strId ^ ".") E)
    
    and transEpsSDecs (E, prefix) (strdec::strdecs) =
          let
            (* FIXME: This illustates some of the performance *)
            (* val _ = print "transEpsSDecs begin" *)
            val (fnXiCode, E') = transEpsSDec (E, prefix) strdec
            (* val _ = print " - transEpsSDec" *)
            val (fnXiCode', E'') = transEpsSDecs (E', prefix) strdecs
            (* val _ = print " - transEpsSDecs end\n" *)
          in
            (fn xiExp => fnXiCode (fnXiCode' xiExp), E'')
          end
      | transEpsSDecs (E, prefix) nil =
          (fn xiExp => xiExp, E)

    (* Translate an entire CeXL program *)
    fun transProgram (E, prefix) strdecs =
          let
            val (fnXiCode, E') = transEpsSDecs (E, prefix) strdecs
            val emptyRecord = KsiAbs.ExpEmptyRecord
                                {ann = {pos = {fileName = "",
                                               fileStartLoc = {line = 0, pos = 0},
                                               fileEndLoc = {line = 0, pos = 0}}}}
          in
            fnXiCode emptyRecord
          end
  end
