(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure Util (* : UTIL *) =
  struct
    exception RangeError (* FIXME: Use general error handling or SML exceptions? *) 

    (* A "finally" clause for handling exceptions (similar to those
       found in Java and Borland's "Delphi" Pascal dialect)
       ought to be part of StandardML in my opinion... *)
    (* But, alas, it is not - so I had to write this method *)
    (*
    fun finalize body final =
          (((body ())
              handle e => (final ();raise e))
           before (ignore (final ())))
    *)
    
    (* FIXME: Better implementation? *)
    fun finalize body final =
          let
            val result = (body ())
                           handle e =>
                             (final (); raise e)
            val _ = final ()
          in
            result
          end

    fun optionMap _ NONE =
          NONE
      | optionMap f (SOME v) =
          SOME (f v)

    fun foldlsep sep f init (x::xs) =
          foldl (fn (e, acc) => f (e, sep acc)) (f (x, init)) xs
      | foldlsep _ _ init nil =
          init

    (* Accumulates a value by folding f left over indices [0, ..., len - 1] *)
    fun indexAccl f init len =
          let
            (* Create list of indices [0, 1, 2, ..., len - 1]
               for fold operation *)
            val indices = List.tabulate (len, fn i => i)
          in
            foldl f init indices
          end

    fun indexAccrImpl f init 0 =
          init
      | indexAccrImpl f init len =
          let
            val nextLen = len - 1
          in
            indexAccrImpl f (f (nextLen, init)) nextLen
          end
          
    (* Accumulates a value by folding f right over indices [0, ..., len - 1] *)
    fun indexAccr f init len =
          if len < 0 then
            raise RangeError
          else
            indexAccrImpl f init len

    fun quickSort _ nil =
          nil
      | quickSort _ [x] =
          [x]
      | quickSort compare (x::xs) =
          let
            fun partition (left, right, nil) =
                  (quickSort compare left) @
                  (x :: quickSort compare right)
              | partition (left, right, e::es) =
                  if compare (e, x) then
                    partition (e::left, right, es)
                  else
                    partition (left, e::right, es)
          in
            partition (nil, nil, xs)
          end

    fun removeSortedDuplicatesImpl eqF prev (x::xs) =
          if eqF (prev, x) then
            removeSortedDuplicatesImpl eqF x xs
          else
            x :: (removeSortedDuplicatesImpl eqF x xs)
      | removeSortedDuplicatesImpl _ _ nil =
          nil

    fun removeSortedDuplicates eqF (x::xs) =
          x :: (removeSortedDuplicatesImpl eqF x xs)
      | removeSortedDuplicates _ nil =
          nil

    fun getSortedDuplicates eqF (x::xs) =
          (* This does not keep the first element, see removeSortedDuplicates *)
          removeSortedDuplicatesImpl (fn arg => not (eqF arg)) x xs
      | getSortedDuplicates _ nil =
          nil

    fun isSortedDistinct ltF (x::xs) =
          #2 (foldl (fn (e, (prev, acc)) => (e, acc andalso ltF (prev, e)))
                    (x, true) xs)
      | isSortedDistinct _ nil =
          true

    fun hasSortedDuplicates eqF (x::xs) =
          #2 (foldl (fn (e, (prev, acc)) => (e, acc orelse eqF (prev, e)))
                    (x, false) xs)
      | hasSortedDuplicates _ nil =
          false

    fun mergeSorted _ nil ys = ys
      | mergeSorted _ xs nil = xs
      | mergeSorted compare (x::xs) (y::ys) =
          if compare (x, y) then
            x :: mergeSorted compare xs (y::ys)
          else
            y :: mergeSorted compare (x::xs) ys
  end

structure TestModuleUtil (* :> TESTER *) =
  struct
    fun finalize () =
          let
            val result1 = ref(false)
            val result2 = ref(false)
            val result3 = ref(false)
            
            val _ = (Util.finalize
              (fn () => raise Match)
              (fn () => result1 := true))
                handle Match => (result2 := true)
            
            val result4 = Util.finalize
                            (fn () => 4)
                            (fn () => result3 := true)
          in
            !result1 andalso !result2 andalso
            !result3 andalso (result4 = 4)
          end  

    fun foldlsep () =
          let
            fun commaSeparate l =
                  Util.foldlsep (fn acc => acc ^ ", ")
                                (fn (e, a) => a ^ e)
                                "" l

            fun emptylist () =
                  "" = commaSeparate []

            fun onelist () =
                  "1" = commaSeparate ["1"]

            fun onetwolist () =
                  "1, 2" = commaSeparate ["1", "2"]

            fun onetofivelist () =
                  "1, 2, 3, 4, 5" = commaSeparate ["1", "2", "3", "4", "5"]
          in
            TestUtil.doTest
              [("emptylist", emptylist),
               ("onelist", onelist),
               ("onetwolist", onetwolist),
               ("onetofivelist", onetofivelist)]
          end

    fun indexAccl () =
          let
            fun intList () =
                  case Util.indexAccl op:: nil 10 of
                    [9,8,7,6,5,4,3,2,1,0] =>
                      true
                  | _ =>
                      false
          in
            TestUtil.doTest
              [("intList", intList)]
          end

    fun indexAccr () =
          let
            fun intList () =
                  case Util.indexAccr op:: nil 10 of
                    [0,1,2,3,4,5,6,7,8,9] =>
                      true
                  | _ =>
                      false
          in
            TestUtil.doTest
              [("intList", intList)]
          end

    fun test () =
          TestUtil.doTest [("finalize", finalize),
                           ("foldlsep", foldlsep),
                           ("indexAccl", indexAccl),
                           ("indexAccr", indexAccr)]
  end 
