(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* This is an implemention of a ternary search tree, as
   described in:
   Dr. Dobb's Journal, April 1998, "Ternary Search Trees"
   It's available online at:
   http://www.ddj.com/articles/1998/9804/9804a/9804a.htm
*)
structure TernaryTree (* :> TERNARY_TREE *) =
  struct
    type index = string
    type dimensions = index list
    datatype traversal = Arbitrary | Alphabetical

    (* The datatype for ternary trees containing data *)
    datatype 'a t =
      Leaf of 'a option
    | Node of Char.char * ('a t) * ('a t) * ('a t) * ('a option)

    exception InternalError
    exception NotImplemented

    (* Supposed to be the default and thus most efficient
       traversal order for the collection datastructure. *)
    val defaultTraversal = Arbitrary

    (* An empty tree *)
    fun empty () =
          Leaf NONE

    (* fLookup : 'a t -> (char * substring) option -> 'a option *)
    
    (* The first 2 pattern cases covers when there are no more characters
       in the key string. Then we have found our data - if there is
       any (remember data is of type 'a option) *)
    fun fLookup (Leaf data) NONE = data
      | fLookup (Node (_, _, _, _, data)) NONE = data

      (* End of tree, but more key characters => nothing found *)
      | fLookup (Leaf _) (SOME _) = NONE

      (* More key characters and more tree data => continue search *)
      | fLookup (Node (foundC, lo, e, hi, data)) (s as (SOME(c, cs))) =
          if Char.< (c, foundC) then
            fLookup lo s
          else if Char.> (c, foundC) then
            fLookup hi s
          else
            fLookup e (Substring.getc cs)

    fun lookup table key =
          fLookup table (Substring.getc (Substring.full key))

    (* Lookup and get all values for keys with the given prefix *)
    fun fLookupKeyedPrefix (Leaf NONE) NONE _ acc =
          acc
      | fLookupKeyedPrefix (Leaf (SOME data)) NONE accKey acc =
          (accKey, data)::acc
      | fLookupKeyedPrefix (Node (foundC, lo, e, hi, data)) NONE accKey acc =
          (* We found our prefix - return everything we found *)
          let
            val acc = case data of
                        NONE =>
                          acc
                      | SOME d =>
                          (accKey, d) :: acc
          in
            fLookupKeyedPrefix lo NONE accKey
              (* accKey ^ foundC gives O(n^2) time complexity, but
                 it doesn't get any faster by accumulating char lists.
                 I have tried it in this implementation! *)
              (fLookupKeyedPrefix e NONE (accKey ^ (str foundC))
                 (fLookupKeyedPrefix hi NONE accKey acc))
          end
      | fLookupKeyedPrefix (Leaf _) (SOME _) _ acc =
          (* End of tree, but more key characters => nothing found *)
          acc
      | fLookupKeyedPrefix (Node (foundC, lo, e, hi, data)) (s as (SOME(c, cs))) accKey acc =
          (* More key characters and more tree data => continue search *)
          if Char.< (c, foundC) then
            fLookupKeyedPrefix lo s accKey acc
          else if Char.> (c, foundC) then
            fLookupKeyedPrefix hi s accKey acc
          else
            (* accKey ^ c gives O(n^2) time complexity, but
               it doesn't get any faster by accumulating char lists.
               I have tried it in this implementation! *)
            fLookupKeyedPrefix e (Substring.getc cs) (accKey ^ (str c)) acc

    fun lookupKeyedPrefix table prefix =
          fLookupKeyedPrefix table (Substring.getc (Substring.full prefix)) "" nil

    fun lookupPrefix table prefix =
          List.map #2 (lookupKeyedPrefix table prefix)

    fun keysWithPrefix table prefix =
          List.map #1 (lookupKeyedPrefix table prefix)

    (* As with fLookup, the first 2 pattern cases convers when there
       are no more characters in the key string. This is where we
       insert the data. Notice that we just overwrite any previous data. *)
    (* Overwriting any previous data is a semantic difference from e.g.
       List.insert and Vector.insert. *)
    fun fInsert (Leaf _) NONE newData =
          Leaf (SOME(newData))
      | fInsert (Node (foundC, lo, e, hi, _)) NONE newData =
          Node (foundC, lo, e, hi, SOME(newData))

      (* End of tree, but more key characters => insert data in middle subtree *)
      | fInsert (Leaf data) (SOME(c, cs)) newData =
          Node (c, empty (),
                   fInsert (empty ()) (Substring.getc cs) newData,
                   empty (), data)

      (* More key characters and more tree data => continue insert *)
      | fInsert (Node (foundC, lo, e, hi, data)) (s as SOME(c, cs)) newData =
          if Char.< (c, foundC) then
            Node (foundC, fInsert lo s newData, e, hi, data)
          else if Char.> (c, foundC) then
            Node (foundC, lo, e, fInsert hi s newData, data)
          else
            Node (foundC, lo,
                          fInsert e (Substring.getc cs) newData,
                          hi, data)

    fun insert key data table =
          fInsert table (Substring.getc (Substring.full key)) data

    fun fDelete (Leaf _) NONE =
          Leaf NONE
      | fDelete (Node (foundC, lo, e, hi, _)) NONE =
          Node (foundC, lo, e, hi, NONE)

      (* End of tree, but more key characters => no data found *)
      | fDelete (Leaf data) (SOME(c, cs)) =
          Leaf data

      (* More key characters and more tree data => continue delete *)
      | fDelete (Node (foundC, lo, e, hi, data)) (s as SOME(c, cs)) =
          if Char.< (c, foundC) then
            Node (foundC, fDelete lo s, e, hi, data)
          else if Char.> (c, foundC) then
            Node (foundC, lo, e, fDelete hi s, data)
          else
            Node (foundC, lo,
                          fDelete e (Substring.getc cs),
                          hi, data)

    fun delete key table =
          fDelete table (Substring.getc (Substring.full key))

    fun substitute key f table =
          (* This can obviously be done more efficiently, but we
             keep it simple and slightly less efficient for now... *)
          case lookup table key of
            NONE =>
              raise Subscript
          | SOME v =>
              insert key (f v) table

    fun tabulate (indices, tabF) =
          List.foldl (fn (index, acc) => insert index (tabF index) acc)
                     (empty ()) indices

    fun tabulateTraversal Arbitrary args =
          tabulate args
      | tabulateTraversal Alphabetical (indices, tabF) =
          (* This would involve sorting the indices
             and then running the normal tabulate. *)
          raise NotImplemented

    fun create (indices, value) =
          tabulate (indices, fn _ => value)
    
    fun sub (aTable, index) =
          case lookup aTable index of
            NONE =>
              raise Subscript
          | SOME v =>
              v
    
    fun rev aTable =
          (* This operation does not make much sense for a
             ternary tree, so we don't do anything *)
          aTable

    fun foldliF f acc (Leaf data) accKey =
          (case data of
             NONE =>
               acc
           | SOME d =>
               f (accKey, d, acc)
          )
      | foldliF f acc (Node (c, lo, e, hi, data)) accKey =
          let
            val accLo = foldliF f acc lo accKey
            val accData = case data of
                            NONE =>
                              accLo
                          | SOME d =>
                              f (accKey, d, accLo)
            val accE = foldliF f accData e (accKey ^ (str c))
            val accHi = foldliF f accE hi accKey
          in
            accHi
          end
          
    fun foldli f init aTable =
          foldliF f init aTable ""

    fun foldl f init aTable =
          foldli (fn (_, v, acc) => f (v, acc)) init aTable

    fun foldlTraversal Arbitrary f init aTable =
          foldl f init aTable
      | foldlTraversal Alphabetical f init aTable =
          raise NotImplemented

    fun foldliTraversal Arbitrary f init aTable =
          foldli f init aTable
      | foldliTraversal Alphabetical f init aTable =
          raise NotImplemented

    (* A quite efficient implementation of dimensions
       (given the efficient implementation of foldli) *)
    fun dimensions aTable =
          List.rev (foldli (fn (i, _, acc) => i::acc) nil aTable)

    fun mapiF f (Leaf data) accKey =
          Leaf
            (case data of
               NONE =>
                 NONE
             | SOME d =>
                 SOME (f (accKey, d))
            )
      | mapiF f (Node (c, lo, e, hi, data)) accKey =
          Node (c,
                mapiF f lo accKey,
                mapiF f e (accKey ^ (str c)),
                mapiF f hi accKey,
                case data of
                  NONE =>
                    NONE
                | SOME d =>
                    SOME (f (accKey, d))
               )
          
    fun mapi f aTable =
          mapiF f aTable ""
      
    fun map f aTable =
          mapi (fn (_, v) => f v) aTable
    
    fun mapTraversal Arbitrary f aTable =
          map f aTable
      | mapTraversal Alphabetical f aTable =
          raise NotImplemented

    fun mapiTraversal Arbitrary f aTable =
          mapi f aTable
      | mapiTraversal Alphabetical f aTable =
          raise NotImplemented

    (* A simple and fairly efficient implementation of append *)
    (* This function has time complexity O(m log n),
       where n is the size of aTable and
       m is the size of (the hopefully smaller) addTable. *)
    fun append (aTable, addTable) =
          let
            fun foldF (index, value, acc) =
                  insert index value acc
          in
            foldli foldF aTable addTable
          end
    
    (* An even more efficient implementation of append:
       time-complexity O(m), where n is the size of 2nd argument. *)
    (* FIXME: However this code is buggy and does not work... *)
    (*
    fun append (Node (oc, olo, oe, ohi, oldData),
                Node (c, lo, e, hi, data)) =
          if Char.< (c, oc) then
            (* This clause may insert data which are never accessed.
               Some kind of culling is necessary to avoid this. *)
            Node (oc,
                  append (olo, Node (c, lo, e, hi, NONE)), (* data surely appended somewhere here *)
                  oe,                (* Surely nothing to append here *)
                  append (ohi, hi),  (* Possibly something to append here *)
                  case data of
                    NONE =>
                      oldData
                  | SOME d =>
                      SOME d)
          else if Char.> (c, oc) then
            (* This clause may insert data which are never accessed.
               Some kind of culling is necessary to avoid this. *)
            Node (oc,
                  append (olo, lo),  (* Possibly something to append here *)
                  oe,                (* Surely nothing to append here *)
                  append (ohi, Node (c, lo, e, hi, NONE)), (* data surely appended somewhere here *)
                  case data of
                    NONE =>
                      oldData
                  | SOME d =>
                      SOME d)
          else
            (* c = oc *)
            Node (oc,
                  append (olo, lo),
                  append (oe, e),
                  append (ohi, hi),
                  case data of
                    NONE =>
                      oldData
                  | SOME data =>
                      SOME data
                 )
      | append (Leaf oldData, Node (c, lo, e, hi, data)) =
          (* The whole subtree appended here without further traversal *)
          Node (c, lo, e, hi,
                case data of
                  NONE =>
                    oldData
                | SOME d =>
                    (* oldData overwritten *)
                    SOME d)
      | append (Node (c, lo, e, hi, oldData), Leaf data) =
          (case data of
             NONE =>
               Node (c, lo, e, hi, oldData)
           | SOME d =>
               (* table in 2nd argument overwrites the first one *)
               Node (c, lo, e, hi, SOME d)
          )
      | append (Leaf oldData, Leaf data) =
          (case data of
             NONE =>
               Leaf oldData
           | SOME d =>
               (* table in 2nd argument overwrites the first one *)
               Leaf (SOME d)
          )
    *)
      
    fun appiF f aTable index =
          case lookup aTable index of
            NONE =>
              raise InternalError
          | SOME v =>
              f (index, v)
              
    fun appi f aTable =
          foldli (fn (i, v, ()) => f (i, v)) () aTable
          
    fun app f aTable =
          (* Can obviously be done more efficiently *)
          appi (fn (_, v) => f v) aTable

    fun appTraversal Arbitrary f aTable =
          app f aTable
      | appTraversal Alphabetical f aTable =
          raise NotImplemented

    fun appiTraversal Arbitrary f aTable =
          appi f aTable
      | appiTraversal Alphabetical f aTable =
          raise NotImplemented

    fun findF f (e, NONE) =
          if f e then
            SOME e
          else
            NONE
      | findF f (_, acc as (SOME _)) =
          acc

    fun find f aTable =
          foldl (findF f) NONE aTable
    
    fun exists f aTable =
          foldl (fn (e, acc) => acc orelse (f e)) false aTable

    fun toList aTable =
          foldl op:: nil aTable

    fun fromList l =
          (* Maybe this doesn't make much sense. We would have
             to invent the keys, but we _could_ do that... *)
          raise NotImplemented
  end