(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003 Anoq of the Sun (alias Johnny Andersen)

   E-mail: anoq at HardcoreProcessing dot com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information. E-mail:
   info at HardcoreProcessing dot com

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure ErrorExceptions =
  struct
    exception Bug of real * string
    exception Error of real * string
    exception Warning of real * string
    exception Output of real * string
  end

structure ErrorStdOut =
  struct
    exception Error
    exception Bug

    fun bug msg = (print (msg ^ "\n"); raise Bug)
    fun error msg = (print (msg ^ "\n"); raise Error)
    fun warning msg = print (msg ^ "\n")
  end

(* This one supports the USER_OUTPUT signature, but has
   support for configuring output. It can be passed as a functor
   argument where an USER_OUTPUT is expected. *)
(* This ought to be a functor so that it can be instantiated,
   since it has reference variables. *)
structure ErrorConfig =
  struct
    
    local
      val bugF = ref (fn (_ : real) => ErrorStdOut.bug : string -> unit)
      val errorF = ref (fn (_ : real) => ErrorStdOut.error : string -> unit)
      val warningF = ref (fn (_ : real) => ErrorStdOut.warning)
      val outputF = ref (fn (_ : real) => print)
    in
      fun bug severity msg =
            ((!bugF) severity msg;
             raise ErrorExceptions.Bug (severity, msg))

      fun error severity msg =
            ((!errorF) severity msg;
             raise ErrorExceptions.Error (severity, msg))

      fun exnToMessage e =
            let
              val msgRef = ref ""
              val _ = Excn.outputUnhandled (fn str => msgRef := (!msgRef) ^ str) e
            in
              !msgRef
            end

      fun addError severity msg e =
            case e of
              (ErrorExceptions.Bug (sev, m)) =>
                 ((!errorF) (Real.max (severity, sev))
                            (msg ^ "\nThe bug which caused this:\n" ^ m);
                  raise ErrorExceptions.Error (severity, msg))
            | (ErrorExceptions.Error (sev, m)) =>
                 ((!errorF) (Real.max (severity, sev))
                            (msg ^ "\nThe error which caused this:\n" ^ m);
                  raise ErrorExceptions.Error (severity, msg))
            | (ErrorExceptions.Warning (sev, m)) =>
                 ((!errorF) (Real.max (severity, sev))
                            (msg ^ "\nThe warning which caused this:\n" ^ m);
                  raise ErrorExceptions.Error (severity, msg))
            | (ErrorExceptions.Output (sev, m)) =>
                 ((!errorF) (Real.max (severity, sev))
                            (msg ^ "\nThe output message which caused this:\n" ^ m);
                  raise ErrorExceptions.Error (severity, msg))
            | e =>
                 ((!errorF) severity
                              (msg ^ "\nThe exception which caused this:\n" ^
                               (exnToMessage e));
                  raise ErrorExceptions.Error (severity, msg))

      fun warning severity msg =
            (!warningF) severity msg

      fun output severity msg =
            (!outputF) severity msg

      fun setBugFunc f =
            bugF := f

      fun setErrorFunc f =
            errorF := f

      fun setWarningFunc f =
            warningF := f

      fun setOutputFunc f =
            outputF := f
    end
  end
