(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiBuiltinDynamic = 
  struct
    structure Err = ErrorConfig
    structure ST = SymbolTable

    fun wrapBinOp binOp (KsiVals.Record fields) =
          (case fields of
             [{lab = "1", value = KsiVals.Const v0},
              {lab = "2", value = KsiVals.Const v1}] =>
               binOp (v0, v1)
           | [{lab = "2", value = KsiVals.Const v1},
              {lab = "1", value = KsiVals.Const v0}] =>
               binOp (v0, v1)
           | _ =>
               (* Try to provoke an error from binOp, and
                  if not successful raise a bug exception. *)
               ((binOp (KsiAbs.SCString "", KsiAbs.SCInt 0);
                 Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "wrapBinOp (1)"))
                   handle e =>
                     (* We always handle the exception and reraise as error. *)
                     (Err.addError 1.0 ("KsiBuiltinDynamic: " ^ "wrapBinOp (1)") e))
          )
      | wrapBinOp _ _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "wrapBinOp (2)")

    fun wrapConstBinOp binOp v =
          wrapBinOp (KsiVals.Const o binOp) v

    fun implOpOnNum opW _ _ (KsiAbs.SCWord w0, KsiAbs.SCWord w1) =
          KsiAbs.SCWord (opW (w0, w1))
      | implOpOnNum _ opI _ (KsiAbs.SCInt i0, KsiAbs.SCInt i1) =
          KsiAbs.SCInt (opI (i0, i1))
      | implOpOnNum _ _ opR (KsiAbs.SCReal r0, KsiAbs.SCReal r1) =
          KsiAbs.SCReal (opR (r0, r1))
      | implOpOnNum _ _ _ _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implOpOnNum")

    fun boolToValue true =
          KsiVals.Constructor {name = "true", value = NONE}
      | boolToValue false =
          KsiVals.Constructor {name = "false", value = NONE}

    fun valueToBool (KsiVals.Constructor {name = "true", value = NONE}) =
          true
      | valueToBool (KsiVals.Constructor {name = "false", value = NONE}) =
          false
      | valueToBool _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "valueToBool")
          
    fun implBoolOpOnNumTxt opW _ _ _ _ (KsiAbs.SCWord w0, KsiAbs.SCWord w1) =
          boolToValue (opW (w0, w1))
      | implBoolOpOnNumTxt _ opI _ _ _ (KsiAbs.SCInt i0, KsiAbs.SCInt i1) =
          boolToValue (opI (i0, i1))
      | implBoolOpOnNumTxt _ _ opR _ _ (KsiAbs.SCReal r0, KsiAbs.SCReal r1) =
          boolToValue (opR (r0, r1))
      | implBoolOpOnNumTxt _ _ _ opS _ (KsiAbs.SCString s0, KsiAbs.SCString s1) =
          boolToValue (opS (s0, s1))
      | implBoolOpOnNumTxt _ _ _ _ opC (KsiAbs.SCChar c0, KsiAbs.SCChar c1) =
          boolToValue (opC (c0, c1))
      | implBoolOpOnNumTxt _ _ _ _ _ _ =
          Err.error 1.0 ("KsiBuiltinDynamic: " ^ "implBoolOpOnNumTxt")

    fun typeErrorOp str _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ str)
    
    fun implOpOnWordInt opW opI =
          implOpOnNum opW opI
                      (typeErrorOp "implOpOnWordInt")

    fun implOpOnReal opR =
          implOpOnNum (typeErrorOp "implOpOnReal")
                      (typeErrorOp "implOpOnReal")
                      opR

    fun implBoolOpOnReal opR =
          implBoolOpOnNumTxt (typeErrorOp "implBoolOpOnReal")
                             (typeErrorOp "implBoolOpOnReal")
                             opR
                             (typeErrorOp "implBoolOpOnReal")
                             (typeErrorOp "implBoolOpOnReal")
    
    fun implBoolOpOnEq opW opI opS opC =
          implBoolOpOnNumTxt opW opI
                             (typeErrorOp "implBoolOpOnEq")
                             opS opC
    
    fun implOpOnInt opI =
          implOpOnNum (typeErrorOp "implOpOnInt")
                      opI
                      (typeErrorOp "implOpOnInt")

    fun implOpOnWord opW =
          implOpOnNum opW
                      (typeErrorOp "implOpOnWord")
                      (typeErrorOp "implOpOnWord")

    val implOpDiv = implOpOnWordInt Word.div Int.div
    val implOpMod = implOpOnWordInt Word.mod Int.mod
    val implOpTimes = implOpOnNum op* op* op*
    val implOpDivide = implOpOnReal op/
    
    val implOpPlus = implOpOnNum op+ op+ op+
    val implOpMinus = implOpOnNum op- op- op-

    val implOpEqual = implBoolOpOnEq op= op= op= op=
    val implOpNotEqual = implBoolOpOnEq op<> op<> op<> op<>
    val implOpLessThan = implBoolOpOnNumTxt op< op< op< op< op<
    val implOpGreaterThan = implBoolOpOnNumTxt op> op> op> op> op>
    val implOpLessThanEqual = implBoolOpOnNumTxt op<= op<= op<= op<= op<=
    val implOpGreaterThanEqual = implBoolOpOnNumTxt op>= op>= op>= op>= op>=
    
    val opDiv = wrapConstBinOp implOpDiv
    val opMod = wrapConstBinOp implOpMod
    val opTimes = wrapConstBinOp implOpTimes
    val opDivide = wrapConstBinOp implOpDivide
    
    val opPlus = wrapConstBinOp implOpPlus
    val opMinus = wrapConstBinOp implOpMinus

    val opEqual = wrapBinOp implOpEqual
    val opNotEqual = wrapBinOp implOpNotEqual
    val opLessThan = wrapBinOp implOpLessThan
    val opGreaterThan = wrapBinOp implOpGreaterThan
    val opLessThanEqual = wrapBinOp implOpLessThanEqual
    val opGreaterThanEqual = wrapBinOp implOpGreaterThanEqual
    
    fun implHat (KsiVals.Record fields) =
          (case fields of
             [{lab = "1", value = KsiVals.Const (KsiAbs.SCString s0)},
              {lab = "2", value = KsiVals.Const (KsiAbs.SCString s1)}] =>
               KsiVals.Const (KsiAbs.SCString (s0 ^ s1))
           | [{lab = "2", value = KsiVals.Const (KsiAbs.SCString s1)},
              {lab = "1", value = KsiVals.Const (KsiAbs.SCString s0)}] =>
               KsiVals.Const (KsiAbs.SCString (s0 ^ s1))
           | _ =>
               Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implHat (2)")
          )
      | implHat _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implHat (1)")

    fun implNegate (KsiVals.Const (KsiAbs.SCReal r)) =
          KsiVals.Const (KsiAbs.SCReal (~r))
      | implNegate (KsiVals.Const (KsiAbs.SCInt i)) =
          KsiVals.Const (KsiAbs.SCInt (~i))
      | implNegate _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implNegate")
          
    fun implAbs (KsiVals.Const (KsiAbs.SCReal r)) =
          KsiVals.Const (KsiAbs.SCReal (abs r))
      | implAbs (KsiVals.Const (KsiAbs.SCInt i)) =
          KsiVals.Const (KsiAbs.SCInt (abs i))
      | implAbs _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implAbs")

    fun implAssign (KsiVals.Record fields) =
          (case fields of
             [{lab = "1", value = KsiVals.Ref r},
              {lab = "2", value = v}] =>
               (r := v; KsiVals.Record nil)
           | [{lab = "2", value = v},
              {lab = "1", value = KsiVals.Ref r}] =>
               (r := v; KsiVals.Record nil)
           | _ =>
               Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implAssign (1)")
          )
      | implAssign _ =
          Err.bug 1.0 ("KsiBuiltinDynamic: " ^ "implAssign (2)")

    (* Build a symbol table of the environment of builtin values.
       This is for evaluation of xi-Calculus. *)
    fun bindVal env name value =
          ST.insert (ST.symbol name) value env

    fun bindFun env name funImpl =
          bindVal env name (KsiVals.Fun funImpl)

    (* Declare the basic top-level environment values, which we are not
       allowed to declare as CeXL source code for syntactic reasons. *)
    fun declareDynValEnv () =
          let
            val env = (ST.empty ()) : KsiVals.t ST.t
            val env = bindFun env ":="      implAssign
            val env = bindVal env "absent"  (KsiVals.Field NONE)
            val env = bindFun env "present" (fn x => KsiVals.Field (SOME x))
            val env = bindVal env "Match"
                        (KsiVals.Exception {name = ref (), value = NONE})
            val env = bindVal env "Bind"
                        (KsiVals.Exception {name = ref (), value = NONE})
                        
            val env = bindFun env "div" opDiv
            val env = bindFun env "mod" opMod
            val env = bindFun env "*" opTimes
            val env = bindFun env "/" opDivide
            val env = bindFun env "+" opPlus
            val env = bindFun env "-" opMinus
            val env = bindFun env "=" opEqual
            val env = bindFun env "<>" opNotEqual
            val env = bindFun env "<" opLessThan
            val env = bindFun env ">" opGreaterThan
            val env = bindFun env "<=" opLessThanEqual
            val env = bindFun env ">=" opGreaterThanEqual
            val env = bindFun env "~" implNegate
            val env = bindFun env "^" implHat
            val env = bindFun env "abs" implAbs
          in
            env
          end
  end
