(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiBuiltinStatic = 
  struct
    structure ST = SymbolTable

    (* All predeclared constructor names.
       This is for Naked CeXL to xi-Calculus translation *)
    val predecConsNames = ["true", "false",
                           "nil", "::", "ref",
                           "absent", "present"]

    (* All predeclared type names.
       This is for Naked CeXL to xi-Calculus translation *)
    val predecTypeNames = ["bool", "int", "real", "word",
                           "string", "char", "list", "ref",
                           "absent", "present", "exn", "?"]

    (* An alpha type variable type, which is handy *)
    fun alphaVarTyRestr restr =
          KsiAbs.TyVar
            {name = "'a",
             restr = restr,
             ann = ()
            }

    val alphaVarTy = alphaVarTyRestr KsiAbs.RstrNone

    (* All the CeXL types *)
    val unitTy = KsiAbs.TyEmptyRecord {ann = ()}

    val boolTy = KsiAbs.TyCon {args = nil,
                               name = "bool",
                               tags = ["true", "false"],
                               ann = ()}

    val absentTy = KsiAbs.TyCon {args = nil,
                                 name = "absent",
                                 tags = ["absent"],
                                 ann = ()}

    val presentTy = KsiAbs.TyCon {args = nil,
                                  name = "present",
                                  tags = ["present"],
                                  ann = ()}

    val intTy = KsiAbs.TyCon {args = nil,
                              name = "int",
                              tags = ["0"],
                              ann = ()}

    val wordTy = KsiAbs.TyCon {args = nil,
                               name = "word",
                               tags = ["0w0"],
                               ann = ()}

    val realTy = KsiAbs.TyCon {args = nil,
                               name = "real",
                               tags = ["0.0"],
                               ann = ()}

    val charTy = KsiAbs.TyCon {args = nil,
                               name = "char",
                               tags = ["#\"a\""],
                               ann = ()}

    val stringTy = KsiAbs.TyCon {args = nil,
                                 name = "string",
                                 tags = ["#\"\""],
                                 ann = ()}

    val listTy = KsiAbs.TyCon {args = [{name = "'a",
                                        ty = alphaVarTy}],
                               name = "list",
                               tags = ["nil", "::"],
                               ann = ()}

    val refTy = KsiAbs.TyCon {args = [{name = "'a",
                                       ty = alphaVarTy}],
                              name = "ref",
                              tags = ["ref"],
                              ann = ()}

    val exnTy = KsiAbs.TyCon {args = [],
                              name = "exn",
                              tags = [],
                              ann = ()}

    fun makePresentField ty =
          KsiAbs.TyField {present = presentTy, ty = ty, ann = ()}

    (* All CeXL constructor and exception type schemes
       and the unit type scheme *)
    val unitTyScheme = {params = nil,
                        ty = unitTy,
                        ann = ()}

    val refTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                       ty = KsiAbs.TyFun
                               {from = alphaVarTy,
                                to = refTy,
                                ann = ()},
                       ann = ()}

    val nilTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                       ty = listTy,
                       ann = ()}

    val consTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                        ty = KsiAbs.TyFun
                               {from = KsiAbs.TyExtRecord
                                         {fields =
                                            [{lab = "1", ty = makePresentField alphaVarTy, ann = ()},
                                             {lab = "2", ty = makePresentField listTy, ann = ()}],
                                          ext = KsiAbs.TyEmptyRecord
                                                  {ann = ()},
                                          ann = ()},
                                to = listTy,
                                ann = ()},
                        ann = ()}

    val trueTyScheme = {params = nil,
                        ty = boolTy,
                        ann = ()}

    val falseTyScheme = trueTyScheme

    val exnTyScheme = {params = nil,
                       ty = exnTy,
                       ann = ()}

    (* Notice that this is the type scheme for the
       absent constructor, not the absent field variable! *)
    val absentTyScheme = {params = nil,
                          ty = absentTy,
                          ann = ()}

    (* Notice that this is the type scheme for the
       present constructor, not the present field variable! *)
    val presentTyScheme = {params = nil,
                           ty = presentTy,
                           ann = ()}

    val matchTyScheme = {params = nil,
                         ty = exnTy,
                         ann = ()}

    val bindTyScheme = matchTyScheme

    (* Basis Library *)
    val arrayTy = KsiAbs.TyCon {args = [{name = "'a",
                                         ty = alphaVarTy}],
                                name = "Array.t",
                                tags = [],
                                ann = ()}

    val textIOOutStreamTy = KsiAbs.TyCon {args = [],
                                name = "TextIO.outstream",
                                tags = [],
                                ann = ()}

    val textIOInStreamTy = KsiAbs.TyCon {args = [],
                                name = "TextIO.instream",
                                tags = [],
                                ann = ()}

    val binIOOutStreamTy = KsiAbs.TyCon {args = [],
                                name = "BinIO.outstream",
                                tags = [],
                                ann = ()}

    val binIOInStreamTy = KsiAbs.TyCon {args = [],
                                name = "BinIO.instream",
                                tags = [],
                                ann = ()}

    (* Some type patterns that we need *)
    val intTyPat = KsiAbs.RstrTyPat {args = nil,
                                     name = "int",
                                     tags = ["0"]}

    val wordTyPat = KsiAbs.RstrTyPat {args = nil,
                                      name = "word",
                                      tags = ["0w0"]}

    val realTyPat = KsiAbs.RstrTyPat {args = nil,
                                      name = "real",
                                      tags = ["0.0"]}

    val charTyPat = KsiAbs.RstrTyPat {args = nil,
                                      name = "char",
                                      tags = ["#\"a\""]}

    val stringTyPat = KsiAbs.RstrTyPat {args = nil,
                                        name = "string",
                                        tags = ["#\"\""]}
    
    (* Some restrictions that we need *)
    val restrRealInt = KsiAbs.RstrTyPats
                         {pats = [realTyPat, intTyPat]}

    val restrWordInt = KsiAbs.RstrTyPats
                         {pats = [wordTyPat, intTyPat]}

    val restrNum = KsiAbs.RstrTyPats
                         {pats = [wordTyPat, realTyPat, intTyPat]}

    val restrNumTxt = KsiAbs.RstrTyPats
                         {pats = [wordTyPat, realTyPat, intTyPat,
                                  stringTyPat, charTyPat]}

    val restrEq = KsiAbs.RstrTyPats
                         {pats = [wordTyPat, intTyPat,
                                  stringTyPat, charTyPat]}

    (* Types (that is, type schemes) of declared variables *)
    val assignTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                          ty = KsiAbs.TyFun
                                 {from = KsiAbs.TyExtRecord
                                           {fields =
                                              [{lab = "1", ty = makePresentField refTy, ann = ()},
                                               {lab = "2", ty = makePresentField alphaVarTy, ann = ()}],
                                            ext = KsiAbs.TyEmptyRecord
                                                    {ann = ()},
                                            ann = ()},
                                  to = unitTy,
                                  ann = ()},
                          ann = ()}

    (* This is the type scheme for the absent field variable *)
    val absentVarTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                              ty = KsiAbs.TyField
                                     {present = absentTy,
                                      ty = alphaVarTy,
                                      ann = ()},
                              ann = ()}

    (* This is the type scheme for the present field variable *)
    val presentVarTyScheme = {params = [{name = "'a", restr = KsiAbs.RstrNone}],
                              ty = KsiAbs.TyFun
                                     {from = alphaVarTy,
                                      to = KsiAbs.TyField
                                             {present = presentTy,
                                              ty = alphaVarTy,
                                              ann = ()},
                                      ann = ()},
                              ann = ()}

    fun operatorTyArg ty =
          KsiAbs.TyExtRecord
            {fields =
               [{lab = "1", ty = makePresentField ty, ann = ()},
                {lab = "2", ty = makePresentField ty, ann = ()}],
             ext = KsiAbs.TyEmptyRecord {ann = ()},
             ann = ()}

    fun operatorArg restr =
          operatorTyArg (alphaVarTyRestr restr)
              
    fun valueTyOp ty =
          {params = [],
           ty = KsiAbs.TyFun
                  {from = operatorTyArg ty,
                   to = ty,
                   ann = ()},
           ann = ()}                      

    fun valueOp restr =
         {params = [{name = "'a", restr = restr}],
           ty = KsiAbs.TyFun
                  {from = operatorArg restr,
                   to = alphaVarTyRestr restr,
                   ann = ()},
           ann = ()}
    
    fun boolOp restr =
          {params = [{name = "'a", restr = restr}],
           ty = KsiAbs.TyFun
                  {from = operatorArg restr,
                   to = boolTy,
                   ann = ()},
           ann = ()}                      

    fun valueFunc restr =
          {params = [{name = "'a", restr = restr}],
           ty = KsiAbs.TyFun
                  {from = alphaVarTyRestr restr,
                   to = alphaVarTyRestr restr,
                   ann = ()},
           ann = ()}                      

    (* Build a symbol table with the types of the predeclared variables *)
    fun bindTyScheme env name tyScheme =
          ST.insert (ST.symbol name)
                    (tyScheme : unit KsiAbs.tyscheme)
                    env

    fun bindTySchemeParamTy env name params ty =
          bindTyScheme env name {params = params, ty = ty, ann = ()}
          
    fun bindTySchemeTy env name ty =
          bindTySchemeParamTy  env name nil ty
    
    fun declareStaVarEnv () =
          let
            val env = ST.empty ()
            val env = bindTyScheme env ":="      assignTyScheme
            val env = bindTyScheme env "absent"  absentVarTyScheme
            val env = bindTyScheme env "present" presentVarTyScheme
            val env = bindTyScheme env "="   (boolOp    restrEq)
            val env = bindTyScheme env "<>"  (boolOp    restrEq)
            val env = bindTyScheme env "+"   (valueOp   restrNum)
            val env = bindTyScheme env "-"   (valueOp   restrNum)
            val env = bindTyScheme env "*"   (valueOp   restrNum)
            val env = bindTyScheme env "/"   (valueTyOp realTy)
            val env = bindTyScheme env "<"   (boolOp    restrNumTxt)
            val env = bindTyScheme env ">"   (boolOp    restrNumTxt)
            val env = bindTyScheme env "<="  (boolOp    restrNumTxt)
            val env = bindTyScheme env ">="  (boolOp    restrNumTxt)
            val env = bindTyScheme env "~"   (valueFunc restrRealInt)
            val env = bindTyScheme env "^"   (valueTyOp stringTy)
            val env = bindTyScheme env "mod" (valueOp   restrWordInt)
            val env = bindTyScheme env "div" (valueOp   restrWordInt)
            val env = bindTyScheme env "abs" (valueFunc restrRealInt)

            (* Note: The operators @, ! and o are declared as xi-Calculus code *)
          in
            env
          end

    (* Build a symbol table with the pre-bound type variables *)
    fun declareStaTyVars () =
          let
            (* No pre-bound type variables *)
            val env = ST.empty ()
          in
            env
          end
  end
