(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiPrettyPrint =
  struct
    (* This module prints xi-Calculus types and restrictions using CeXL syntax. *)
    (* This should really be rewritten to use proper
       combinator pretty printing similar to what is found in
       the book "ML for the Working Programmer" by L.C. Paulson.
       But it is hand-coded here to keep the code dependencies minimal. *)
    fun labsIterToStrs (lab::labs) acc =
          ", "::lab::(labsIterToStrs labs acc)
      | labsIterToStrs nil acc =
          acc

    fun labsToStrsAcc (lab::labs) acc =
          lab::(labsIterToStrs labs acc)
      | labsToStrsAcc nil acc =
          acc

    fun labsToStrs labs =
          labsToStrsAcc labs nil

    fun labsToStr labs =
          String.concat (labsToStrs labs)

    fun restrToStrs KsiAbs.RstrNone acc =
          acc
      | restrToStrs (KsiAbs.RstrExclude {labs = nil}) acc =
          "~{}"::acc
      | restrToStrs (KsiAbs.RstrExclude {labs = lab::labs}) acc =
          "~{"::lab::(labsToStrsAcc labs ("}"::acc))
      | restrToStrs (KsiAbs.RstrTyPats {pats = pat::pats}) acc =
          restrTyPatToStrs pat (restrTyPatsToStrs pats acc)
      | restrToStrs _ acc =
          (* A RstrTyPats without patterns would be an error in xi-Calculus.
             Our job here is only to display though - even errors. *)
          acc

    and restrTyPatsToStrs (pat::pats) acc =
          " | "::(restrTyPatToStrs pat (restrTyPatsToStrs pats acc))
      | restrTyPatsToStrs nil acc =
          acc

    and restrTyPatToStrs (KsiAbs.RstrTyPat {args = arg::args, name, tags}) acc =
          "["::(restrTyPatArgToStrs arg
                  (restrTyPatArgsToStrs args ("] "::name::acc)))
      | restrTyPatToStrs (KsiAbs.RstrTyPat {args = nil, name, tags}) acc =
          "[] "::name::acc

    and restrTyPatArgToStrs {name, pat} acc =
          name::" = "::(restrTyPatToStrs pat acc)

    and restrTyPatArgsToStrs (arg::args) acc =
          ", "::(restrTyPatArgToStrs arg (restrTyPatArgsToStrs args acc))
      | restrTyPatArgsToStrs nil acc =
          acc

    fun restrToStr restr =
          String.concat (restrToStrs restr nil)

    fun tyToStrs (KsiAbs.TyCon {args = [{name = "'a", ty}], name, tags, ann}) acc =
          (* We handle the special case of only one type parameter 'a *)
          tyToStrs ty (" "::name::acc)
      | tyToStrs (KsiAbs.TyCon {args = arg::args, name, tags, ann}) acc =
          "["::(tyTyConArgToStrs arg
                  (tyTyConArgsToStrs args ("] "::name::acc)))
      | tyToStrs (KsiAbs.TyCon {args = nil, name, tags, ann}) acc =
          (* No tycon parameters - so we just display the name *)
          name::acc
      | tyToStrs (KsiAbs.TyFun {from, to, ann}) acc =
          tyToStrs from (" -> "::(tyToStrs to acc))
      | tyToStrs (KsiAbs.TyField {present, ty, ann}) acc =
          "("::(tyToStrs present (", "::(tyToStrs ty (") ?"::acc))))
      | tyToStrs (KsiAbs.TyExtRecord {fields = field::fields, ext, ann}) acc =
          "{"::(tyRecordFieldToStrs field
                  (tyRecordFieldsToStrs fields
                     (tyRecordExtToStrs ext ("}"::acc))))
      | tyToStrs (KsiAbs.TyExtRecord {fields = nil, ext, ann}) acc =
          (* We cannot display this type correctly if there are no
             fields at all. This situation is checked for elsewhere. *)
          "{<internal error: ext record with no fields>"::
          (tyRecordExtToStrs ext ("}"::acc))
      | tyToStrs (KsiAbs.TyEmptyRecord {ann}) acc =
          "{}"::acc
      | tyToStrs (KsiAbs.TyVar {name, restr, ann}) acc =
          name::acc
      | tyToStrs (KsiAbs.TyMeta {name, restr, bound, ann}) acc =
          (case !bound of
             NONE =>
               name::acc
           | SOME t =>
               tyToStrs t acc
          )

    and tyTyConArgsToStrs (arg::args) acc =
          ", "::(tyTyConArgToStrs arg (tyTyConArgsToStrs args acc))
      | tyTyConArgsToStrs nil acc =
          acc

    and tyTyConArgToStrs {name, ty} acc =
          name::" = "::(tyToStrs ty acc)

    and tyRecordFieldsToStrs (field::fields) acc =
          ", " :: (tyRecordFieldToStrs field
                     (tyRecordFieldsToStrs fields acc))
      | tyRecordFieldsToStrs nil acc =
          acc

    and tyRecordFieldToStrs {lab, ty, ann} acc =
          tyRecordFieldTyToStrs lab ty acc

    and tyRecordFieldTyToStrs lab (KsiAbs.TyField {present, ty, ann}) acc =
          tyRecordFieldPresentTyToStrs present
            (lab::" : "::(tyToStrs ty acc))
      | tyRecordFieldTyToStrs lab ty acc =
          "<non-field> "::lab::" : "::(tyToStrs ty acc)
          
    and tyRecordFieldPresentTyToStrs (KsiAbs.TyCon {name, args, tags, ann = _}) acc =
          if name = "present" then
            acc
          else if name = "absent" then
            "- " :: acc
          else
            "<invalid cons> " :: acc
      | tyRecordFieldPresentTyToStrs (KsiAbs.TyVar {name, restr, ann}) acc =
          name::" "::acc
      | tyRecordFieldPresentTyToStrs (KsiAbs.TyMeta {name, restr, bound, ann}) acc =
          (case !bound of
             NONE =>
               name::" "::acc
           | SOME t =>
               tyRecordFieldPresentTyToStrs t acc
          )
      | tyRecordFieldPresentTyToStrs ty acc =
          "<no cons> " :: (tyToStrs ty acc)

    and tyRecordExtToStrs (KsiAbs.TyEmptyRecord _) acc =
          (* Non-extensible record - no extension part to display *)
          acc
      | tyRecordExtToStrs (KsiAbs.TyExtRecord {fields, ext, ann}) acc =
          (* The record is longer still - continue to display *)
          tyRecordFieldsToStrs fields (tyRecordExtToStrs ext acc)
      | tyRecordExtToStrs (KsiAbs.TyMeta {name, restr, bound, ann}) acc =
          (case !bound of
             NONE =>
               (* Extension is an unbound meta variable *)
               ", ... : "::name::acc
           | SOME t =>
               (* The record is longer still - continue to display *)
               tyRecordExtToStrs t acc
          )
      | tyRecordExtToStrs (KsiAbs.TyVar {name, restr, ann}) acc =
          (* Extension is a type variable *)
          ", ... : "::name::acc
      | tyRecordExtToStrs t acc =
          (* Any other type (i.e. TyCon, TyFun or TyField)
             would be an error here - but we just display it. *)
          ", ... : "::(tyToStrs t acc)
    
    fun tyToStr ty =
          String.concat (tyToStrs ty nil)

    fun tyParamToStrs {name, restr} acc =
          name :: " : " :: (restrToStrs restr acc)

    fun tyParamsToStrs [param] acc =
          tyParamToStrs param acc
      | tyParamsToStrs (param::params) acc =
          tyParamToStrs param ("; " :: (tyParamsToStrs params acc))
      | tyParamsToStrs nil acc =
          acc

    fun tyParamsToStr params =
          String.concat (tyParamsToStrs params nil)

    fun tySchemeToStr {params, ty, ann} =
          String.concat ["[", tyParamsToStr params, "]", tyToStr ty]

  end