(*
   This file is part of Anoq's simple implementation of CeXL.
   CeXL is a programming language developed by Anoq of the Sun
   at Hardcore Processing.

   Copyright (C) 2003-2010 Anoq of the Sun (alias Johnny Bock Andersen)

   Contact: http://www.hardcoreprocessing.com/contact

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing for more information at:
   http://www.hardcoreprocessing.com/contact

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

structure KsiTypeInfer =
  struct
    (* exception InternalError *)

    structure ST = SymbolTable

    datatype printdescrs =
      PDTopLevel
    | PDNested
    | PDNone

    val showDescrs = ref PDNone

    (* Unique names *)
    val uniqueMetaNames = CountingInteger.create ()

    (* Functions for reporting errors *)
    fun reset () =
          CountingInteger.reset uniqueMetaNames

    val subSystemCode = (0wx10 : Word32.word)
    
    val errorReported =
          KsiReportTypeErrors.errorReported

    val reportError =
          KsiReportTypeErrors.reportError subSystemCode
    val reportErrorStrs =
          KsiReportTypeErrors.reportErrorStrs subSystemCode

    val reportErrorFurther =
          KsiReportTypeErrors.reportErrorFurther
    val reportErrorFurtherStrs =
          KsiReportTypeErrors.reportErrorFurtherStrs

    val reportErrorAtPos =
          KsiReportTypeErrors.reportErrorAtPos subSystemCode
    val reportErrorAtPosStrs =
          KsiReportTypeErrors.reportErrorAtPosStrs subSystemCode

    fun reportInternalError errCode errMsg =
          KsiReportTypeErrors.reportInternalError
            subSystemCode errCode errMsg
    fun reportInternalErrorStrs errCode strs =
          KsiReportTypeErrors.reportInternalErrorStrs
            subSystemCode errCode strs

    fun reportInternalErrorAtPos pos errCode errMsg =
          KsiReportTypeErrors.reportInternalErrorAtPos
            subSystemCode pos errCode errMsg
    fun reportInternalErrorAtPosStrs pos errCode strs =
          KsiReportTypeErrors.reportInternalErrorAtPosStrs
            subSystemCode pos errCode strs
    
    (* This is for reporting errors in a way less confusing for the user *)
    datatype error =
      NewErr
    | ReportedErr
    | NoErr

    fun combineErrorCodes (NoErr, err) =
          err
      | combineErrorCodes (ReportedErr, err) =
          ReportedErr
      | combineErrorCodes (NewErr, ReportedErr) =
          ReportedErr
      | combineErrorCodes (NewErr, err) =
          NewErr

    (* Utilities for making new meta types and abstract syntax snippets *)
    fun newMetaName () =
          "'meta#" ^ (Int.toString (CountingInteger.next uniqueMetaNames))

    fun newMetaTyRestr restr =
          KsiAbs.TyMeta
            {name = newMetaName (),
             restr = restr,
             bound = ref NONE,
             ann = ()}

    fun newMetaTy () =
          newMetaTyRestr KsiAbs.RstrNone

    fun makeAbsentTyCon () =
          KsiAbs.TyCon
            {args = nil,
             name = "absent",
             tags = ["absent"],
             ann = ()}

    fun makePresentTyCon () =
          KsiAbs.TyCon
            {args = nil,
             name = "present",
             tags = ["present"],
             ann = ()}

    fun newAbsentField () =
          KsiAbs.TyField
            {present = makeAbsentTyCon (),
             ty = newMetaTy (),
             ann = ()}

    fun makeAbsentRestrPat () =
          KsiAbs.RstrTyPat
            {args = nil, name = "absent", tags = ["absent"]}

    fun makePresentRestrPat () =
          KsiAbs.RstrTyPat
            {args = nil, name = "present", tags = ["present"]}

    fun makeAbsentPresentRestr () =
          KsiAbs.RstrTyPats
            {pats = [makePresentRestrPat (),
                     makeAbsentRestrPat ()]
            }

    (* More predefined types *)
    fun makeExnTyCon () =
          KsiAbs.TyCon
            {args = nil, name = "exn",
             tags = nil, ann = ()}

    fun newField () =
          KsiAbs.TyField
            {present = newMetaTyRestr
                         (makeAbsentPresentRestr ()),
             ty = newMetaTy (),
             ann = ()}

    (* Convert an integer >= 1 to a type variable name in
       lexicographical order. *)
    fun nextTyVar (c::cs) =
          if ord c < ord #"z" then
            chr ((ord c) + 1)::cs
          else
            #"a" :: (nextTyVar cs)
      | nextTyVar nil =
          [#"a"]

    fun makeTyVar cs =
          implode (#"'" :: (rev cs))

    (* Produce type variable names in lexicographical order *)
    fun lexicogTyvarProducer () =
          let
            val iter = ref nil

            fun produce () =
                  let
                    val _ = iter := (nextTyVar (!iter))
                  in
                    makeTyVar (!iter)
                  end
          in
            produce
          end

    (* Convert an integer >= 1 to a type variable name in
       lexicographical order. Only used for testing actually. *)
    fun lexicogTyvar n =
          let
            val producer = lexicogTyvarProducer ()
          in
            case rev (List.tabulate (n, fn i => producer ())) of
              lastVar::_ =>
                lastVar
            | nil =>
                (reportInternalErrorStrs 0wx0101
                   ["Error producing ", Int.toString n, " type variables"];
                 "'invalid")
          end

    (* General utilities *)
    (* A quick-and-dirty but inefficient implementation of sets *)
    (* Test set membership *)
    fun hasElem eqF elem s =
          List.exists (fn e => eqF(elem, e)) s

    (* Test subset *)
    fun isSubset eqF subS superS =
          foldl (fn (e, acc) => acc andalso hasElem eqF e superS) true subS

    (* Test sets disjoint *)
    fun isDisjointSets eqF s1 s2 =
          foldl (fn (e, acc) => acc andalso not (hasElem eqF e s2)) true s1

    (* Test set equality *)
    fun isEqualSets eqF s1 s2 =
          isSubset eqF s1 s2 andalso
          isSubset eqF s2 s2

    (* Make set intersection *)
    (* This unification implementation relies on the fact that
       the elements of the resulting set is taken from s1 (and not s2!) *)
    fun makeIntersectSet eqF s1 s2 =
          List.filter (fn e => hasElem eqF e s2) s1

    (* Make set difference *)
    fun makeDifferenceSet eqF s removeS =
          List.filter (fn e => not (hasElem eqF e removeS)) s

    (* Make set union *)
    fun makeUnionSet eqF s1 s2 =
          (makeDifferenceSet eqF s1 s2) @ s2
    
    (* More efficient implementation of sets and named values,
       based on sorting the elements. This requires an ordering of elements. *)
    fun sortSet lt elems =
          Util.quickSort lt elems

    fun isSortedDistinct lt elems =
          Util.isSortedDistinct lt elems

    fun getSortedDuplicates eqF elems =
          Util.getSortedDuplicates eqF elems

    fun isEqualSortedSets eqF (e::es) (e2::es2) =
          (eqF (e, e2)) andalso (isEqualSortedSets eqF es es2)
      | isEqualSortedSets eqF nil nil =
          true
      | isEqualSortedSets _ _ _ =
          false

    fun isSubsetSortedSets eqF (sub::subs) (super::supers) =
          if eqF (sub, super) then
            isSubsetSortedSets eqF subs supers
          else
            (* For this to be a subset, super must be an extra element *)
            isSubsetSortedSets eqF (sub::subs) supers
      | isSubsetSortedSets eqF nil _ =
          true
      | isSubsetSortedSets eqF _ _ =
          false

    fun isDisjointSortedSets ltF eqF (e::es) (e2::es2) =
          if ltF (e, e2) then
            isDisjointSortedSets ltF eqF es (e2::es2)
          else if eqF (e, e2) then
            false
          else
            isDisjointSortedSets ltF eqF (e::es) es2
      | isDisjointSortedSets _ _ _ _ =
          true

    fun makeIntersectSortedSets ltF eqF (e::es) (e2::es2) =
          if ltF (e, e2) then
            makeIntersectSortedSets ltF eqF es (e2::es2)
          else if eqF (e, e2) then
            e::(makeIntersectSortedSets ltF eqF es es2)
          else
            makeIntersectSortedSets ltF eqF (e::es) es2
      | makeIntersectSortedSets _ _ _ _ =
          nil
      
    fun makeUnionSortedSets ltF eqF (e::es) (e2::es2) =
          if eqF (e, e2) then
            e ::(makeUnionSortedSets ltF eqF es es2)
          else if ltF (e, e2) then
            e ::(makeUnionSortedSets ltF eqF es (e2::es2))
          else
            e2::(makeUnionSortedSets ltF eqF (e::es) es2)
      | makeUnionSortedSets _ _ nil es2 =
          es2
      | makeUnionSortedSets _ _ es nil =
          es

    fun makeDifferenceSortedSets ltF eqF (e::es) (remove::removes) =
          if eqF (e, remove) then
            makeDifferenceSortedSets ltF eqF es removes
          else if ltF (e, remove) then
            e::(makeDifferenceSortedSets ltF eqF es (remove::removes))
          else
            makeDifferenceSortedSets ltF eqF (e::es) removes
      | makeDifferenceSortedSets _ _ nil removes =
          nil
      | makeDifferenceSortedSets _ _ es nil =
          es

    (* Equality tests *)
    fun eqTyPat (KsiAbs.RstrTyPat {args, name, tags},
                 KsiAbs.RstrTyPat {args = args2, name = name2, tags = tags2}) =
          name = name2 andalso
          isEqualSets eqTyPatArg args args2

    and eqTyPatArg ({name, pat}, {name = name2, pat = pat2}) =
          name = name2 andalso
          eqTyPat (pat, pat2)

    fun eqRestrict (KsiAbs.RstrNone, KsiAbs.RstrNone) =
          true
      | eqRestrict (KsiAbs.RstrExclude {labs},
                    KsiAbs.RstrExclude {labs = labs2}) =
          isEqualSortedSets op= labs labs2
      | eqRestrict (KsiAbs.RstrTyPats {pats},
                    KsiAbs.RstrTyPats {pats = pats2}) =
          isEqualSets eqTyPat pats pats2
      | eqRestrict _ =
          false

    fun eqRecordField ({lab, ty, ann}, {lab = lab2, ty = _, ann = _}) =
          lab = lab2

    (* Ordering tests *)
    fun ltRecordField ({lab : string, ty, ann}, {lab = lab2, ty = _, ann = _}) =
          lab < lab2

    fun ltTyConArg ({name : string, ty}, {name = name2, ty = ty2}) =
          name < name2

    (* Equality predicates *)
    fun isNamedMetaVar theName (KsiAbs.TyMeta {name, ...}) =
          theName = name
      | isNamedMetaVar _ _ =
          false

    fun isAbsentField (KsiAbs.TyField {present, ty, ann}) =
          (case present of
             KsiAbs.TyCon {args, name, tags, ann} =>
               name = "absent"
           | KsiAbs.TyMeta {name, bound, restr, ann = _} =>
               (case !bound of
                  NONE =>
                    (* If unbound meta variable is restricted to having
                       to be absent, it will be replaced with absent during
                       closure, so it counts as an absent field. *)
                    eqRestrict
                      (restr,
                       KsiAbs.RstrTyPats {pats = [makeAbsentRestrPat ()]})
                | SOME ty' =>
                    (* Follow bound meta variables *)
                    isAbsentField
                      (KsiAbs.TyField {present = ty', ty = ty, ann = ann})
               )
           | _ =>
               false
          )
      | isAbsentField _ =
          false

    (* Makes an "etcetera" type, for showing missing type information *)
    fun makeEtc (ty as (KsiAbs.TyCon {args, name = "..", tags, ann})) =
          ty
      | makeEtc ty =
          KsiAbs.TyCon
            {args = [{name = "'a", ty = ty}],
             name = "..",
             tags = nil,
             ann = ()}

    fun makeEtcNoParam () =
          KsiAbs.TyCon
            {args = nil, name = "..", tags = nil, ann = ()}

    (* Generate the difference between 2 types.
       The result of diffTy(ty1, ty2) is NONE if the types are equal and
       (SOME ty) if not, where ty is the type information contained in
       ty1 which is not contained in ty2. *)
    fun diffTy (ty as (KsiAbs.TyCon {args, name, tags, ann}),
                KsiAbs.TyCon {args = args2, name = name2, tags = tags2, ann = ann2}) =
          if name = name2 then
            let
              val args  = sortSet ltTyConArg args
              val args2 = sortSet ltTyConArg args2

              fun makeDiffArgs ({name, ty}::args) ({name = name2, ty = ty2}::args2) =
                    if name = name2 then
                      (case diffTy (ty, ty2) of
                         NONE =>
                           makeDiffArgs args args2
                       | SOME ty' =>
                           {name = name, ty = ty'}
                             ::(makeDiffArgs args args2)
                      )
                    else if name < name2 then
                      {name = name, ty = ty}
                        ::(makeDiffArgs args ({name = name2, ty = ty2}::args2))
                    else
                      makeDiffArgs ({name = name, ty = ty}::args) args2
                | makeDiffArgs nil _ =
                    nil
                | makeDiffArgs args nil =
                    args
            in
              case makeDiffArgs args args2 of
                nil =>
                  NONE
              | diffArgs =>
                  SOME (KsiAbs.TyCon
                          {args = diffArgs,
                           name = name,
                           tags = tags,
                           ann = ann})
            end
          else
            SOME ty
      | diffTy (KsiAbs.TyFun {from, to, ann},
                KsiAbs.TyFun {from = from2, to = to2, ann = ann2}) =
          (case (diffTy (from, from2), diffTy (to, to2)) of
             (NONE, NONE) =>
                NONE
           | (SOME ty, NONE) =>
                SOME (makeEtc ty)
           | (NONE, SOME ty) =>
                SOME (makeEtc ty)
           | (SOME from', SOME to') =>
                SOME (KsiAbs.TyFun
                        {from = from', to = to', ann = ann})
          )
      | diffTy (KsiAbs.TyField {present, ty, ann},
                KsiAbs.TyField {present = present2, ty = ty2, ann = ann2}) =
          (* We do not replace TyField with "etcetera" type,
             to display records correctly. *)
          (case (diffTy (present, present2), diffTy (ty, ty2)) of
             (NONE, NONE) =>
                NONE
           | (SOME _, NONE) =>
                SOME (KsiAbs.TyField
                        {present = present,
                         ty = makeEtcNoParam (),
                         ann = ann})
           | (NONE, SOME ty') =>
                SOME (KsiAbs.TyField
                        {present = present,
                         ty = ty',
                         ann = ann})
           | (SOME present', SOME ty') =>
                SOME (KsiAbs.TyField
                        {present = present',
                         ty = ty',
                         ann = ann})
          )
      | diffTy (KsiAbs.TyExtRecord {fields, ext, ann},
                KsiAbs.TyExtRecord {fields = fields2, ext = ext2, ann = ann2}) =
          let
            fun makeDiffFields ({lab, ty, ann}::fields) ({lab = lab2, ty = ty2, ann = ann2}::fields2) =
                    if lab = lab2 then
                      (case diffTy (ty, ty2) of
                         NONE =>
                           makeDiffFields fields fields2
                       | SOME ty' =>
                           {lab = lab, ty = ty', ann = ann}
                             ::(makeDiffFields fields fields2)
                      )
                    else if lab < lab2 then
                      {lab = lab, ty = ty, ann = ann}
                        ::(makeDiffFields fields ({lab = lab2, ty = ty2, ann = ann2}::fields2))
                    else
                      makeDiffFields ({lab = lab, ty = ty, ann = ann}::fields) fields2
                | makeDiffFields nil _ =
                    nil
                | makeDiffFields fields nil =
                    fields
          in
              case makeDiffFields fields fields2 of
                nil =>
                  (case diffTy (ext, ext2) of
                     NONE =>
                       NONE
                   | SOME ty' =>
                       (* Notice: An empty list of record fields means that
                                  the difference is really only the extension *)
                       SOME ty'
                  )
              | diffFields =>
                  SOME (KsiAbs.TyExtRecord
                          {fields = diffFields,
                           ext = case diffTy (ext, ext2) of
                                   NONE =>
                                     (* Notice: The annotation on this empty
                                        record is not type-preserving. *)
                                     KsiAbs.TyEmptyRecord {ann = ann}
                                 | SOME ty' =>
                                     ty',
                           ann = ann})
          end
      | diffTy (KsiAbs.TyEmptyRecord {ann},
                KsiAbs.TyEmptyRecord {ann = ann2}) =
          NONE
      | diffTy (ty as (KsiAbs.TyExtRecord {fields, ext, ann}),
                KsiAbs.TyEmptyRecord {ann = ann2}) =
          (* We don't do the symmetric equivalent of this, since
             this corresponds to diffTy with its parameters switched. *)
          SOME ty 
      | diffTy (ty as (KsiAbs.TyVar {name, restr, ann}),
                KsiAbs.TyVar {name = name2, restr = restr2, ann = ann2}) =
          if name = name2 then
            NONE
          else
            SOME ty
      | diffTy (ty  as (KsiAbs.TyMeta {name, restr, bound, ann}),
                ty2 as (KsiAbs.TyMeta {name = name2, restr = restr2, bound = bound2, ann = ann2})) =
          if name = name2 then
            NONE
          else
            (* Meta variables are treated transitively when comparing types *)
            (case !bound of
               NONE =>
                 (case !bound2 of
                    NONE =>
                      SOME ty
                  | SOME ty2' =>
                      diffTy (ty, ty2')
                 )
             | SOME ty' =>
                 diffTy (ty', ty2)
            )
      | diffTy (ty as (KsiAbs.TyMeta {name, restr, bound, ann}), ty2) =
          (* Meta variables are treated transitively when comparing types *)
          (case !bound of
             NONE =>
               SOME ty
           | SOME ty' =>
               diffTy (ty', ty2)
          )
      | diffTy (ty, ty2 as (KsiAbs.TyMeta {name = name2, restr = restr2, bound = bound2, ann = ann2})) =
          (* Meta variables are treated transitively when comparing types *)
          (case !bound2 of
             NONE =>
               SOME ty
           | SOME ty2' =>
               diffTy (ty, ty2')
          )
      | diffTy (ty, ty2) =
          SOME ty

    (* Traverse *)
    fun foldEmbedTy f acc (t as (KsiAbs.TyCon {args, name, tags, ann})) =
          f (t, foldEmbedTyTyConArgs f acc args)
      | foldEmbedTy f acc (t as (KsiAbs.TyFun {from, to, ann})) =
          f (t, foldEmbedTy f (foldEmbedTy f acc from) to)
      | foldEmbedTy f acc (t as (KsiAbs.TyField {present, ty, ann})) =
          f (t, foldEmbedTy f (foldEmbedTy f acc present) ty)
      | foldEmbedTy f acc (t as (KsiAbs.TyExtRecord {fields, ext, ann})) =
          f (t, foldEmbedTy f (foldEmbedTyExtRecordFields f acc fields) ext)
      | foldEmbedTy f acc (t as (KsiAbs.TyEmptyRecord _)) =
          f (t, acc)
      | foldEmbedTy f acc (t as (KsiAbs.TyVar _)) =
          f (t, acc)
      | foldEmbedTy f acc (t as (KsiAbs.TyMeta {name, restr, bound, ann})) =
          f (t,
             case !bound of
               NONE    => acc
             | SOME ty => foldEmbedTy f acc ty
            )

    (* If we had extensible records here, it would be enough to write
       2 functions here rather than 4! We only use the field ty :-) *)
    and foldEmbedTyTyConArgs f acc args =
          foldl (foldEmbedTyTyConArg f) acc args

    and foldEmbedTyTyConArg f ({name, ty}, acc) =
          foldEmbedTy f acc ty

    and foldEmbedTyExtRecordFields f acc args =
          foldl (foldEmbedTyExtRecordField f) acc args

    and foldEmbedTyExtRecordField f ({lab, ty, ann}, acc) =
          foldEmbedTy f acc ty

    (* Replace embedded types bottom-up by a given map function *)
    fun mapEmbedTy f (KsiAbs.TyCon {args, name, tags, ann}) =
          f (KsiAbs.TyCon
               {args = mapEmbedTyTyConArgs f args,
                name = name,
                tags = tags,
                ann = ann})
      | mapEmbedTy f (KsiAbs.TyFun {from, to, ann}) =
          f (KsiAbs.TyFun
               {from = mapEmbedTy f from,
                to = mapEmbedTy f to,
                ann = ann})
      | mapEmbedTy f (KsiAbs.TyField {present, ty, ann}) =
          f (KsiAbs.TyField
               {present = mapEmbedTy f present,
                ty = mapEmbedTy f ty,
                ann = ann})
      | mapEmbedTy f (KsiAbs.TyExtRecord {fields, ext, ann}) =
          f (KsiAbs.TyExtRecord
               {fields = mapEmbedTyExtRecordFields f fields,
                ext = mapEmbedTy f ext,
                ann = ann})
      | mapEmbedTy f (t as (KsiAbs.TyEmptyRecord _)) =
          f t
      | mapEmbedTy f (t as (KsiAbs.TyVar _)) =
          f t
      | mapEmbedTy f (t as (KsiAbs.TyMeta {name, restr, bound, ann})) =
          (case !bound of
             NONE =>
               (* Map unbound meta variables properly *)
               f t
           | SOME ty =>
               (* Ignore bound meta variables (treating them transitively) *)
               f (mapEmbedTy f ty)
          )
        
    (* If we had extensible records here, it would be enough to write
       2 functions here rather than 4!
       We only update the field ty :-) *)
    and mapEmbedTyTyConArgs f args =
          map (mapEmbedTyTyConArg f) args

    and mapEmbedTyTyConArg f {name, ty} =
          {name = name,
           ty = mapEmbedTy f ty}

    and mapEmbedTyExtRecordFields f args =
          map (mapEmbedTyExtRecordField f) args

    and mapEmbedTyExtRecordField f {lab, ty, ann} =
          {lab = lab,
           ty = mapEmbedTy f ty,
           ann = ann}

    (* Occurs check *)
    fun occursNamedMetaVar theName ty =
          foldEmbedTy (fn (t, acc) => acc orelse isNamedMetaVar theName t)
                      false ty

    (* The mapping function for instantiating type variables *)
    fun instantiateTyVarsM mapEnv (t as (KsiAbs.TyVar {name, restr, ann})) =
          (case ST.lookup mapEnv (ST.symbol name) of
             NONE    => t
           | SOME ty => ty
          )
      | instantiateTyVarsM _ t = t

    (* Instantiate type variables in ty *)
    fun instantiateTyVars mapEnv ty =
          mapEmbedTy (instantiateTyVarsM mapEnv) ty

    (* Make an environment from parameters mapping them
       into fresh meta variables. *)
    fun paramsTyFreshMetaVarsF ({name, restr}, mapEnv) =
          ST.insert (ST.symbol name) (newMetaTyRestr restr) mapEnv

    fun paramsToFreshMetaVars mapEnv params =
          foldl paramsTyFreshMetaVarsF mapEnv params

    (* Make a type generalized by a type scheme,
       so that it becomes a type ready for further unification. *)
    fun generalizingTyScheme {params, ty, ann} =
          instantiateTyVars (paramsToFreshMetaVars (ST.empty ()) params) ty

    (* Extract explicit (unbound) meta variables from a type *)
    fun metavarsTyF (KsiAbs.TyMeta {name, restr, bound, ann}, acc) =
          (case !bound of
             NONE =>
               ST.insert (ST.symbol name) restr acc
           | SOME _ =>
               acc
          )
      | metavarsTyF (_, acc) =
          acc

    fun metavarsTy ty =
          foldEmbedTy metavarsTyF (ST.empty ()) ty

    (* Extract explicit type variables from a type *)
    fun tyvarsTyF (KsiAbs.TyVar {name, restr, ann}, acc) =
          ST.insert (ST.symbol name) restr acc
      | tyvarsTyF (_, acc) =
          acc

    fun tyvarsTy ty =
          foldEmbedTy tyvarsTyF (ST.empty ()) ty

    (* Extract parameters for a type to produce a type scheme *)
    fun tySchemeFromTy ty =
          let
            val tyvars   = tyvarsTy   ty
            val metavars = metavarsTy ty

            fun accParamF (name, restr, acc) =
                  {name = ST.name name, restr = restr}::acc

            val params = ST.foldli accParamF nil    metavars
            val params = ST.foldli accParamF params tyvars
          in
            {params = params,
             ty     = ty,
             ann    = ()} : unit KsiAbs.tyscheme
          end

    fun tyToStr ty =
          KsiPrettyPrint.tySchemeToStr (tySchemeFromTy ty)

    (* Lookup function for order rows *)
    fun getOrderRowField r theLab =
          case List.find (fn {lab, ty, ann} => theLab = lab) r of
            NONE =>
              reportInternalErrorStrs 0wx0102
                ["Order row label ", theLab, " not found!"]
          | SOME {lab, ty, ann} =>
              ty

    (* Lookup function for TyCon arguments *)
    fun getTyConArg args theName =
          case List.find (fn {name, ty} => theName = name) args of
            NONE =>
              reportInternalErrorStrs 0wx0103
                ["TyCon argument name ", theName, " not found!"]
          | SOME {name, ty} =>
              ty

    (* Lookup function for TyPat arguments *)
    fun getTyPatArg patArgs theName =
          case List.find (fn {name, pat} => theName = name) patArgs of
            NONE =>
              reportInternalErrorStrs 0wx0104
                ["TyPat argument name ", theName, " not found!"]
          | SOME {name, pat} =>
              pat

    (* Convert type pattern to type *)
    fun restrTyPatToTy (KsiAbs.RstrTyPat {args, name, tags}) =
          KsiAbs.TyCon
            {args = map (fn {name, pat} =>
                           {name = name,
                            ty = restrTyPatToTy pat})
                        args,
             name = name,
             tags = tags,
             ann = ()}

    (* Checks on unique variable bindings *)
    fun checkUniqueVarBind boundE name =
          (case ST.lookup boundE (ST.symbol name) of
             NONE =>
               ST.insert (ST.symbol name) () boundE
           | SOME _ =>
               (reportErrorStrs 0wx0105
                  ["The variable ", name, " is bound more than once ",
                   "in the same val-binding, which is not legal"];
                boundE)
          )

    fun checkUniqueVarBindPat boundE (KsiAbs.PatVar {name, ann}) =
          checkUniqueVarBind boundE name
      | checkUniqueVarBindPat boundE (KsiAbs.PatTyped {pat, ty, ann}) =
          checkUniqueVarBindPat boundE pat
      | checkUniqueVarBindPat boundE (KsiAbs.PatLayered {vid, pat, ann}) =
          checkUniqueVarBindPat
            (checkUniqueVarBind boundE vid) pat
      | checkUniqueVarBindPat boundE (KsiAbs.PatCon {tag, tyScheme, ann}) =
          boundE
      | checkUniqueVarBindPat boundE (KsiAbs.PatParamCon {tag, tyScheme, pat, ann}) =
          checkUniqueVarBindPat boundE pat
      | checkUniqueVarBindPat boundE (KsiAbs.PatEx {tag, ann}) =
          boundE
      | checkUniqueVarBindPat boundE (KsiAbs.PatParamEx {tag, pat, ty, ann}) =
          checkUniqueVarBindPat boundE pat
      | checkUniqueVarBindPat boundE (KsiAbs.PatSCon {scon, ann}) =
          boundE
      | checkUniqueVarBindPat boundE (KsiAbs.PatWild {ann}) =
          boundE
      | checkUniqueVarBindPat boundE (KsiAbs.PatEmptyRecord {ann}) =
          boundE
      | checkUniqueVarBindPat boundE (KsiAbs.PatExtRecord {fields, ext, ann}) =
          checkUniqueVarBindPat
            (checkUniqueVarBindFields boundE fields) ext
      | checkUniqueVarBindPat boundE (KsiAbs.PatOptField {present, pat, ann}) =
          checkUniqueVarBindPat
            (checkUniqueVarBindPat boundE present) pat
      
    and checkUniqueVarBindFields boundE fields =
          foldl checkUniqueVarBindFieldF boundE fields

    and checkUniqueVarBindFieldF ({lab, pat, ann}, boundE) =
          checkUniqueVarBindPat boundE pat

    fun checkUniqueVarBindBindF ({pat, exp, ann}, boundE) =
          checkUniqueVarBindPat boundE pat

    fun checkUniqueVarBindBinds boundE binds =
          foldl checkUniqueVarBindBindF boundE binds

    (* Unification *)
    fun moreRestrictive (_, KsiAbs.RstrNone) =
          true
      | moreRestrictive (KsiAbs.RstrExclude {labs},
                         KsiAbs.RstrExclude {labs = labs2}) =
          (* labs2 must be a subset of labs *)
          isSubsetSortedSets op= labs2 labs
      | moreRestrictive (KsiAbs.RstrTyPats {pats},
                         KsiAbs.RstrTyPats {pats = pats2}) =
          let
            val len = length pats
            val len2 = length pats2
          in
            if 1 <= len andalso 1 <= len2 then
              len <= len2 andalso isSubset eqTyPat pats pats2
            else
              (reportError 0wx0106 "Type pattern sequence pats may not be empty";
               false)
          end
      | moreRestrictive (r1, r2) =
          (* r1 is not more restrictive than r2 *)
          false

    fun tyPatRespectsRestr (t, KsiAbs.RstrNone) =
          true
      | tyPatRespectsRestr (KsiAbs.RstrTyPat {args = _, name = n, tags = _},
                            KsiAbs.RstrTyPats {pats}) =
          (* It should be enough to check equality of the names to
             determine if the type patterns are entirely the same,
             because of the uniqueness of names. *)
          List.exists (fn (KsiAbs.RstrTyPat {args, name, tags}) => name = n)
                      pats
      | tyPatRespectsRestr _ =
          false

    fun tyRespectsRestr (KsiAbs.TyVar {name, restr, ann}, theRestr) =
          moreRestrictive (restr, theRestr)
      | tyRespectsRestr (t, KsiAbs.RstrNone) =
          true
      | tyRespectsRestr (t, KsiAbs.RstrTyPats {pats}) =
          List.exists (fn pat => tyRespectsRestrTyPat (t, pat)) pats
      | tyRespectsRestr (KsiAbs.TyEmptyRecord {ann}, theRestr) =
          (case theRestr of
             KsiAbs.RstrExclude _ =>
               true
           | _ =>
               false
          )
      | tyRespectsRestr _ =
          (* Anything else does not respect the restriction *)
          false

    and tyRespectsRestrTyPat (KsiAbs.TyMeta {name, restr, bound, ann}, thePat) =
          (case !bound of
             NONE =>
               (* We treat this as an error in the semantics.
                  Otherwise it would require unification of restrictions
                  which is NOT trivial to add to xi-Calculus! *)
               false
           | SOME t =>
               tyRespectsRestrTyPat (t, thePat)
          )
      | tyRespectsRestrTyPat (KsiAbs.TyVar {name, restr, ann}, thePat) =
          moreRestrictive (restr, KsiAbs.RstrTyPats {pats = [thePat]})
      | tyRespectsRestrTyPat (KsiAbs.TyCon {args, name, tags, ann = _},
                              KsiAbs.RstrTyPat {args = patArgs, name = patName, tags = patTags}) =
          if name = patName then
            (* The type names match - check the parameters *)
            let
              val argNames = map #name args
              val patArgNames = map #name patArgs
            in
              if isEqualSets op= argNames patArgNames then
                tyRespectsRestrTyConArgs args patArgs
              else
                (* If args and patArgs does not have the same parameters
                   it must be an error in the parser or the translation to
                   xi-Calculus! *)
                reportInternalErrorStrs 0wx0107
                  ["The constructor ", name,
                   " occurs with differing sets of parameter names ",
                   "in a restriction!"]
            end
          else
            (* Type names do not match *)
            false
      | tyRespectsRestrTyPat (_, _) =
          (* Anything else is irrespectful of the type :-) *)
          false
      
    and tyRespectsRestrTyConArgs args patArgs =
          foldl (fn ({name, ty}, acc) =>
                   acc andalso
                   (tyRespectsRestrTyPat (ty, getTyPatArg patArgs name)))
                true args

    (* Make record fields independent of order *)
    fun makeOrderRow errPos (ty as (KsiAbs.TyExtRecord {fields, ext, ann})) exclLabs =
          (case fields of
             nil =>
               (reportErrorAtPos errPos 0wx0108
                  "Encountered extensible record without any fields";
                makeOrderRow errPos ext exclLabs)
           | _ =>
               (* There is at least one field as required *)
               let
                 val labs = map #lab fields
               in
                 if isDisjointSortedSets op< op= labs exclLabs then
                   let
                     val (orderRow, ty) = makeOrderRow errPos ext
                                            (makeUnionSortedSets op< op= labs exclLabs)
                   in
                     (makeUnionSortedSets
                        ltRecordField eqRecordField fields orderRow,
                      ty)
                   end
                 else
                   let
                     val forbiddenLabs = makeIntersectSortedSets op< op= labs exclLabs
                   in
                     (reportErrorAtPosStrs errPos 0wx0109
                        ("Record contains the following forbidden fields: "::
                         (KsiPrettyPrint.labsToStrs forbiddenLabs));
                      makeOrderRow errPos ext
                        (makeUnionSortedSets op< op= labs exclLabs))
                   end
               end
          )
      | makeOrderRow errPos (ty as (KsiAbs.TyEmptyRecord _)) exclLabs =
          (nil, ty)
      | makeOrderRow errPos (ty as (KsiAbs.TyMeta {name, restr, bound, ann})) exclLabs =
          (case !bound of
             NONE =>
               (case restr of
                  KsiAbs.RstrExclude {labs} =>
                    if isSubsetSortedSets op= exclLabs labs then
                      (nil, ty)
                    else
                      let
                        val freshLabs = makeUnionSortedSets op< op= exclLabs labs
                        val freshTy = newMetaTyRestr
                                        (KsiAbs.RstrExclude
                                           {labs = freshLabs})
                        val _ = bound := SOME freshTy
                      in
                        (nil, freshTy)
                      end
                | KsiAbs.RstrNone =>
                    let
                      val freshTy = newMetaTyRestr
                                      (KsiAbs.RstrExclude
                                         {labs = exclLabs})
                      val _ = bound := SOME freshTy
                    in
                      (nil, freshTy)
                    end
                | _ =>
                    (reportErrorAtPosStrs errPos 0wx0110
                       ["Record does not unify with the meta variable: ", name,
                        " since it is restricted to be a non-record type"];
                     (nil, ty))
               )
           | SOME t =>
               makeOrderRow errPos t exclLabs
          )
      | makeOrderRow errPos (ty as (KsiAbs.TyVar {name, restr, ann})) exclLabs =
          (case restr of
             KsiAbs.RstrExclude {labs} =>
               if isSubsetSortedSets op= exclLabs labs then
                 (nil, ty)
               else
                 let
                   val forbiddenLabs = makeDifferenceSortedSets op< op= exclLabs labs
                 in
                   (reportErrorAtPosStrs errPos 0wx0111
                      ("Record does not unify with the type variable: "::name::
                       " since it contains the following forbidden fields: "::
                       (KsiPrettyPrint.labsToStrs forbiddenLabs));
                    (nil, ty))
                 end
           | _ =>
               (reportErrorAtPosStrs errPos 0wx0112
                  ["Record does not unify with the type variable: ", name,
                   " since it is not restricted to be a record type"];
                (nil, ty))
          )
      | makeOrderRow errPos ty exclLabs =
          (reportErrorAtPosStrs errPos 0wx0113
             ["The type:\n", tyToStr ty,
              "\nis not a record where this is expected (CXL2007)"];
           (nil, ty))
    
    fun reportTooManyLabelsDifferences pos (nil, nil) =
          (* Hey! The sets were supposed to be different! *)
          reportInternalErrorAtPos pos 0wx0114
            "Internal error - most likely in the set-operations!"
      | reportTooManyLabelsDifferences pos (nil, tooMany2) =
          reportErrorAtPosStrs pos 0wx0115
            ["Two non-extensible records failed to unify ",
             "because the second record contains the following labels ",
             "which are not present in the first record:\n",
             KsiPrettyPrint.labsToStr tooMany2]
      | reportTooManyLabelsDifferences pos (tooMany, nil) =
          reportErrorAtPosStrs pos 0wx0116
            ["Two non-extensible records failed to unify ",
             "because the first record contains the following labels ",
             "which are not present in the second record:\n",
             KsiPrettyPrint.labsToStr tooMany]
      | reportTooManyLabelsDifferences pos (tooMany, tooMany2) =
          reportErrorAtPosStrs pos 0wx0117
            ["Two non-extensible records failed to unify ",
             "because the first record contains the following labels ",
             "which are not present in the second record:\n",
             KsiPrettyPrint.labsToStr tooMany,
             "\nand the second record contains the following labels ",
             "which are not present in the first record:\n",
             KsiPrettyPrint.labsToStr tooMany2]

    fun reportOrderRowDifferences pos (labs, labs2) =
          let
            val tooMany = makeDifferenceSet op= labs labs2
            val tooMany2 = makeDifferenceSet op= labs2 labs
          in
            reportTooManyLabelsDifferences pos (tooMany, tooMany2)
          end

    fun unify errPos (t1 as (KsiAbs.TyMeta {name, restr, bound, ann}) : unit KsiAbs.ty, t2 : unit KsiAbs.ty) =
          (case !bound of
             SOME t =>
               (unify errPos (t, t2)) : error
           | NONE =>
               (case t2 of
                  (KsiAbs.TyMeta
                     {name = name2, restr = restr2,
                      bound = bound2, ann}) =>
                     (case !bound2 of
                        SOME t =>
                          (* Since the mete variable name2 is bound to
                             a type, we have that name2 <> name,
                             since meta var name is not bound to any type. *)
                          unify errPos (t1, t)
                      | NONE =>
                          if name = name2 then
                            (* This is OK *)
                            NoErr
                          else
                            if moreRestrictive (restr2, restr) then
                              (bound := (SOME t2);
                               NoErr)
                            else if moreRestrictive (restr, restr2) then
                              (bound2 := (SOME t1);
                               NoErr)
                            else
                              (reportErrorAtPosStrs errPos 0wx0118
                                 ["Incompatiple restrictions: ",
                                  KsiPrettyPrint.restrToStr restr,
                                  " and: ",
                                  KsiPrettyPrint.restrToStr restr2];
                               ReportedErr)
                     )
                | (KsiAbs.TyExtRecord _) =>
                    (case restr of
                       KsiAbs.RstrExclude _ =>
                         (* The syntactic restrictions must ensure that
                            fields and fields2 are not nil.
                            It is also double-checked by makeOrderRow. *)
                         unifyOrderRow errPos (nil, t1) (makeOrderRow errPos t2 nil)
                     | KsiAbs.RstrNone =>
                         let
                           val freshMeta = newMetaTyRestr
                                             (KsiAbs.RstrExclude
                                                {labs = nil})
                         in
                           if occursNamedMetaVar name t2 then
                             (reportErrorAtPosStrs errPos 0wx0119
                                [name, " occurs in type ",
                                 tyToStr t2];
                              ReportedErr)
                           else
                             (bound := (SOME freshMeta);
                              (* The syntactic restrictions must ensure that
                                 fields and fields2 are not nil.
                                 It is also double-checked by makeOrderRow. *)
                              unifyOrderRow errPos (nil, freshMeta)
                                                   (makeOrderRow errPos t2 nil)
                             )
                         end
                     | _ =>
                         (reportErrorAtPosStrs errPos 0wx0120
                            ["Illegal restriction: ",
                             KsiPrettyPrint.restrToStr restr,
                             " for extensible record type:\n",
                             tyToStr t2];
                          ReportedErr)
                    )
                | _ =>
                    (* Here t2 is neither a TyMeta nor a TyExtRecord *)
                    if occursNamedMetaVar name t2 then
                      (reportErrorAtPosStrs errPos 0wx0121
                         [name, " occurs in type ", tyToStr t2];
                       ReportedErr)
                    else
                      if tyRespectsRestr (t2, restr) then
                        (bound := (SOME t2);
                         NoErr)
                      else
                        (reportErrorAtPosStrs errPos 0wx0122
                           ["The type:\n", tyToStr t2,
                            "\ndoes not respect the restriction:\n  ",
                            KsiPrettyPrint.restrToStr restr];
                         ReportedErr)
               )
          )
      | unify errPos (t1, t2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unify errPos (t2, t1)
      | unify errPos (KsiAbs.TyVar {name, restr, ann},
                      KsiAbs.TyVar {name = name2, restr = restr2, ann = _}) =
          if name = name2 then
            (* This is OK *)
            NoErr
          else
            (reportErrorAtPosStrs errPos 0wx0123
               ["Type variables ", name, " and ", name2,
                " cannot be unified"];
             ReportedErr)
      | unify errPos (KsiAbs.TyFun {from, to, ann},
                      KsiAbs.TyFun {from = from2, to = to2, ann = _}) =
          combineErrorCodes
            (unify errPos (from, from2),
             unify errPos (to, to2))
      | unify errPos (KsiAbs.TyField {present, ty, ann},
                      KsiAbs.TyField {present = present2, ty = ty2, ann = _}) =
          combineErrorCodes
            (unify errPos (present, present2),
             unify errPos (ty, ty2))
      | unify errPos (KsiAbs.TyCon {args, name, tags, ann},
                      KsiAbs.TyCon {args = args2, name = name2, tags = tags2, ann = _}) =
          if name = name2 then
            let
              val argNames = map #name args
              val argNames2 = map #name args2
            in
              if isEqualSets op= argNames argNames2 then
                unifyTyConArgs errPos args args2
              else
                (* If args and args2 does not have the same parameters
                   it must be an error in the parser or the translation to
                   xi-Calculus! *)
                (reportInternalErrorStrs 0wx0124
                   ["The constructor ", name,
                    " occurs 2 places with differing sets of parameter names!"];
                 ReportedErr)
            end
          else
            (reportErrorAtPosStrs errPos 0wx0125
               ["Differing type constructors ", name, " and ",
                name2, " cannot be unified"];
             ReportedErr)
      | unify errPos (t as (KsiAbs.TyExtRecord {fields, ext, ann}),
                      t2 as (KsiAbs.TyExtRecord {fields = fields2, ext = ext2, ann = _})) =
          (* The syntactic restrictions must ensure that
             fields and fields2 are not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRow errPos (makeOrderRow errPos t  nil)
                               (makeOrderRow errPos t2 nil)
      | unify errPos (t as (KsiAbs.TyExtRecord {fields, ext, ann}),
                      t2 as (KsiAbs.TyEmptyRecord _)) =
          (* The syntactic restrictions must ensure that fields is not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRowWithEmpty errPos (makeOrderRow errPos t nil)
      | unify errPos (t as (KsiAbs.TyEmptyRecord _),
                      t2 as (KsiAbs.TyExtRecord {fields = fields2, ext = ext2, ann = _})) =
          (* The syntactic restrictions must ensure that fields is not nil.
             It is also double-checked by makeOrderRow. *)
          unifyOrderRowWithEmpty errPos (makeOrderRow errPos t2 nil)
      | unify errPos (KsiAbs.TyEmptyRecord _, KsiAbs.TyEmptyRecord _) =
          (* This is OK *)
          NoErr
      | unify errPos (t, t2) =
          (reportErrorAtPosStrs errPos 0wx0126
             ["Cannot unify the type:\n",
              tyToStr t,
              "\nwith the type:\n",
              tyToStr t2];
           ReportedErr)

    and unifyTyConArgs errPos args args2 =
          List.foldl (fn ({name, ty}, acc) =>
                        combineErrorCodes
                          (acc, unify errPos (ty, getTyConArg args2 name)))
                     NoErr args

    and unifyOrderRow errPos
            ((r,  KsiAbs.TyEmptyRecord _) : ((unit KsiAbs.orderrow) * (unit KsiAbs.ty)))
            ((r2, KsiAbs.TyEmptyRecord _) : ((unit KsiAbs.orderrow) * (unit KsiAbs.ty))) =
          let
            val rInters = makeIntersectSortedSets
                            ltRecordField eqRecordField r r2
            val rExtra  = makeDifferenceSortedSets
                            ltRecordField eqRecordField r r2
            val rExtra2 = makeDifferenceSortedSets
                            ltRecordField eqRecordField r2 r
            val labsExtra = map #lab rExtra
            val labsExtra2 = map #lab rExtra2

            (* Unify the record fields which are in r
               but not in r2 with absent fields. *)
            val uniErr = unifyAbsentFields errPos rExtra

            (* Unify the record fields which are in r2
               but not in r with absent fields. *)
            val uniErr2 = unifyAbsentFields errPos rExtra2
          in
            combineErrorCodes
              (combineErrorCodes (uniErr, uniErr2),
               (* Unify the intersection record (from r) with r2 *)
               unifyOrderRowFields errPos (rInters : unit KsiAbs.orderrow)
                                          (r2 : unit KsiAbs.orderrow))
          end
      | unifyOrderRow errPos (r,  KsiAbs.TyVar {name, restr, ann})
                             (r2, KsiAbs.TyVar {name = name2, restr = restr2, ann = _}) =
          if name = name2 then
            (* The type variables are the same as required *)
            (* Since the type variables come from order rows
               it is already guaranteed that the restrictions are
               excluding labels and that they are OK. *)
            let
              val rInters = makeIntersectSortedSets
                              ltRecordField eqRecordField r r2
              val rExtra  = makeDifferenceSortedSets
                              ltRecordField eqRecordField r r2
              val rExtra2 = makeDifferenceSortedSets
                              ltRecordField eqRecordField r2 r
              val labsExtra = map #lab rExtra
              val labsExtra2 = map #lab rExtra2

              (* Unify the record fields which are in r
                 but not in r2 with absent fields. *)
              val uniErr = unifyAbsentFields errPos rExtra

              (* Unify the record fields which are in r2
                 but not in r with absent fields. *)
              val uniErr2 = unifyAbsentFields errPos rExtra2
            in
              combineErrorCodes
                (combineErrorCodes (uniErr, uniErr2),
                 (* Unify the intersection record (from r) with r2 *)
                 unifyOrderRowFields errPos rInters r2)
            end
          else
            (reportErrorAtPosStrs errPos 0wx0127
               ["Differing type variables: ", name, " and: ", name2,
                " cannot be unified"];
             ReportedErr)
      | unifyOrderRow errPos (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                             (r2, KsiAbs.TyMeta {name = name2, restr = restr2,
                                                 bound = bound2, ann = ann2}) =
          (case (restr, restr2) of
             (KsiAbs.RstrExclude {labs = labs},
              KsiAbs.RstrExclude {labs = labs2}) =>
                (* Restrictions are records as required *)
                let
                  val rInters = makeIntersectSortedSets
                                  ltRecordField eqRecordField r r2
                  val rExtra  = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r r2
                  val rExtra2 = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r2 r
                  val labsExtra  = map #lab rExtra
                  val labsExtra2 = map #lab rExtra2
                in
                  if isDisjointSortedSets String.< op= labsExtra labs2 then
                    if isDisjointSortedSets String.< op= labsExtra2 labs then
                      (case (!bound, !bound2) of
                         (NONE, NONE) =>
                            let
                              (* First we handle meta variables
                                 since it is a side effect *)
                              val freshLabs = makeUnionSortedSets op< op= labs labs2
                              val freshTy = newMetaTyRestr
                                              (KsiAbs.RstrExclude
                                                 {labs = freshLabs})
                              val rest = case rExtra of
                                           nil =>
                                             freshTy
                                         | _ =>
                                             KsiAbs.TyExtRecord
                                               {fields = rExtra,
                                                ext = freshTy,
                                                ann = ann}
                              val rest2 = case rExtra2 of
                                            nil =>
                                              freshTy
                                          | _ =>
                                              KsiAbs.TyExtRecord
                                                {fields = rExtra2,
                                                 ext = freshTy,
                                                 ann = ann2}
                            in
                              (* Make the occurs checks before binding
                                 the types. *)
                              if occursNamedMetaVar name rest2 then
                                (reportErrorAtPosStrs errPos 0wx0128
                                   [name, " occurs in type ",
                                    tyToStr rest2];
                                 ReportedErr)
                              else if occursNamedMetaVar name2 rest then
                                (reportErrorAtPosStrs errPos 0wx0129
                                   [name2, " occurs in type ",
                                    tyToStr rest];
                                 ReportedErr)
                              else
                                let
                                  val _ = bound := SOME rest2
                                  val _ = bound2 := SOME rest
                                in
                                  (* Unify the record fields in the
                                     intersection of the 2 records *)
                                  unifyOrderRowFields errPos rInters r2
                                end
                            end
                       | _ =>
                            (reportInternalErrorStrs 0wx012a
                               ["A meta variable for an ",
                                "order row is bound"];
                             ReportedErr)
                      )
                    else
                      let
                        val forbiddenLabs2 = makeIntersectSortedSets
                                               String.< op= labsExtra2 labs
                      in
                        (reportErrorAtPosStrs errPos 0wx012b
                           ("Two records do not unify because one record "::
                            "contains the following forbidden fields: "::
                            (KsiPrettyPrint.labsToStrs forbiddenLabs2));
                         ReportedErr)
                      end
                  else
                    let
                      val forbiddenLabs = makeIntersectSortedSets
                                            String.< op= labsExtra labs2
                    in
                      (reportErrorAtPosStrs errPos 0wx012c
                         ("Two records do not unify because one record "::
                          "contains the following forbidden fields: "::
                          (KsiPrettyPrint.labsToStrs forbiddenLabs));
                       ReportedErr)
                    end
                end
           | _ =>
                (reportInternalErrorStrs 0wx012d
                   ["Meta variables for order row are not ",
                    "restricted to be records as expected"];
                 ReportedErr)
          )
      | unifyOrderRow errPos (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                             (r2, ty2 as (KsiAbs.TyVar {name = name2, restr = restr2, ann = ann2})) =
          (case (restr, restr2) of
             (KsiAbs.RstrExclude {labs = labs},
              KsiAbs.RstrExclude {labs = labs2}) =>
                (* Restrictions are records as required *)
                let
                  val rExtra  = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r r2
                  val rExtra2 = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r2 r
                  val labsExtra  = map #lab rExtra
                  val labsExtra2 = map #lab rExtra2
                in
                  if isSubsetSortedSets op= labs labs2 then
                    if isDisjointSortedSets String.< op= labsExtra labs2 then
                      if isDisjointSortedSets String.< op= labsExtra2 labs then
                        (case !bound of
                           NONE =>
                              let
                                (* First we handle meta variables
                                   since it is a side effect *)
                                val rest = case rExtra2 of
                                             nil =>
                                               ty2
                                           | _ =>
                                               KsiAbs.TyExtRecord
                                                 {fields = rExtra2,
                                                  ext = ty2,
                                                  ann = ann2}
                              in
                                (* Make the occurs check before binding
                                   the type. *)
                                if occursNamedMetaVar name rest then
                                  (reportErrorAtPosStrs errPos 0wx012e
                                     [name, " occurs in type ",
                                      tyToStr rest];
                                   ReportedErr)
                                else
                                  let
                                    val _ = bound := SOME rest
                                  in
                                    (* Unify the record fields in the record r *)
                                    unifyOrderRowFields errPos r r2
                                  end
                              end
                         | _ =>
                              (reportInternalErrorStrs 0wx012f
                                 ["A meta variable for an ",
                                  "order row is bound"];
                               ReportedErr)
                        )
                      else
                        let
                          val forbiddenLabs2 = makeIntersectSortedSets
                                                 String.< op= labsExtra2 labs
                        in
                          (* If this happens, either the error should already be
                             reported or there is an internal error. *)
                          if errorReported () then
                            ReportedErr
                          else
                            (reportInternalErrorAtPosStrs errPos 0wx0130
                               ("Two records do not unify because one record "::
                                "contains the following forbidden fields: "::
                                (KsiPrettyPrint.labsToStrs forbiddenLabs2));
                             ReportedErr)
                        end
                    else
                      let
                        val forbiddenLabs = makeDifferenceSortedSets
                                              String.< op= labsExtra labs2
                      in
                        (reportErrorAtPosStrs errPos 0wx0131
                           ("Two records do not unify because one record "::
                            "contains the following forbidden fields: "::
                            (KsiPrettyPrint.labsToStrs forbiddenLabs));
                         ReportedErr)
                      end
                  else
                    let
                      val forbiddenLabs = makeDifferenceSortedSets
                                            String.< op= labs labs2
                    in
                      (reportErrorAtPosStrs errPos 0wx0132
                         ("Two records do not unify because one record "::
                          "is extended with a meta variable restricted "::
                          "with the following extra fields: "::
                          (KsiPrettyPrint.labsToStrs forbiddenLabs));
                       ReportedErr)
                    end
                end
           | (KsiAbs.RstrExclude {labs}, _) =>
                (reportErrorAtPosStrs errPos 0wx0133
                   ["Type variable " ^ name2 ^ " for ",
                    "order row is not restricted to be a record as expected"];
                 ReportedErr)
           | _ =>
                (* We expect the rest of the semantics to ensure that this
                   doesn't happen, so this is an internal error. *)
                (reportInternalErrorAtPosStrs errPos 0wx0134
                   ["Meta variable " ^ name ^ " for ",
                    "order row is not restricted to be a record as expected"];
                 ReportedErr)
          )
      | unifyOrderRow errPos (r,  KsiAbs.TyMeta {name, restr, bound, ann})
                             (r2, ty2 as (KsiAbs.TyEmptyRecord {ann = ann2})) =
          (case restr of
             KsiAbs.RstrExclude {labs} =>
                (* Restrictions are records as required *)
                let
                  val ls = map #lab r
                  val ls2 = map #lab r2
                  val rInters = makeIntersectSortedSets
                                  ltRecordField eqRecordField r r2
                  val rExtra  = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r r2
                  val rExtra2 = makeDifferenceSortedSets
                                  ltRecordField eqRecordField r2 r
                  val labsExtra2 = map #lab rExtra2
                in
                  if isDisjointSortedSets String.< op= labsExtra2 labs then
                    (case !bound of
                       NONE =>
                         let
                           (* First we handle meta variables
                              since it is a side effect *)
                           val rest = case rExtra2 of
                                        nil =>
                                          ty2
                                      | _ =>
                                          KsiAbs.TyExtRecord
                                            {fields = rExtra2,
                                             ext = ty2,
                                             ann = ann2}
                         in
                           (* Make the occurs check before binding
                              the type. *)
                           if occursNamedMetaVar name rest then
                             (reportErrorAtPosStrs errPos 0wx0135
                                [name, " occurs in type ",
                                 tyToStr rest];
                              ReportedErr)
                           else
                             let
                               val _ = bound := SOME rest
                             
                               (* Unify the record fields which are in r
                                  but not in r2 with absent fields. *)
                               val uniErr = unifyAbsentFields errPos rExtra
                             in
                               (* Unify the record fields in the
                                  intersection of the 2 records *)
                               combineErrorCodes
                                 (unifyOrderRowFields errPos rInters r2,
                                  uniErr)
                             end
                         end
                     | _ =>
                         (reportInternalErrorAtPosStrs errPos 0wx0136
                            ["A meta variable for an ",
                             "order row is bound"];
                          ReportedErr)
                    )
                  else
                    let
                      val forbiddenLabs2 = makeIntersectSortedSets
                                             String.< op= labsExtra2 labs
                    in
                      (reportErrorAtPosStrs errPos 0wx0137
                         ("Two records do not unify because one record "::
                          "contains the following forbidden fields: "::
                          (KsiPrettyPrint.labsToStrs forbiddenLabs2));
                       ReportedErr)
                    end
                end
           | _ =>
                (reportInternalErrorAtPosStrs errPos 0wx0138
                   ["Meta variable for ",
                    "order row is not restricted to be a record as expected"];
                 ReportedErr)
          )
      | unifyOrderRow errPos (r,  ty  as (KsiAbs.TyVar _))
                             (r2, ty2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unifyOrderRow errPos (r2, ty2) (r, ty)
      | unifyOrderRow errPos (r,  ty  as (KsiAbs.TyEmptyRecord _))
                             (r2, ty2 as (KsiAbs.TyMeta _)) =
          (* Unification is symmetric *)
          unifyOrderRow errPos (r2, ty2) (r, ty)
      | unifyOrderRow errPos (_, ty1) (_, ty2) =
          (reportErrorAtPosStrs errPos 0wx0139
             ["Two extensible records do not unify. One records is extended with:\n",
              tyToStr ty1,
              "\nwhile the other is extended with:\n",
              tyToStr ty2];
           ReportedErr)

    (* Inefficient unifyOrderRowFields, which works with unsorted fields *)
    (*
    and unifyOrderRowFields errPos r r2 =
          List.foldl (fn ({lab, ty, ann}, acc) =>
                        combineErrorCodes
                          (acc, unify errPos (ty, getOrderRowField r2 lab)))
                     NoErr r
    *)

    (* Efficient unifyOrderRowFields, which only works with sorted fields *)
    and unifyOrderRowFields errPos r r2 =
          unifyOrderRowFieldsIter errPos NoErr r r2

    and unifyOrderRowFieldsIter errPos acc
                                ({lab : string, ty, ann}::r)
                                ({lab = lab', ty = ty', ann = _}::r2) =
          if lab = lab' then
            unifyOrderRowFieldsIter errPos
              (combineErrorCodes
                 (acc, unify errPos (ty, ty')))
              r r2
          else
            (* skip element in r2 *)
            unifyOrderRowFieldsIter
              errPos acc ({lab = lab, ty = ty, ann = ann}::r) r2
      | unifyOrderRowFieldsIter errPos acc _ _ =
          acc
    
    and unifyAbsentFields errPos r =
          List.foldl (fn ({lab, ty, ann}, acc) =>
                        combineErrorCodes
                          (acc, unify errPos (ty, newAbsentField ())))
                     NoErr r

    and unifyOrderRowWithEmpty errPos (r, KsiAbs.TyEmptyRecord _) =
          unifyAbsentFields errPos r
      | unifyOrderRowWithEmpty errPos (r, KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               (* First bind type variable since this causes a side-effect. *)
               (bound := SOME (KsiAbs.TyEmptyRecord {ann = ()});
                unifyAbsentFields errPos r)
           | _ =>
               (reportInternalErrorAtPosStrs errPos 0wx013a
                  ["A meta variable for an ",
                   "order row is bound"];
                ReportedErr)
          )
      | unifyOrderRowWithEmpty errPos (r, ty) =
          (reportErrorAtPosStrs errPos 0wx013b
             ["Cannot unify the type:\n",
              tyToStr ty,
              "\nwith the empty record"];
           ReportedErr)

    fun reportTypeDiffOneWay msg code (ty1, ty2) =
          (case diffTy (ty1, ty2) of
             NONE =>
               ()
           | SOME ty =>
               (case diffTy (ty1, ty) of
                  NONE =>
                    () (* Don't report anything if the difference is 100% *)
                | SOME _ =>
                    reportErrorFurtherStrs
                      [msg, tyToStr ty]
               )
          )

    fun reportTypeDiff (ty1, ty2) =
          (reportTypeDiffOneWay
             "Type information in first type which is not in second type:\n"
             (ty1, ty2);
           reportTypeDiffOneWay
             "Type information in second type which is not in first type:\n"
             (ty2, ty1)
          )

    fun unifyReport errPos (ty1, ty2) =
          (case unify errPos (ty1, ty2) of
             NoErr =>
               NoErr
           | NewErr =>
               (reportErrorAtPosStrs errPos 0wx013c
                  ["Error occurred while trying to unify the type:\n",
                   tyToStr ty1, "\nwith the type:\n",
                   tyToStr ty2];
                reportTypeDiff (ty1, ty2);
                ReportedErr)
           | ReportedErr =>
               (reportErrorFurtherStrs
                  ["Error occurred while trying to unify the type:\n",
                   tyToStr ty1, "\nwith the type:\n",
                   tyToStr ty2];
                reportTypeDiff (ty1, ty2);
                ReportedErr)
          )

    (* Extract explict type variables from a pattern *)
    fun tyvarsPat (KsiAbs.PatVar {name, ann}) =
          ST.empty ()
      | tyvarsPat (KsiAbs.PatLayered {vid, pat, ann}) =
          tyvarsPat pat
      | tyvarsPat (KsiAbs.PatTyped {pat, ty, ann}) =
          ST.append (tyvarsPat pat, tyvarsTy ty)
      | tyvarsPat (KsiAbs.PatCon {tag, tyScheme, ann}) =
          (* We assume that type schemes are closed and
             thus have no free type variables. *)
          ST.empty ()
      | tyvarsPat (KsiAbs.PatParamCon {tag, tyScheme, pat, ann}) =
          (* We assume that type schemes are closed and
             thus have no free type variables. *)
          tyvarsPat pat
      | tyvarsPat (KsiAbs.PatEx {tag, ann}) =
          ST.empty ()
      | tyvarsPat (KsiAbs.PatParamEx {tag, pat, ty, ann}) =
          ST.append (tyvarsPat pat, tyvarsTy ty)
      | tyvarsPat (KsiAbs.PatSCon {scon, ann}) =
          ST.empty ()
      | tyvarsPat (KsiAbs.PatWild {ann}) =
          ST.empty ()
      | tyvarsPat (KsiAbs.PatEmptyRecord {ann}) =
          ST.empty ()
      | tyvarsPat (KsiAbs.PatExtRecord {fields, ext, ann}) =
          (* The premise that there must be at least one
             field is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({lab, pat, ann}, accTyvE) =>
                   ST.append (accTyvE, tyvarsPat pat))
                (tyvarsPat ext) fields
      | tyvarsPat (KsiAbs.PatOptField {present, pat, ann}) =
          (* The premise that present must be a present constructor
             is checked elsewhere, so we don't bother with it here. *)
          (* We assume that type schemes (i.e. also present) are closed and
             thus have no free type variables. *)
          tyvarsPat pat

    (* Extract explict type variables from an expression *)
    fun tyvarsExp (KsiAbs.ExpFun {param, exp, ann}) =
          tyvarsExp exp
      | tyvarsExp (KsiAbs.ExpApp {funExp, argExp, ann}) =
          ST.append (tyvarsExp funExp, tyvarsExp argExp)
      | tyvarsExp (KsiAbs.ExpVar {name, ann}) =
          ST.empty ()
      | tyvarsExp (KsiAbs.ExpTyped {exp, ty, ann}) =
          ST.append (tyvarsExp exp, tyvarsTy ty)
      | tyvarsExp (KsiAbs.ExpCon {tag, tyScheme, ann}) =
          (* We assume that type schemes are closed and
             thus have no free type variables. *)
          ST.empty ()
      | tyvarsExp (KsiAbs.ExpEx {tag, ty, ann}) =
          tyvarsTy ty
      | tyvarsExp (KsiAbs.ExpSCon {scon, ann}) =
          ST.empty ()
      | tyvarsExp (KsiAbs.ExpEmptyRecord {ann}) =
          ST.empty ()
      | tyvarsExp (KsiAbs.ExpExtRecord {fields, ext, ann}) =
          (* The premise that there must be at least one
             field is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({lab, exp, ann}, accTyvE) =>
                   ST.append (accTyvE, tyvarsExp exp))
                (tyvarsExp ext) fields
      | tyvarsExp (KsiAbs.ExpOptField {present, exp, ann}) =
          ST.append (tyvarsExp present, tyvarsExp exp)
      | tyvarsExp (KsiAbs.ExpLet {binds, exp, ann}) =
          ST.append (tyvarsBinds binds, tyvarsExp exp)
      | tyvarsExp (KsiAbs.ExpLetRec {binds, exp, ann}) =
          (* The premise that there must be at least one
             binding is checked elsewhere, so we don't bother with it here. *)
          ST.append (tyvarsBinds binds, tyvarsExp exp)
      | tyvarsExp (KsiAbs.ExpCase {exp, matches, ann}) =
          (* The premise that there must be at least one
             match is checked elsewhere, so we don't bother with it here. *)
          tyvarsMatches (tyvarsExp exp) matches
      | tyvarsExp (KsiAbs.ExpFieldCase {exp, tyvar, absExp,
                                        prePat, preExp, tyOut, ann}) =
          let
            val tyvarE = ST.insert (ST.symbol tyvar)
                                   (makeAbsentPresentRestr ())
                                   (ST.empty ()) 
          in
            ST.append
              (ST.append
                 (tyvarsTy tyOut, tyvarE),
               ST.append
                 (ST.append
                    (tyvarsExp exp, tyvarsExp absExp),
                  ST.append
                    (tyvarsPat prePat, tyvarsExp preExp))
              )
          end
      | tyvarsExp (KsiAbs.ExpLetEx {binds, exp, ann}) =
          (* The premise that there must be at least one
             binding is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({tag, ty, ann}, accTyvE) =>
                   ST.append (accTyvE, tyvarsTy ty))
                (tyvarsExp exp) binds
      | tyvarsExp (KsiAbs.ExpHandle {exp, matches, ann}) =
          (* The premise that there must be at least one
             match is checked elsewhere, so we don't bother with it here. *)
          tyvarsMatches (tyvarsExp exp) matches
      | tyvarsExp (KsiAbs.ExpRaise {exp, ann}) =
          tyvarsExp exp

    and tyvarsMatches initTyvE matches =
          foldl (fn ({pat, exp, ann}, accTyvE) =>
                   ST.append
                     (accTyvE,
                      ST.append
                        (tyvarsPat pat,
                         tyvarsExp exp)))
                initTyvE matches

    and tyvarsBindF ({pat, exp, ann}, accTyvE) =
          ST.append (ST.append (accTyvE, tyvarsPat pat), tyvarsExp exp)
          
    and tyvarsBinds binds =
          foldl tyvarsBindF (ST.empty ()) binds

    (* Make forbidden type variables for the closure operation *)
    fun makeForbiddenTyvars tyvE binds =
          foldl tyvarsBindF tyvE binds

    (* Function for choosing legal names for type variables during closure. *)
    fun makeTyvarProducer tyvE binds =
          let
            val forbidTyvars = makeForbiddenTyvars tyvE binds

            val p = lexicogTyvarProducer ()

            fun produce () =
                  let
                    val name = p ()
                  in
                    case ST.lookup forbidTyvars (ST.symbol name) of
                      NONE =>
                        name
                    | SOME _ =>
                        produce ()
                  end
          in
            produce
          end

    (* Exclude tyvars *)
    fun tyvarsExcludeInsertF tyvE exclE (index, accE) =
          (case ST.lookup exclE index of
             NONE =>
               (* Don't exclude this, insert it *)
               (case ST.lookup tyvE index of
                  NONE =>
                    (reportInternalErrorStrs 0wx013d
                       ["tyvarsExcludeInsertF: Type variable ",
                        ST.name index, " not found where it came from!"];
                     accE)
                | SOME data =>
                    ST.insert index data accE
               )
           | SOME _ =>
               (* Exclude this *)
               accE
          )

    fun tyvarsExclude tyvE exclE =
          foldl (tyvarsExcludeInsertF tyvE exclE)
                (ST.empty ()) (ST.dimensions tyvE)

    (* Check if pat binds the variable vid *)
    fun patHasVariable pat vid =
          (* We use checkUniqueVarBindPat to build an
             environment of all variables bound by pat. *)
          (case ST.lookup (checkUniqueVarBindPat (ST.empty ()) pat)
                          (ST.symbol vid) of
             NONE =>
               false
           | SOME _ =>
               true)

    (* Find the exp for the pat that vid is bound to! *)
    fun findExpForVidBinding vid binds =
          (case List.find (fn {pat, exp, ann} => patHasVariable pat vid) binds of
             NONE =>
               NONE
           | SOME {pat, exp, ann} =>
               SOME exp)

    fun getVEMetaVars VE =
          let
            fun appendVarF ({params, ty, ann}, accMeta) =
                  ST.append (accMeta, metavarsTy ty)
          in
            ST.foldl appendVarF (ST.empty ()) VE
          end

    (* Bind an unbound meta variable to a type variable *)
    fun bindAsTyVar (tyvProd, bound, restr, ann) =
          let
            val name = tyvProd ()
            val ty = KsiAbs.TyVar
                       {name = name,
                        restr = restr,
                        ann = ann}
            
            val _ = bound := SOME ty
          in
            ty
          end

    fun checkBindAsTyVar (name, metaE, defaultTy, tyvProd, bound, restr, ann) =
          case ST.lookup metaE (ST.symbol name) of
            NONE =>
              bindAsTyVar (tyvProd, bound, restr, ann)
          | SOME restr =>
              defaultTy

    (* Function for doing initial tidying up of type by mapping bottom-up *)
    fun closureTidyNoInstMapF (ty as (KsiAbs.TyMeta {name, restr, bound, ann})) =
           (case !bound of
              NONE =>
                (* Unbound meta variable *)
                (* Check restriction *)
                (case restr of
                   KsiAbs.RstrTyPats {pats} =>
                     (case pats of
                        nil =>
                          reportInternalError 0wx013e
                            "Restriction is an empty set of type patterns!"
                      | [pat] =>
                          (* This is a restriction of a single concrete type *)
                          let
                            (* Replace with conrete type *)
                            val ty = restrTyPatToTy pat
                            val _ = bound := SOME ty
                          in
                            ty
                          end
                      | _ =>
                          (* OK, this is a proper restriction *)
                          ty
                     )
                 | _ =>
                     (* OK, this is a proper restriction *)
                     ty
                )
            | SOME ty' =>
                (reportInternalErrorStrs 0wx013f
                   ["closureTidyNoInstMapF: mapEmbedTy should be transitive!"];
                 (* Bound meta variable, replace with what it is bound to *)
                 ty'
                )
           )
      | closureTidyNoInstMapF (KsiAbs.TyExtRecord {fields, ext, ann}) =
          let
            (* Remove all absent fields *)
            val fields' = List.filter
                            (fn {lab, ty, ann} => not (isAbsentField ty))
                            fields
          in
            case ext of
              KsiAbs.TyExtRecord
                  {fields = nestedFields, ext = nestedExt, ann = _} =>
                (* Flatten extensible records *)
                KsiAbs.TyExtRecord
                  {fields = makeUnionSortedSets
                              ltRecordField eqRecordField fields' nestedFields,
                   ext = nestedExt,
                   ann = ann}
            | _ =>
                (* Record is already as flattened as it can be *)
                KsiAbs.TyExtRecord
                  {fields = fields', ext = ext, ann = ann}
          end
      | closureTidyNoInstMapF ty =
          (* All other types are kept as they are *)
          ty

    (* Do the initial tidying up of a type for the closure operation *)
    fun closureTidyNoInst ty =
          mapEmbedTy closureTidyNoInstMapF ty

    (* Function for final tidying up type by mapping bottom-up *)
    fun closureTidyMapF (tyvProd, metaE) (ty as (KsiAbs.TyMeta {name, restr, bound, ann})) =
           (case !bound of
              NONE =>
                (* Unbound meta variable *)
                checkBindAsTyVar (name, metaE, ty, tyvProd, bound, restr, ann)
            | SOME ty' =>
                (reportInternalErrorStrs 0wx0140
                   ["closureTidyMapF: mapEmbedTy should be transitive!"];
                 (* Bound meta variable, replace with what it is bound to *)
                 ty'
                )
           )
      | closureTidyMapF _ ty =
          (* All other types are kept as they are *)
          ty

    (* Do the final tidying up of a type for the closure operation *)
    fun closureTidy (tyvProd, metaE) ty =
          mapEmbedTy (closureTidyMapF (tyvProd, metaE)) ty

    (* Check if expression is non-expansive constructor expression,
       for value restriction *)
    fun isNonExpansiveConsExp (KsiAbs.ExpTyped {exp, ty, ann}) =
          isNonExpansiveConsExp exp
      | isNonExpansiveConsExp (KsiAbs.ExpCon {tag, tyScheme, ann}) =
          (* Constructors are non-expansive,
             if they are not the special ref constructor. *)
          tag <> "ref"
      | isNonExpansiveConsExp (KsiAbs.ExpEx {tag, ty, ann}) =
          true
      | isNonExpansiveConsExp _ =
          (* Anything else is not a constructor expression *)
          false
    
    (* Check if expression is non-expansive, for value restriction *)
    fun isNonExpansiveExp (KsiAbs.ExpFun {param, exp, ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpApp {funExp, argExp, ann}) =
          (isNonExpansiveConsExp funExp) andalso
          (isNonExpansiveExp argExp)
      | isNonExpansiveExp (KsiAbs.ExpVar {name, ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpTyped {exp, ty, ann}) =
          isNonExpansiveExp exp
      | isNonExpansiveExp (KsiAbs.ExpCon {tag, tyScheme, ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpEx {tag, ty, ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpSCon {scon, ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpEmptyRecord {ann}) =
          true
      | isNonExpansiveExp (KsiAbs.ExpExtRecord {fields, ext, ann}) =
          (isNonExpansiveExp ext) andalso
          (isNonExpansiveFields fields)
      | isNonExpansiveExp (KsiAbs.ExpOptField {present, exp, ann}) =
          (isNonExpansiveExp present) andalso
          (isNonExpansiveExp exp)
      | isNonExpansiveExp _ =
          (* Anything else is expansive *)
          false

    and isNonExpansiveFields fields =
          foldl (fn ({lab, exp, ann}, acc) => acc andalso isNonExpansiveExp exp)
                true fields

    (* Convert symbol table of type variables to a type parameter list *)
    fun tyvarsToParams tyvE =
          let
            val indices = ST.dimensions tyvE

            fun makeParam index =
                  case ST.lookup tyvE index of
                    NONE =>
                      (reportInternalErrorStrs 0wx0141
                         ["tyvarsToParams: Type variable ", ST.name index,
                          " not found where it came from"];
                       {name = ST.name index, restr = KsiAbs.RstrNone})
                  | SOME restr =>
                      {name = ST.name index, restr = restr}
          in
            map makeParam indices
          end

    (* Convert symbol table of type variables to a constructor argument list *)
    fun tyvarsToConArgs tyvE =
          let
            val indices = ST.dimensions tyvE

            fun makeArg index =
                  case ST.lookup tyvE index of
                    NONE =>
                      (reportInternalErrorStrs 0wx0142
                         ["tyvarsToConArgs: Type variable ", ST.name index,
                          " not found where it came from"];
                       {name = ST.name index,
                        ty = KsiAbs.TyVar
                               {name = ST.name index,
                                restr = KsiAbs.RstrNone,
                                ann = ()}})
                  | SOME restr =>
                      {name = ST.name index,
                       ty = KsiAbs.TyVar
                              {name = ST.name index,
                               restr = restr,
                               ann = ()}}
          in
            map makeArg indices
          end

    fun addMetaVarsForExpansiveF tyvE binds (vid, {params, ty, ann}, accMetaE) =
          (case findExpForVidBinding (ST.name vid) binds of
             NONE =>
               (reportInternalErrorStrs 0wx0144
                  ["Expression for variable ", ST.name vid,
                   " not found in KsiTypeInfer.addMetaVarsForExpansiveF"];
                accMetaE)
           | SOME exp =>
               (* Check value restriction on exp *)
               if isNonExpansiveExp exp then
                 accMetaE
               else
                 ST.append (accMetaE, metavarsTy ty)
          )

    (* The closure operation on a type in a binding, which assumes
       that initial tidying up of types is done and that metaE includes
       meta variables in expansive expressions. *)
    fun closureT errPos tyvProd tyvE metaE isTop binds (vid, {params, ty, ann}) =
          let
            (* Tidy up type for closure *)
            val ty' = closureTidy (tyvProd, metaE) ty
            (* Get type parameters and exclude type variables in tyvE *)
            val tyvE' = tyvarsExclude (tyvarsTy ty') tyvE
            (* Convert to a parameter list *)
            val params' = tyvarsToParams tyvE'
            
            (* Check value restriction *)
            val _ = case findExpForVidBinding (ST.name vid) binds of
                      NONE =>
                        reportInternalErrorAtPosStrs errPos 0wx0145
                          ["Expression for variable ", ST.name vid,
                           " not found in KsiTypeInfer.closureT"]
                    | SOME exp =>
                        (* Check value restriction on exp *)
                        (case params' of
                           nil =>
                             () (* This is OK *)
                         | _ =>
                             if isNonExpansiveExp exp then
                               () (* This is OK *)
                             else
                               reportErrorAtPosStrs errPos 0wx0146
                                 ["The expression for the bound variable ", ST.name vid,
                                  " is expansive and thus violates the value restriction. Its type is:\n",
                                  tyToStr ty']
                        )

            (* Here we print the type of declared values, if desired *)
            val show = case !showDescrs of
                         PDNone =>
                           false
                       | PDTopLevel =>
                           isTop
                       | PDNested =>
                           true
            
            val _ = if show then
                      print (String.concat
                               ["val ", ST.name vid, " :\n",
                                tyToStr ty', "\n"])
                    else
                      ()
          in
            {params = params',
             ty = ty',
             ann = ann}
          end

    (* The closure operation for a value environment *)
    fun closureVE errPos (tyvE, tyvE', nestVE, isTop, binds) vE =
          let
            val tyvProd = makeTyvarProducer tyvE binds

            val vE' = ST.map (fn {params, ty, ann} =>
                                {params = params,
                                 ty = closureTidyNoInst ty,
                                 ann = ann}) vE

            val metaE = if isTop then
                          ST.empty () (* nestVE is supposed to be empty *)
                        else
                          ST.foldli (addMetaVarsForExpansiveF tyvE binds)
                                    (getVEMetaVars nestVE) vE'
          in
            ST.mapi (closureT errPos tyvProd tyvE metaE isTop binds) vE'
          end

    (* Rules for implicitly scoped type variables for xi-Calculus *)
    (* The combineScoped operator *)
    fun combineScopedF tyvE (symbol, accTyvE) =
          (case ST.lookup tyvE symbol of
             NONE =>
               (reportInternalErrorStrs 0wx0148
                  ["combineScopedF: Could not get symbol from the symbol table ",
                   "where the symbol itself comes from"];
                accTyvE)
           | SOME restr =>
               (case ST.lookup accTyvE symbol of
                  NONE =>
                    (* The symbol does not exists yet, so just insert it *)
                    ST.insert symbol restr accTyvE
                | SOME restr2 =>
                    (if eqRestrict (restr, restr2) then
                       (* The symbol already exists with the right restriction *)
                       ()
                     else
                       reportErrorStrs 0wx0149
                         ["The type variable ", ST.name symbol, " was ",
                          "encountered more than once in the same implicit ",
                          "scope with differing restrictions"];
                     accTyvE)
               )
          )

    fun combineScoped (tyvE, tyvE') =
          foldl (combineScopedF tyvE) tyvE' (ST.dimensions tyvE)

    (* Inference rules for implicitly scoped type variables *)
    (* Implicitly scoped type variables in expressions *)
    fun implScopedTySpec tyvE (KsiAbs.TyVar {name, restr, ann}) =
          (case ST.lookup tyvE (ST.symbol name) of
             SOME restr2 =>
               (if eqRestrict (restr, restr2) then
                  () (* OK *)
                else
                  reportErrorStrs 0wx014a
                    ["The type variable ", name, " encountered where ",
                     "it is already implicitly scoped with a ",
                     "different restriction"];
                ST.empty ())
           | NONE =>
               ST.insert (ST.symbol name) restr (ST.empty ())
          )
      | implScopedTySpec tyvE (KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               ST.empty ()
           | SOME ty =>
               implScopedTySpec tyvE ty
          )
      | implScopedTySpec tyvE (KsiAbs.TyFun {from, to, ann}) =
          combineScoped
            (implScopedTySpec tyvE from,
             implScopedTySpec tyvE to)
      | implScopedTySpec tyvE (KsiAbs.TyField {present, ty, ann}) =
          combineScoped
            (implScopedTySpec tyvE present,
             implScopedTySpec tyvE ty)
      | implScopedTySpec tyvE (KsiAbs.TyCon {args, name, tags, ann}) =
          foldl (fn ({name, ty}, accTyvE) =>
                   combineScoped (accTyvE, implScopedTySpec tyvE ty))
                (ST.empty ()) args
      | implScopedTySpec tyvE (KsiAbs.TyEmptyRecord {ann}) =
          ST.empty ()
      | implScopedTySpec tyvE (KsiAbs.TyExtRecord {fields, ext, ann}) =
          (* The premise that there must be at least one
             field is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({lab, ty, ann}, accTyvE) =>
                   combineScoped (accTyvE, implScopedTySpec tyvE ty))
                (implScopedTySpec tyvE ext) fields

    (* Implicitly scoped type variables in patterns *)
    fun implScopedPat tyvE (KsiAbs.PatVar {name, ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatLayered {vid, pat, ann}) =
          implScopedPat tyvE pat
      | implScopedPat tyvE (KsiAbs.PatTyped {pat, ty, ann}) =
          combineScoped
            (implScopedPat tyvE pat,
             implScopedTySpec tyvE ty)
      | implScopedPat tyvE (KsiAbs.PatCon {tag, tyScheme, ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatParamCon {tag, tyScheme, pat, ann}) =
          implScopedPat tyvE pat
      | implScopedPat tyvE (KsiAbs.PatEx {tag, ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatParamEx {tag, pat, ty, ann}) =
          combineScoped
            (implScopedPat tyvE pat,
             implScopedTySpec tyvE ty)
      | implScopedPat tyvE (KsiAbs.PatSCon {scon, ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatWild {ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatEmptyRecord {ann}) =
          ST.empty ()
      | implScopedPat tyvE (KsiAbs.PatExtRecord {fields, ext, ann}) =
          (* The premise that there must be at least one
             field is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({lab, pat, ann}, accTyvE) =>
                   combineScoped (accTyvE, implScopedPat tyvE pat))
                (implScopedPat tyvE ext) fields
      | implScopedPat tyvE (KsiAbs.PatOptField {present, pat, ann}) =
          (* The premise that present must be a present constructor
             is checked elsewhere, so we don't bother with it here. *)
          implScopedPat tyvE pat

    (* Implicitly scoped type variables in expressions *)
    fun implScopedExp tyvE (KsiAbs.ExpFun {param, exp, ann}) =
          implScopedExp tyvE exp
      | implScopedExp tyvE (KsiAbs.ExpApp {funExp, argExp, ann}) =
          combineScoped
            (implScopedExp tyvE funExp,
             implScopedExp tyvE argExp)
      | implScopedExp tyvE (KsiAbs.ExpVar {name, ann}) =
          ST.empty ()
      | implScopedExp tyvE (KsiAbs.ExpTyped {exp, ty, ann}) =
          combineScoped
            (implScopedExp tyvE exp,
             implScopedTySpec tyvE ty)
      | implScopedExp tyvE (KsiAbs.ExpCon {tag, tyScheme, ann}) =
          ST.empty ()
      | implScopedExp tyvE (KsiAbs.ExpEx {tag, ty, ann}) =
          implScopedTySpec tyvE ty
      | implScopedExp tyvE (KsiAbs.ExpSCon {scon, ann}) =
          ST.empty ()
      | implScopedExp tyvE (KsiAbs.ExpEmptyRecord {ann}) =
          ST.empty ()
      | implScopedExp tyvE (KsiAbs.ExpExtRecord {fields, ext, ann}) =
          (* The premise that there must be at least one
             field is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({lab, exp, ann}, accTyvE) =>
                   combineScoped (accTyvE, implScopedExp tyvE exp))
                (implScopedExp tyvE ext) fields
      | implScopedExp tyvE (KsiAbs.ExpOptField {present, exp, ann}) =
          combineScoped
            (implScopedExp tyvE present,
             implScopedExp tyvE exp)
      | implScopedExp tyvE (KsiAbs.ExpLet {binds, exp, ann}) =
          implScopedExp tyvE exp
      | implScopedExp tyvE (KsiAbs.ExpLetRec {binds, exp, ann}) =
          (* The premise that there must be at least one
             binding is checked elsewhere, so we don't bother with it here. *)
          implScopedExp tyvE exp
      | implScopedExp tyvE (KsiAbs.ExpCase {exp, matches, ann}) =
          (* The premise that there must be at least one
             match is checked elsewhere, so we don't bother with it here. *)
          implScopedMatches tyvE (implScopedExp tyvE exp) matches
      | implScopedExp tyvE (KsiAbs.ExpFieldCase {exp, tyvar, absExp,
                                                 prePat, preExp, tyOut, ann}) =
          let
            val abspre = makeAbsentPresentRestr ()
            val tyvarTyvE =
                  case ST.lookup tyvE (ST.symbol tyvar) of
                    SOME restr =>
                      (if eqRestrict (restr, abspre) then
                         () (* OK *)
                       else
                         reportErrorStrs 0wx014b
                           ["The type variable ", tyvar, " encountered in a ",
                            "fieldcase expression, while ",
                            "it is already implicitly scoped with a ",
                            "restriction which is not absent | present"];
                       ST.empty ())
                  | NONE =>
                      ST.insert (ST.symbol tyvar) abspre (ST.empty ())
          in
            combineScoped
              (combineScoped
                 (implScopedTySpec tyvE tyOut, tyvarTyvE),
               combineScoped
                 (combineScoped
                    (implScopedExp tyvE exp, implScopedExp tyvE absExp),
                  combineScoped
                    (implScopedPat tyvE prePat, implScopedExp tyvE preExp))
              )
          end
      | implScopedExp tyvE (KsiAbs.ExpLetEx {binds, exp, ann}) =
          (* The premise that there must be at least one
             binding is checked elsewhere, so we don't bother with it here. *)
          foldl (fn ({tag, ty, ann}, accTyvE) =>
                   combineScoped (accTyvE, implScopedTySpec tyvE ty))
                (implScopedExp tyvE exp) binds
      | implScopedExp tyvE (KsiAbs.ExpHandle {exp, matches, ann}) =
          (* The premise that there must be at least one
             match is checked elsewhere, so we don't bother with it here. *)
          implScopedMatches tyvE (implScopedExp tyvE exp) matches
      | implScopedExp tyvE (KsiAbs.ExpRaise {exp, ann}) =
          implScopedExp tyvE exp

    and implScopedMatches tyvE initTyvE matches =
          foldl (fn ({pat, exp, ann}, accTyvE) =>
                   combineScoped
                     (accTyvE,
                      combineScoped
                        (implScopedPat tyvE pat,
                         implScopedExp tyvE exp)))
                initTyvE matches

    (* Implicitly scoped type variables in let and letrec bindings.
       This is the function needed for the semantic rules of
       let and letrec. *)
    fun implScopedBinds tyvE binds =
          foldl (fn ({pat, exp, ann}, accTyvE) =>
                   combineScoped
                     (accTyvE,
                      combineScoped
                        (implScopedPat tyvE pat,
                         implScopedExp tyvE exp)))
                (ST.empty ()) binds

    (* Inference rules for xi-Calculus *)
    (* Get the type of a constant expression *)
    fun typeOfSCon (KsiAbs.SCReal _) =
          KsiAbs.TyCon
            {args = nil, name = "real", tags = ["0.0"], ann = ()}
      | typeOfSCon (KsiAbs.SCInt _) =
          KsiAbs.TyCon
            {args = nil, name = "int", tags = ["0"], ann = ()}
      | typeOfSCon (KsiAbs.SCWord _) =
          KsiAbs.TyCon
            {args = nil, name = "word", tags = ["0w0"], ann = ()}
      | typeOfSCon (KsiAbs.SCString _) =
          KsiAbs.TyCon
            {args = nil, name = "string", tags = ["\"\""], ann = ()}
      | typeOfSCon (KsiAbs.SCChar _) =
          KsiAbs.TyCon
            {args = nil, name = "char", tags = ["#\"a\""], ann = ()}

    (* Check that 2 type patterns are disjoint *)
    fun checkTwoTyPatsDisjoint
            (KsiAbs.RstrTyPat {args, name, tags},
             KsiAbs.RstrTyPat {args = args2, name = name2, tags = tags2}) =
          if name = name2 then
            (* Check that args have the same parameter variables,
               and that at least one of them are disjoint in
               args and args2. *)
            let
              val argNames = map #name args
              val argNames2 = map #name args2
            in
              if isEqualSets op= argNames argNames2 then
                (* Parameter variables are the same, check that
                   at least one pair of the arguments are disjoint. *)
                checkTwoTyPatsDisjointTyConArgs args args2
              else
                (reportErrorStrs 0wx014c
                   ["The type name ", name, " occurs two times with ",
                    "a differing set of parameter variables"];
                 false)
            end
          else
            (* Two different type names are disjoint *)
            true

    and checkTwoTyPatsDisjointTyConArgs args args2 =
          foldl (fn ({name, pat}, acc) =>
                   acc orelse
                   (checkTwoTyPatsDisjoint (pat, getTyPatArg args2 name)))
                false args

    (* Check that type patterns are mutually disjoint *)
    fun checkTyPatsMutuallyDisjointWithTyPat pat pats =
          foldl (fn (pat2, acc) => acc andalso
                                   (checkTwoTyPatsDisjoint (pat, pat2)))
                true pats

    fun checkTyPatsMutuallyDisjointIter (pat::pats) =
          (checkTyPatsMutuallyDisjointWithTyPat pat pats) andalso
          checkTyPatsMutuallyDisjointIter pats
      | checkTyPatsMutuallyDisjointIter nil =
          true

    fun checkTyPatsMutuallyDisjoint (pats as (_::_)) =
          checkTyPatsMutuallyDisjointIter pats
      | checkTyPatsMutuallyDisjoint _ =
          (reportErrorStrs 0wx014d
             ["Encountered list of alternative type patterns in a restriction ",
              "which does not have at least one pattern as required"];
           false)

    (* Check that a restriction is well-formed *)
    fun checkRestr (KsiAbs.RstrNone) =
          true
      | checkRestr (KsiAbs.RstrExclude {labs}) =
          true
      | checkRestr (KsiAbs.RstrTyPats {pats}) =
          (* The function also checks that there is at least 1 pat *)
          checkTyPatsMutuallyDisjoint pats

    (* Check that an optional field type specification is valid *)
    fun checkTySpecOpt errPos tyvE (KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               let
                 val abspre = makeAbsentPresentRestr ()
               in
                 if moreRestrictive (restr, abspre) then
                   NoErr
                 else if moreRestrictive (abspre, restr) then
                   (bound := SOME (newMetaTyRestr abspre);
                    NoErr)
                 else
                   (reportErrorAtPosStrs errPos 0wx014e
                      ["Restriction is not compatiple with an optional field: ",
                       KsiPrettyPrint.restrToStr restr];
                    ReportedErr)
               end
           | SOME ty =>
               checkTySpecOpt errPos tyvE ty
          )
      | checkTySpecOpt errPos tyvE (KsiAbs.TyVar {name, restr, ann}) =
          (case ST.lookup tyvE (ST.symbol name) of
             NONE =>
               (reportErrorAtPosStrs errPos 0wx014f
                  ["Type variable ", name, " occurs explicitly in an ",
                   "optional field type specification, but it is not bound as an ",
                   "implicitly scoped type variable as it should be"];
                ReportedErr)
           | SOME restrict =>
               if eqRestrict (restr, restrict) then
                 if moreRestrictive (restr, makeAbsentPresentRestr ()) then
                   NoErr
                 else
                   (reportErrorAtPosStrs errPos 0wx0150
                      ["Type variable ", name, " occurs explicitly in an ",
                       "optional field type specification, but it has a ",
                       "restriction which is not appropriate for an ",
                       "optional field"];
                    ReportedErr)
               else
                 (reportErrorAtPosStrs errPos 0wx0151
                    ["Type variable ", name, " occurs explicitly in an ",
                     "optional field type specification, but it is bound as an ",
                     "implicitly scoped type variable with a ",
                     "different restriction"];
                  ReportedErr)
          )
      | checkTySpecOpt errPos tyvE (KsiAbs.TyCon {args, name, tags, ann}) =
          if (name = "absent") orelse
             (name = "present") then
            NoErr
          else
            NewErr
      | checkTySpecOpt errPos tyvE _ =
          NewErr

    (* Check that an end-of-record type specification is valid *)
    fun checkTySpecEndRec errPos tyvE exclLabs (KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               let
                 val excl = KsiAbs.RstrExclude {labs = exclLabs}
               in
                 if moreRestrictive (restr, excl) then
                   NoErr
                 else if moreRestrictive (excl, restr) then
                   (bound := SOME (newMetaTyRestr excl);
                    NoErr)
                 else
                   (reportErrorAtPosStrs errPos 0wx0152
                      ["Restriction is not compatiple with a record extension: ",
                       KsiPrettyPrint.restrToStr restr];
                    ReportedErr)
               end
           | SOME ty =>
               checkTySpecEndRec errPos tyvE exclLabs ty
          )
      | checkTySpecEndRec errPos tyvE exclLabs (KsiAbs.TyVar {name, restr, ann}) =
          (case ST.lookup tyvE (ST.symbol name) of
             NONE =>
               (reportErrorAtPosStrs errPos 0wx0153
                  ["Type variable ", name, " occurs explicitly in an ",
                   "end-of-record type specification, but it is not bound as an ",
                   "implicitly scoped type variable as it should be"];
                ReportedErr)
           | SOME restrict =>
               if eqRestrict (restr, restrict) then
                 if moreRestrictive
                      (restr, KsiAbs.RstrExclude {labs = exclLabs}) then
                   NoErr
                 else
                   (reportErrorAtPosStrs errPos 0wx0154
                      ["Type variable ", name, " occurs explicitly in an ",
                       "end-of-record type specification, but it has a ",
                       "restriction which is not appropriate for the ",
                       "extensible record!\nThe record requires at least the restriction:\n",
                       KsiPrettyPrint.restrToStr
                         (KsiAbs.RstrExclude {labs = exclLabs}),
                       "\nbut the restriction of ", name," is:\n",
                       KsiPrettyPrint.restrToStr restrict];
                    ReportedErr)
               else
                 (reportErrorAtPosStrs errPos 0wx0155
                    ["Type variable ", name, " occurs explicitly in an ",
                     "end-of-record type specification, but it is bound as an ",
                     "implicitly scoped type variable with a ",
                     "different restriction"];
                  ReportedErr)
          )
      | checkTySpecEndRec errPos tyvE exclLabs (KsiAbs.TyEmptyRecord _) =
          NoErr
      | checkTySpecEndRec errPos tyvE exclLabs (KsiAbs.TyExtRecord {fields, ext, ann}) =
          (* This clause of the function should never be used, since
             we always check all type specifications in the
             inference rules before any other unification is done. *)
          let
            (* fields are assumed to be sorted *)
            val sortedLabs = map #lab fields
          in
            if Util.hasSortedDuplicates op= sortedLabs then
              (reportErrorAtPosStrs errPos 0wx0156
                 ["Duplicate record labels in type specification: ",
                  KsiPrettyPrint.labsToStr
                    (Util.getSortedDuplicates op= sortedLabs)];
               ReportedErr)
            else
              (* Labels are distinct as required *)
              case sortedLabs of
                (_::_) =>
                  (* There is at least one label as required *)
                  if isDisjointSortedSets op< op= sortedLabs exclLabs then
                    (* Labels are disjoint as required *)
                    combineErrorCodes
                      (checkTySpecEndRec errPos tyvE
                         (makeUnionSortedSets op< op= sortedLabs exclLabs) ext,
                       foldl (fn (f, acc) =>
                                 combineErrorCodes
                                   (acc, checkTySpecRecordField errPos tyvE f))
                             NoErr fields)
                  else
                    (reportErrorAtPosStrs errPos 0wx0157
                       ["Type specification contains the following labels, ",
                        "which are already in the record: ",
                        KsiPrettyPrint.labsToStr
                          (makeIntersectSortedSets op< op= exclLabs sortedLabs)];
                     ReportedErr)
              | nil =>
                  (reportErrorAtPosStrs errPos 0wx0158
                     ["Encountered type specification with ",
                      "extensible record without labels"];
                   ReportedErr)
          end
      | checkTySpecEndRec errPos tyvE _ _ =
          NewErr

    (* Check that a type specification of an extensible record field
       is valid. The fields may be the result of some other parts of the
       unification, so we have to unify the field type with a record field. *)
    and checkTySpecRecordField errPos tyvE {lab, ty, ann} =
          let
            val presentTy = newMetaTyRestr (makeAbsentPresentRestr ())
            val tyTy = newMetaTy ()
            val fieldTy = KsiAbs.TyField
                            {present = presentTy,
                             ty = tyTy,
                             ann = ()}
            val uniErr = unifyReport errPos (fieldTy, ty)
          in
            combineErrorCodes
              (combineErrorCodes
                 (checkTySpecOpt errPos tyvE presentTy,
                  checkTySpec errPos tyvE tyTy),
               uniErr)
          end

    (* Check that a type specification is valid *)
    and checkTySpec errPos tyvE (KsiAbs.TyMeta {name, restr, bound, ann}) =
          (case !bound of
             NONE =>
               NoErr
           | SOME ty =>
               checkTySpec errPos tyvE ty
          )
      | checkTySpec errPos tyvE (KsiAbs.TyVar {name, restr, ann}) =
          (case ST.lookup tyvE (ST.symbol name) of
             NONE =>
               (reportErrorAtPosStrs errPos 0wx0159
                  ["Type variable ", name, " occurs explicitly in a ",
                   "type specification, but it is not bound as an ",
                   "implicitly scoped type variable as it should be"];
                ReportedErr)
           | SOME restrict =>
               if eqRestrict (restr, restrict) then
                 NoErr
               else
                 (reportErrorAtPosStrs errPos 0wx015a
                    ["Type variable ", name, " occurs explicitly in a ",
                     "type specification, but it is bound as an ",
                     "implicitly scoped type variable with a ",
                     "different restriction"];
                  ReportedErr)
          )
      | checkTySpec errPos tyvE (KsiAbs.TyFun {from, to, ann}) =
          combineErrorCodes
            (checkTySpec errPos tyvE from,
             checkTySpec errPos tyvE to)
      | checkTySpec errPos tyvE (KsiAbs.TyCon {args, name, tags, ann}) =
          foldl (fn ({name, ty}, acc) =>
                   combineErrorCodes (acc, checkTySpec errPos tyvE ty))
                NoErr args
      | checkTySpec errPos tyvE (KsiAbs.TyField {present, ty, ann}) =
          combineErrorCodes
            (checkTySpecOpt errPos tyvE present,
             checkTySpec errPos tyvE ty)
      | checkTySpec errPos tyvE (KsiAbs.TyEmptyRecord {ann}) =
          NoErr
      | checkTySpec errPos tyvE (KsiAbs.TyExtRecord {fields, ext, ann}) =
          let
            (* fields are assumed to be sorted *)
            val sortedLabs = map #lab fields
          in
            if Util.hasSortedDuplicates op= sortedLabs then
              (reportErrorAtPosStrs errPos 0wx015b
                 ["Duplicate record labels in type specification: ",
                  KsiPrettyPrint.labsToStr
                    (Util.getSortedDuplicates op= sortedLabs)];
               ReportedErr)
            else
              (* Labels are distinct as required *)
              case sortedLabs of
                (_::_) =>
                  (* There is at least one label as required *)
                  combineErrorCodes
                    (checkTySpecEndRec errPos tyvE sortedLabs ext,
                     foldl (fn (f, acc) =>
                              combineErrorCodes
                                (acc, checkTySpecRecordField errPos tyvE f))
                           NoErr fields)
              | nil =>
                  (reportErrorAtPosStrs errPos 0wx015c
                     ["Encountered type specification with ",
                      "extensible record without labels"];
                   ReportedErr)
          end

    (* Do type inference on a xi-Calculus pattern while
       producing an identical pattern annotated with types. *)
    (* nestVE is an optimization (compared to the sematics) which
       contains the environment variables introduced nested from top-level.
       This makes it faster to find all meta variables in the environment. *)
    fun transPat (vE, tyvE, nestVE) (KsiAbs.PatVar {name, ann = {pos}}) =
          let
            val annTy = newMetaTy ()
            val ann' = {pos = pos, ty = annTy}

            (* vE binds type schemes. Representing the type as
               an unparametrized type scheme makes the generalizing
               type scheme work as if it is just a type. *)
            val tyScheme = {params = nil, ty = annTy, ann = ()}
          in
            (KsiAbs.PatVar {name = name, ann = ann'},
             annTy,
             ST.insert (ST.symbol name) tyScheme vE,
             ST.insert (ST.symbol name) tyScheme nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatLayered {vid, pat, ann = {pos}}) =
          let
            val (pat', ty, vE', nestVE') = transPat (vE, tyvE, nestVE) pat

            (* vE binds type schemes. Representing the type as
               an unparametrized type scheme makes the generalizing
               type scheme work as if it is just a type. *)
            val tyScheme = {params = nil, ty = ty, ann = ()}
          in
            (KsiAbs.PatLayered {vid = vid, pat = pat', ann = {pos = pos, ty = ty}},
             ty,
             ST.insert (ST.symbol vid) tyScheme vE',
             ST.insert (ST.symbol vid) tyScheme nestVE')
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatTyped {pat, ty, ann = {pos}}) =
          (* Here we check the type specification before any
             other unification is done. This means that, part
             of the function checkTySpecEndRec will not be used. *)
          (case checkTySpec pos tyvE ty of
             NoErr =>
               ()
           | NewErr =>
               reportErrorAtPosStrs pos 0wx015d
                 ["Pattern contains invalid type specification:\n",
                  tyToStr ty]
           | ReportedErr =>
               reportErrorFurtherStrs
                 ["Error occurred in pattern with type specification:\n",
                  tyToStr ty];
           let
             val (pat', ty', vE', nestVE') = transPat (vE, tyvE, nestVE) pat
             val _ = unifyReport pos (ty, ty')
           in
             (KsiAbs.PatTyped
                {pat = pat',
                 ty = ty,
                 ann = {pos = pos, ty = ty'}},
              ty',
              vE',
              nestVE')
           end)
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatCon {tag, tyScheme, ann = {pos}}) =
          let
            (* The tyScheme must generalize the type *)
            val annTy = generalizingTyScheme tyScheme
          in
            (KsiAbs.PatCon
               {tag = tag,
                tyScheme = tyScheme,
                ann = {pos = pos, ty = annTy}},
             annTy,
             vE,
             nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatParamCon {tag, tyScheme, pat, ann = {pos}}) =
          let
            (* The tyScheme must generalize the type *)
            val conTy = generalizingTyScheme tyScheme

            val (pat', patTy, vE', nestVE') = transPat (vE, tyvE, nestVE) pat
            val annTy = newMetaTy ()
            val funTy = KsiAbs.TyFun {from = patTy, to = annTy, ann = ()}
            val _ = unifyReport pos (conTy, funTy)
          in
            (KsiAbs.PatParamCon
               {tag = tag,
                tyScheme = tyScheme,
                pat = pat',
                ann = {pos = pos, ty = annTy}},
             annTy,
             vE',
             nestVE')
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatEx {tag, ann = {pos}}) =
          let
            val exnTy = makeExnTyCon ()
          in
            (KsiAbs.PatEx {tag = tag, ann = {pos = pos, ty = exnTy}},
             exnTy,
             vE,
             nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatParamEx {tag, pat, ty, ann = {pos}}) =
          let
            val (pat', patTy, vE', nestVE') = transPat (vE, tyvE, nestVE) pat
            val _ = unifyReport pos (patTy, ty)
            val exnTy = makeExnTyCon ()
          in
            (KsiAbs.PatParamEx
               {tag = tag,
                pat = pat',
                ty = ty,
                ann = {pos = pos, ty = exnTy}},
             exnTy,
             vE',
             nestVE')
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatSCon {scon, ann = {pos}}) =
          let
            val annTy = typeOfSCon scon
          in
            (KsiAbs.PatSCon {scon = scon, ann = {pos = pos, ty = annTy}},
             annTy,
             vE,
             nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatWild {ann = {pos}}) =
          let
            val annTy = newMetaTy ()
          in
            (KsiAbs.PatWild {ann = {pos = pos, ty = annTy}},
             annTy,
             vE,
             nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatEmptyRecord {ann = {pos}}) =
          let
            val annTy = KsiAbs.TyEmptyRecord {ann = ()}
          in
            (KsiAbs.PatEmptyRecord {ann = {pos = pos, ty = annTy}},
             annTy,
             vE,
             nestVE)
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatExtRecord {fields, ext, ann = {pos}}) =
          let
            (* fields are assumed to be sorted *)
            val sortedLabs = map #lab fields

            (* This check is redundant *)
            val _ = if Util.hasSortedDuplicates op= sortedLabs then
                      reportErrorAtPosStrs pos 0wx015f
                        ["Duplicate record labels in record pattern: ",
                         KsiPrettyPrint.labsToStr
                         (Util.getSortedDuplicates op= sortedLabs)]
                    else
                      ()
            
            val _ = case sortedLabs of
                      (_::_) =>
                         (* There is at least one label as required *)
                         ()
                    | nil =>
                        reportErrorAtPos pos 0wx0160
                          "Encountered extensible record pattern without labels"
                  
            (* Unify all field types to be fields
               and produce a list of typed pattern fields *)
            val (typedFields, vE', nestVE') =
                   foldl (transPatRecordFieldToTypedF tyvE)
                         (nil, vE, nestVE) fields
            val typedFields = rev typedFields
            
            val fields' = map (fn {lab, pat, ty, ann} =>
                                 {lab = lab,
                                  pat = pat,
                                  ann = ann})
                              typedFields
            val tyFields = map (fn {lab, pat, ty, ann} =>
                                  {lab = lab,
                                   ty = ty,
                                   ann = ()})
                               typedFields
                                          
            val (ext', extTy, vE'', nestVE'') = transPat (vE', tyvE, nestVE') ext
            val exclRestr = KsiAbs.RstrExclude {labs = sortedLabs}
            val extTyRestr = newMetaTyRestr exclRestr
            val _ = unifyReport pos (extTy, extTyRestr)
            
            val annTy = KsiAbs.TyExtRecord
                          {fields = tyFields,
                           ext = extTy,
                           ann = ()}
          in
            (KsiAbs.PatExtRecord
               {fields = fields',
                ext = ext',
                ann = {pos = pos, ty = annTy}},
             annTy,
             vE'',
             nestVE'')
          end
      | transPat (vE, tyvE, nestVE) (KsiAbs.PatOptField {present, pat, ann = {pos}}) =
          let
            val presentTy = makePresentTyCon ()
            val present' = case present of
                               KsiAbs.PatCon {tag = "present", tyScheme, ann = {pos = pos'}} =>
                                 KsiAbs.PatCon
                                   {tag = "present",
                                    tyScheme = tyScheme,
                                    ann = {pos = pos', ty = presentTy}}
                             | _ =>
                               (reportErrorAtPosStrs pos 0wx0161
                                  ["Encountered optional field pattern which ",
                                   "is not a present field as expected"];
                                (* Here we discard the syntax present and
                                   generate this new syntax instead. *)
                                KsiAbs.PatCon
                                   {tag = "present",
                                    tyScheme = KsiBuiltinStatic.presentTyScheme,
                                    ann = {pos = pos, ty = presentTy}})

            val (pat', patTy, vE', nestVE') = transPat (vE, tyvE, nestVE) pat
            val annTy = KsiAbs.TyField
                          {present = presentTy,
                           ty = patTy,
                           ann = ()}
          in
            (KsiAbs.PatOptField
               {present = present',
                pat = pat',
                ann = {pos = pos, ty = annTy}},
             annTy,
             vE',
             nestVE')
          end

    and transPatRecordFieldToTypedF tyvE ({lab, pat, ann = {pos}}, (acc, vE, nestVE)) =
          let
            val (pat', ty, vE', nestVE') = transPat (vE, tyvE, nestVE) pat
            val newFieldTy = newField ()
            val _ = unifyReport pos (ty, newFieldTy)
          in
            ({lab = lab, pat = pat', ty = newFieldTy,
              ann = {pos = pos, ty = newFieldTy}}
                :: acc,
             vE', nestVE')
          end

    (* Do type inference on a xi-Calculus expression while
       producing an identical expression annotated with types. *)
    (* isTop keeps track of what is top-level let and letrec expressions.
       nestVE is an optimization (compared to the sematics) which
       contains the environment variables introduced nested from top-level.
       This makes it faster to find all meta variables in the environment. *)
    fun transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpFun {param, exp, ann = {pos}}) =
          let
            val paramTy = newMetaTy ()
            val vE' = ST.insert (ST.symbol param)
                                {params = nil, ty = paramTy, ann = ()}
                                vE
            val nestVE' = ST.insert (ST.symbol param)
                                    {params = nil, ty = paramTy, ann = ()}
                                    nestVE
            val (exp', toTy) = transExp (vE', tyvE, false, nestVE') exp
            val annTy = KsiAbs.TyFun {from = paramTy, to = toTy, ann = ()}
          in
            (KsiAbs.ExpFun
               {param = param,
                exp = exp',
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpApp {funExp, argExp, ann = {pos}}) =
          let
            val (argExp', argTy) =  transExp (vE, tyvE, false, nestVE) argExp
            val (funExp', funTy) =  transExp (vE, tyvE, false, nestVE) funExp
            
            val annTy = newMetaTy ()
            val newFunTy = KsiAbs.TyFun
                             {from = argTy,
                              to = annTy,
                              ann = ()}
            val _ = unifyReport pos (funTy, newFunTy)
          in
            (KsiAbs.ExpApp
               {funExp = funExp',
                argExp = argExp',
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpVar {name, ann = {pos}}) =
          let
            val annTy = case ST.lookup vE (ST.symbol name) of
                          NONE =>
                            (reportErrorAtPosStrs pos 0wx0162
                               ["Variable ", name, " not declared"];
                             (* We don't know anything about the type here *)
                             newMetaTy ())
                        | SOME tyScheme =>
                            (* The tyScheme must generalize the type *)
                            generalizingTyScheme tyScheme
          in
            (KsiAbs.ExpVar {name = name, ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpTyped {exp, ty, ann = {pos}}) =
          (* Here we check the type specification before any
             other unification is done. This means that, part
             of the function checkTySpecEndRec will not be used. *)
          (case checkTySpec pos tyvE ty of
             NoErr =>
               ()
           | NewErr =>
               reportErrorAtPosStrs pos 0wx0163
                 ["Expression contains invalid type specification:\n",
                  tyToStr ty]
           | ReportedErr =>
               reportErrorFurtherStrs
                 ["Error occurred in expression with type specification:\n",
                  tyToStr ty];
           let
             val (exp', ty') = transExp (vE, tyvE, false, nestVE) exp
             val _ = unifyReport pos (ty, ty')
           in
             (KsiAbs.ExpTyped
                {exp = exp',
                 ty = ty,
                 ann = {pos = pos, ty = ty'}},
              ty')
           end)
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpCon {tag, tyScheme, ann = {pos}}) =
          let
            (* The tyScheme must generalize the type *)
            val annTy = generalizingTyScheme tyScheme
          in
            (KsiAbs.ExpCon
               {tag = tag,
                tyScheme = tyScheme,
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpEx {tag, ty, ann = {pos}}) =
          (KsiAbs.ExpEx
             {tag = tag,
              ty = ty,
              ann = {pos = pos, ty = ty}},
           ty)
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpSCon {scon, ann = {pos}}) =
          let
            val annTy = typeOfSCon scon
          in
            (KsiAbs.ExpSCon
               {scon = scon,
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpEmptyRecord {ann = {pos}}) =
          let
            val annTy = KsiAbs.TyEmptyRecord {ann = ()}
          in
            (KsiAbs.ExpEmptyRecord {ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpExtRecord {fields, ext, ann = {pos}}) =
          let
            (* fields are assumed to be sorted *)
            val sortedLabs = map #lab fields

            (* This check is redundant *)
            val _ = if Util.hasSortedDuplicates op= sortedLabs then
                      reportErrorAtPosStrs pos 0wx0165
                        ["Duplicate record labels in record expression: ",
                         KsiPrettyPrint.labsToStr
                         (Util.getSortedDuplicates op= sortedLabs)]
                    else
                      ()
            
            val _ = case sortedLabs of
                      (_::_) =>
                         (* There is at least one label as required *)
                         ()
                    | nil =>
                        reportErrorAtPos pos 0wx0166
                          "Encountered extensible record expression without labels"
                  
            (* Unify all field types to be fields
               and produce a list of typed expression fields*)
            val typedFields = map (transExpRecordFieldToTyped (vE, tyvE, nestVE))
                                  fields
            val fields' = map (fn {lab, exp, ty, ann} =>
                                 {lab = lab,
                                  exp = exp,
                                  ann = ann})
                              typedFields
            val tyFields = map (fn {lab, exp, ty, ann} =>
                                  {lab = lab,
                                   ty = ty,
                                   ann = ()})
                               typedFields
                                          
            val (ext', extTy) = transExp (vE, tyvE, false, nestVE) ext
            val exclRestr = KsiAbs.RstrExclude {labs = sortedLabs}
            val extTyRestr = newMetaTyRestr exclRestr
            val _ = unifyReport pos (extTy, extTyRestr)
            
            val annTy = KsiAbs.TyExtRecord
                          {fields = tyFields,
                           ext = extTy,
                           ann = ()}
          in
            (KsiAbs.ExpExtRecord
               {fields = fields',
                ext = ext',
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpOptField {present, exp, ann = {pos}}) =
          let
            val (present', presentTy) = transExp (vE, tyvE, false, nestVE) present
            val (exp', expTy) = transExp (vE, tyvE, false, nestVE) exp
            val newTy = newMetaTyRestr (makeAbsentPresentRestr ())
            val _ = unifyReport pos (presentTy, newTy)

            val annTy = KsiAbs.TyField
                          {present = presentTy,
                           ty = expTy,
                           ann = ()}
          in
            (KsiAbs.ExpOptField
               {present = present',
                exp = exp',
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpLetEx {binds, exp, ann = {pos}}) =
          (* Here we check the type specification before any
             other unification is done. This means that, part
             of the function checkTySpecEndRec will not be used. *)
          (case foldl (fn ({tag, ty, ann}, acc) =>
                         combineErrorCodes (acc, checkTySpec pos tyvE ty))
                      NoErr binds of
             NoErr =>
             (* Exception types are OK *)
             let
               val (exp', ty) = transExp (vE, tyvE, isTop, nestVE) exp
               val _ = case binds of
                         (_::_) =>
                           (* There is at least one exception as required *)
                           ()
                       | nil =>
                           reportErrorAtPos pos 0wx0167
                             "Encountered exception declaration with no exceptions"

               val binds' = map (fn {tag, ty, ann = {pos}} =>
                                   {tag = tag,
                                    ty = ty,
                                    ann = {pos = pos, ty = ty}})
                            binds
             in
               (KsiAbs.ExpLetEx
                  {binds = binds',
                   exp = exp',
                   ann = {pos = pos, ty = ty}},
                ty)
             end
           | NewErr =>
               (reportErrorAtPosStrs pos 0wx0168
                  ["Exception declared with invalid type"];
                transExp (vE, tyvE, false, nestVE) exp)
           | ReportedErr =>
               (reportErrorFurtherStrs
                  ["Exception declared with invalid type"];
                transExp (vE, tyvE, false, nestVE) exp)
          )
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpLet {binds, exp, ann = {pos}}) =
          let
            val tyvE' = implScopedBinds tyvE binds
            val tyvE'' = ST.append (tyvE, tyvE')
            
            val (binds', vE', nestVE') = transBindsPats (tyvE'', nestVE) binds
            (* The expressions in a non-recursive let does not know
               about the values bound by the let. *)
            val binds'' = transBindsExps (vE, tyvE'', nestVE) binds'

            (* Check that all patterns together does not
               bind the same variable more than once. *)
            val _ = checkUniqueVarBindBinds (ST.empty ()) binds

            (* FIXME: Check that each pattern is exhaustive on its type. *)

            (* Do closure operation to produce cvE from vE' *)
            val cvE = closureVE pos (tyvE, tyvE', nestVE, isTop, binds'') vE'
            val cvE' = ST.append (vE, cvE)

            val nextNestVE = if isTop then
                               ST.empty ()
                             else
                               nestVE'
            val (exp', ty) = transExp (cvE', tyvE, isTop, nextNestVE) exp
          in
            (KsiAbs.ExpLet
               {binds = binds'',
                exp = exp',
                ann = {pos = pos, ty = ty}},
             ty)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpLetRec {binds, exp, ann = {pos}}) =
          let
            val tyvE' = implScopedBinds tyvE binds
            val tyvE'' = ST.append (tyvE, tyvE')
            
            val _ = case binds of
                      (_::_) =>
                        ()
                    | nil =>
                        reportErrorAtPosStrs pos 0wx016a
                           ["Encountered a recursive val-binding with no ",
                            "bindings"]

            val (binds', vE', nestVE') = transBindsPats (tyvE'', nestVE) binds
            (* The expressions in letrec knows
               the values bound by the letrec. *)
            val vE'' = ST.append (vE, vE')
            val binds'' = transBindsExps (vE'', tyvE'', nestVE') binds'

            (* Check that all patterns together does not
               bind the same variable more than once. *)
            val _ = checkUniqueVarBindBinds (ST.empty ()) binds

            (* FIXME: Check that each pattern is exhaustive on its type. *)

            (* Do closure operation to produce cvE from vE' *)
            val cvE = closureVE pos (tyvE, tyvE', nestVE, isTop, binds'') vE'
            val cvE' = ST.append (vE, cvE)

            val nextNestVE = if isTop then
                               ST.empty ()
                             else
                               nestVE'
            val (exp', ty) = transExp (cvE', tyvE, isTop, nextNestVE) exp
          in
            (KsiAbs.ExpLetRec
               {binds = binds'',
                exp = exp',
                ann = {pos = pos, ty = ty}},
             ty)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpHandle {exp, matches, ann = {pos}}) =
          let
            val (exp', ty) = transExp (vE, tyvE, false, nestVE) exp
            val exnTy = makeExnTyCon ()
            val matches' = transExpMatches (vE, tyvE, nestVE) (exnTy, ty) matches
            (* FIXME: Check that matches are not redundant on exnTy *)
            val _ = case matches of
                      (_::_) =>
                         (* At least one match as required *)
                         ()
                    | nil =>
                        reportErrorAtPosStrs pos 0wx016b
                          ["Encountered exception handler without any ",
                           "match clauses"]
          in
            (KsiAbs.ExpHandle
               {exp = exp',
                matches = matches',
                ann = {pos = pos, ty = ty}},
             ty)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpRaise {exp, ann = {pos}}) =
          let
            val (exp', ty) = transExp (vE, tyvE, false, nestVE) exp
            val exnTy = makeExnTyCon ()
            val _ = unifyReport pos (ty, exnTy)
            
            val annTy = newMetaTy ()
          in
            (KsiAbs.ExpRaise {exp = exp', ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE) (KsiAbs.ExpCase {exp, matches, ann = {pos}}) =
          let
            val (exp', ty) = transExp (vE, tyvE, false, nestVE) exp
            val annTy = newMetaTy ()
            val matches' = transExpMatches (vE, tyvE, nestVE) (ty, annTy) matches
            (* FIXME: Check that matches are not redundant on ty *)
            (* FIXME: Issue warning if matches are not exhaustive on ty *)
            val _ = case matches of
                      (_::_) =>
                         (* At least one match as required *)
                         ()
                    | nil =>
                        reportErrorAtPosStrs pos 0wx016c
                          ["Encountered case expression without any cases"]
          in
            (KsiAbs.ExpCase
               {exp = exp',
                matches = matches',
                ann = {pos = pos, ty = annTy}},
             annTy)
          end
      | transExp (vE, tyvE, isTop, nestVE)
                 (KsiAbs.ExpFieldCase {exp, tyvar, absExp,
                                       prePat, preExp, tyOut, ann = {pos}}) =
          (* Here we check the type specification before any
             other unification is done. This means that, part
             of the function checkTySpecEndRec will not be used. *)
          let
            val _ = case checkTySpec pos tyvE tyOut of
                      NoErr =>
                        ()
                    | NewErr =>
                        reportErrorAtPosStrs pos 0wx016d
                          ["Fieldcase expression contains invalid type specification:\n",
                           tyToStr tyOut]
                    | ReportedErr =>
                        reportErrorFurtherStrs
                          ["Error occurred in fieldcase expression with invalid type specification:\n",
                           tyToStr tyOut]

            val abspre = makeAbsentPresentRestr ()
            val _ = case ST.lookup tyvE (ST.symbol tyvar) of
                      NONE =>
                        reportErrorAtPosStrs pos 0wx016f
                           ["Type variable ", tyvar, " occurs explicitly in a ",
                            "fieldcase expression, but it is not bound as an ",
                            "implicitly scoped type variable as it should be"]
                    | SOME restrict =>
                        if eqRestrict (abspre, restrict) then
                          ()
                        else
                          reportErrorAtPosStrs pos 0wx0170
                             ["Type variable ", tyvar, " occurs explicitly in a ",
                              "fieldcase expression, but it is bound as an ",
                              "implicitly scoped type variable with a ",
                              "different restriction"]

            val (prePat', prePatTy, vE', nestVE') = transPat (vE, tyvE, nestVE) prePat
            val (exp', expTy) = transExp (vE, tyvE, false, nestVE) exp
            val fieldTy = KsiAbs.TyField
                            {present = KsiAbs.TyVar
                                         {name = tyvar,
                                          restr = abspre,
                                          ann = ()},
                             ty = prePatTy,
                             ann = ()}
            val _ = unifyReport pos (expTy, fieldTy)

            (* Translate absent expression and unify with tyOut
               where tyvar is replaced with the absent type constructor *)
            val (absExp', absExpTy) = transExp (vE, tyvE, false, nestVE) absExp
            val tyvarToAbsentEnv = ST.insert (ST.symbol tyvar)
                                             (makeAbsentTyCon ())
                                             (ST.empty ())
            val absTyOut = instantiateTyVars tyvarToAbsentEnv tyOut
            val _ = unifyReport pos (absExpTy, absTyOut)

            (* Translate present expression and unify with tyOut
               where tyvar is replaced with the present type constructor *)
            val (preExp', preExpTy) = transExp (vE', tyvE, false, nestVE') preExp
            val tyvarToPresentEnv = ST.insert (ST.symbol tyvar)
                                              (makePresentTyCon ())
                                              (ST.empty ())
            val preTyOut = instantiateTyVars tyvarToPresentEnv tyOut
            val _ = unifyReport pos (preExpTy, preTyOut)
                    
            (* FIXME: Check that prePat is exhaustive on prePatTy *)

            (* Check that each variable in prePat occurs only once *)
            val _ = checkUniqueVarBindPat (ST.empty ()) prePat
          in
            (KsiAbs.ExpFieldCase
               {exp = exp',
                tyvar = tyvar,
                absExp = absExp',
                prePat = prePat',
                preExp = preExp',
                tyOut = tyOut,
                ann = {pos = pos, ty = tyOut}},
             tyOut)
          end
          
    and transExpRecordFieldToTyped (vE, tyvE, nestVE) {lab, exp, ann = {pos}} =
          let
            val (exp', ty) = transExp (vE, tyvE, false, nestVE) exp
            val newFieldTy = newField ()
            val _ = unifyReport pos (ty, newFieldTy)
          in
            {lab = lab,
             exp = exp',
             ty = newFieldTy,
             ann = {pos = pos, ty = newFieldTy}}
          end

    and transExpMatch (vE, tyvE, nestVE) (patTy, expTy) {pat, exp, ann = {pos}} =
          let
            (* Check that each variable in pat occurs only once! *)
            val _ = checkUniqueVarBindPat (ST.empty ()) pat

            val (pat', patTy', vE', nestVE') = transPat (vE, tyvE, nestVE) pat
            val _ = unifyReport pos (patTy, patTy')
            val (exp', expTy') = transExp (vE', tyvE, false, nestVE') exp
            val _ = unifyReport pos (expTy, expTy')
            val annTy = KsiAbs.TyFun {from = patTy', to = expTy', ann = ()}
          in
            {pat = pat', exp = exp',
             ann = {pos = pos, ty = annTy}}
          end

    and transExpMatches (vE, tyvE, nestVE) (patTy, expTy) matches =
          map (transExpMatch (vE, tyvE, nestVE) (patTy, expTy)) matches

    and transBindsPatF tyvE ({pat, exp, ann = {pos}}, (accBinds, vE, nestVE)) =
          let
            val (pat', ty, vE', nestVE') = transPat (vE, tyvE, nestVE) pat
          in
            ({pat = pat', exp = exp, ann = {pos = pos, ty = ty}}
                :: accBinds,
             vE', nestVE')
          end

    and transBindsPats (tyvE, nestVE) binds =
          (* vE is initially empty, because the result is appended in
             the rules for let and letrec, but nestVE is just propagated. *)
          foldl (transBindsPatF tyvE) (nil, ST.empty (), nestVE) binds

    and transBindsExpM (vE, tyvE, nestVE) {pat, exp, ann = ann as {pos, ty = annTy}} =
          let
            val (exp', ty) = transExp (vE, tyvE, false, nestVE) exp
            val _ = unifyReport pos (ty, annTy)
          in
            {pat = pat, exp = exp', ann = ann}
          end

    and transBindsExps (vE, tyvE, nestVE) binds =
          map (transBindsExpM (vE, tyvE, nestVE)) binds

    fun transProgramExp (vE, tyvE) exp =
          transExp (vE, tyvE, true, ST.empty ()) exp
  end
